/*  -*- Mode: c++ -*-
 *
 *  $Id: xotcl.c,v 1.37 1999/10/20 16:59:48 neumann Exp $
 *
 *  XOTcl - Extended OTcl
 *
 *  Copyright (C) 1999 Gustaf Neumann, Uwe Zdun
 *
 *     University of Essen
 *     Information Systems and Software Techniques
 *     Universitaetsstr. 9
 *     D-45129 Essen, Germany
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. We make no representations about the suitability 
 *  of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 *
 *
 *  This software is based upon MIT Object Tcl by David Wetherall and
 *  Christopher J. Lindblad, that contains the following copyright message:
 *
 *   "Copyright 1993 Massachusetts Institute of Technology
 *
 *    Permission to use, copy, modify, distribute, and sell this software and its
 *    documentation for any purpose is hereby granted without fee, provided that
 *    the above copyright notice appear in all copies and that both that
 *    copyright notice and this permission notice appear in supporting
 *    documentation, and that the name of M.I.T. not be used in advertising or
 *    publicity pertaining to distribution of the software without specific,
 *    written prior permission.  M.I.T. makes no representations about the
 *    suitability of this software for any purpose.  It is provided "as is"
 *    without express or implied warranty."
 *
 */
#define NDEBUG   /* activate/deacticate assert */

#include <stdlib.h>
#include <string.h>
#include <tclInt.h>
#include "xotcl.h"
#include <assert.h>

/*#define AUTOVARS*/  /* make self, proc and class in instproc and procs */

#undef PROCCACHE /* turn the proc cache on/off */
/*#define PROCCACHE*/

/*
 * Makros
 */

/* for the time being avoid macro string constants with VCC */
#if defined(VISUAL_CC)
#  define XOLIBPKG "/Programme/Tcl/XOTcl"
#  define XOTCLVERSION "0.8"
#endif

/* turn tracing output on/off
#define DISPATCH_TRACE
#define STACK_TRACE
*/

#define ALLOC_NAME_NS(DSP, NS, NAME) \
     Tcl_DStringInit(DSP),\
     Tcl_DStringAppend(DSP, NS, -1),\
     Tcl_DStringAppend(DSP, "::", 2),\
     Tcl_DStringAppend(DSP, NAME, -1)
#define ALLOC_TOP_NS(DSP, NAME) \
     Tcl_DStringInit(DSP),\
     Tcl_DStringAppend(DSP, "::", 2),\
     Tcl_DStringAppend(DSP, NAME, -1)
#define ALLOC_DSTRING(DSP,ENTRY) \
     Tcl_DStringInit(DSP),\
     Tcl_DStringAppend(DSP, ENTRY, -1)

#define nr_elements(arr)  ((int) (sizeof(arr) / sizeof(arr[0])))

#define setOption(value,options,bitToChange) \
    if (value)  options |= bitToChange;      \
    else        options &= ~bitToChange;

#define isArgsString(m) (\
	*m   == 'a' && m[1] == 'r' && m[2] == 'g' && m[3] == 's' && \
	m[4] == '\0')
#define isAutoString(m) (\
	*m   == 'a' && m[1] == 'u' && m[2] == 't' && m[3] == 'o' && \
	m[4] == '\0')
#define isClassString(m) (\
	*m   == 'c' && m[1] == 'l' && m[2] == 'a' && m[3] == 's' && \
	m[4] == 's' && m[5] == '\0')
#define isCreateString(m) (\
	*m   == 'c' && m[1] == 'r' && m[2] == 'e' && m[3] == 'a' && \
	m[4] == 't' && m[5] == 'e' && m[6] == '\0')
#define isDecrString(m) (\
	*m   == 'd' && m[1] == 'e' && m[2] == 'c' && m[3] == 'r' && \
	m[4] == '\0')
#define isDestroyString(m) (\
	*m   == 'd' && m[1] == 'e' && m[2] == 's' && m[3] == 't' && \
	m[4] == 'r' && m[5] == 'o' && m[6] == 'y' && m[7] == '\0')
#define isInitString(m) (\
	*m   == 'i' && m[1] == 'n' && m[2] == 'i' && m[3] == 't' && \
	m[4] == '\0')
#define isIncrString(m) (\
	*m   == 'i' && m[1] == 'n' && m[2] == 'c' && m[3] == 'r' && \
	m[4] == '\0')
#define isLevelString(m) (\
	*m   == 'l' && m[1] == 'e' && m[2] == 'v' && m[3] == 'e' && \
	m[4] == 'l' && m[5] == '\0')
#define isNextString(m) (\
	*m   == 'n' && m[1] == 'e' && m[2] == 'x' && m[3] == 't' && \
	m[4] == '\0')
#define isProcString(m) (\
	*m   == 'p' && m[1] == 'r' && m[2] == 'o' && m[3] == 'c' && \
	m[4] == '\0')

/*
 * prevent old TCL-versions
 */

#if TCL_MAJOR_VERSION < 8
# error Tcl distribution is TOO OLD, we require tcl8.0
#endif

#if TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION<1
# define PRE81
#else
# if TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==1
#  define V81
# endif
#endif
#if TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION<2
# define PRE82
#endif

#ifdef sun
# define USE_ALLOCA
#endif

#if defined(VISUAL_CC)
#  define USE_MALLOC
#endif

#if defined(USE_MALLOC)
#  define DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov) \
    Tcl_Obj** ov = (Tcl_Obj**) ckalloc(oc*sizeof(Tcl_Obj*))
#  define FREE_TCL_OBJECTS_ON_STACK(ov) ckfree((char*) ov)
#  define DEFINE_NEW_CHAR_ARRAY(c,v) \
    char** v = (char**) ckalloc(c*sizeof(char*))
#  define FREE_CHAR_ARRAY(v) ckfree((char*) v)
#elif defined(USE_ALLOCA)
#  define DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov) \
    Tcl_Obj** ov = (Tcl_Obj**) alloca(oc*sizeof(Tcl_Obj*))
#  define FREE_TCL_OBJECTS_ON_STACK(ov)
#  define DEFINE_NEW_CHAR_ARRAY(c,v) \
    char** v = (char**) alloca(c*sizeof(char*))
#  define FREE_CHAR_ARRAY(v)
#else
#  define DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov) \
    Tcl_Obj* ov[oc]
#  define FREE_TCL_OBJECTS_ON_STACK(ov)
#  define DEFINE_NEW_CHAR_ARRAY(c,v) \
    char* v[c]
#  define FREE_CHAR_ARRAY(v)
#endif

#ifdef USE_ALLOCA
# include <alloca.h>
#endif

#ifdef  __WIN32__
#       define XOTCLINLINE
#       undef  NODOUBLECHECK
#       define XOTclNewObj(A) A=Tcl_NewObj()
#	define DECR_REF_COUNT(A) Tcl_DecrRefCount(A)
#else
# ifdef sun
#       define XOTCLINLINE
# else
#       define XOTCLINLINE inline
# endif
#       define NODOUBLECHECK
#       define XOTclNewObj(A) TclNewObj(A)
#	define DECR_REF_COUNT(A) TclDecrRefCount(A)
#endif

#ifdef V81
# define ObjStr(obj) Tcl_GetString(obj)
# define EvalObj(in,cmd) Tcl_EvalObj (in,cmd,0)
# define TclIsVarArgument(args) (args->isArg)
#else
# define ObjStr(obj) Tcl_GetStringFromObj(obj, ((int*)NULL))
# define EvalObj(in,cmd) Tcl_EvalObj (in,cmd)
# if defined(PRE81) && TCL_RELEASE_SERIAL<3
#  define TclIsVarArgument(args) (args->isArg)
# endif
#endif

#define INCR_REF_COUNT(A) Tcl_IncrRefCount(A)

#define XOTCLSMALLARGS 10

/* TCL_CONTINUE is defined as 4, from 5 on we can
   use app-specific return codes */
#define XOTCL_UNKNOWN 5
#define XOTCL_NOT_FILTERED 6
#define XOTCL_STDARGS_COMPUTED 7
#define XOTCL_USE_TCL_COMMAND 8

/*
 * Filter:  list and info-structure for the objects/classes
 */
typedef struct Filter {
  Tcl_Obj* simpleName;
  Command* cmdPtr;
  struct XOTclClass* filterClass;
  struct Filter* next;
} Filter;

typedef struct ObjList {
  struct XOTclObject* objPtr;
  struct ObjList* next;
} ObjList;

typedef struct FilterChainInfo {
  Filter* filterPtr;
  Tcl_Obj* calledProc;
  Tcl_Obj* callingProc;
  Tcl_Obj* callingClass;
  Tcl_Obj* callingObject;
  CallFrame* procVarFramePtr;
  struct XOTclClass* regClass;
  struct FilterChainInfo* next;
  int destroyCalled;
  int actualProcCalled;
} FilterChainInfo;

typedef struct FilterInfo {
  int valid;
  int doFilters;
  struct FilterChainInfo* chains;
} FilterInfo;

Tcl_HashTable XOTclFilterList;

typedef struct XOTclFilterListEntry {
  struct ObjList* activeObjs;
  int count;
} XOTclFilterListEntry;

/*
 * Assertion structures
 */
typedef struct Assertion {
  Tcl_Obj* content;
  struct Assertion* next;
} Assertion;

typedef struct ProcAssertion {
  Assertion* pre;
  Assertion* post;
} ProcAssertion;

typedef struct AssertionStore {
  Assertion* invariants;
  Tcl_HashTable procs;
} AssertionStore;

typedef enum { /* powers of 2; add to ALL, if default; */
  CHECK_NONE = 0, CHECK_CLINVAR = 1, CHECK_OBJINVAR = 2,
  CHECK_PRE= 4, CHECK_POST = 8,
  CHECK_ALL = CHECK_CLINVAR+CHECK_OBJINVAR+CHECK_PRE+CHECK_POST,
} CheckOptions;

/*
 * per-obj mixins
 */
typedef struct Mixin {
  Command* cmdPtr;
  struct Mixin* next;
} Mixin;

typedef struct MixinStack {
  Command* currentCmdPtr;
  int mixinChainOn;
  int mixinInit;
  struct MixinStack* next;
} MixinStack;

typedef struct CmdList {
  Command* cmdPtr;
  struct CmdList* next;
} CmdList;

/*
 * PROCCACHE
 */
#ifdef PROCCACHE
typedef struct ProcCacheEntry {
  Command* cmdPtr;
  struct XOTclClass* cl;
} ProcCacheEntry;
#endif

/*
 * object and class internals
 */
typedef struct XOTclObject {
  Tcl_Obj* cmdName;
  Tcl_Command id;
  int destroyCalled;
  Tcl_Interp* teardown;
  struct XOTclClass* cl;
  struct XOTclClass* type;
  Namespace *nsPtr;
  AssertionStore assertions;
  int checkoptions;
  FilterInfo filterinfo;

  Mixin* mixins;
  struct CmdList* mixinOrder;
  MixinStack* mixinStack;

#ifdef PROCCACHE
  Tcl_HashTable procCache;
#endif
  Tcl_HashTable metaData;
} XOTclObject;

typedef struct XOTclClass {
  struct XOTclObject object;
  struct XOTclClasses* super;
  struct XOTclClasses* sub;
  int color;
  struct XOTclClasses* order;
  struct XOTclClass* parent;
  Tcl_HashTable instances;
  Tcl_HashTable* objectdata;
  Namespace *nsPtr;
  Filter* filters;
  AssertionStore assertions;
  Tcl_Obj* parameterClass;
} XOTclClass;

typedef struct XOTclClasses {
  struct XOTclClass* cl;
  struct XOTclClasses* next;
} XOTclClasses;

/*
 * XOTcl CallStack
 */
typedef struct CallStackContent {
  XOTclObject* self;
  int destroyed;
  int isFilterEntry;
  int isMixinEntry;
  XOTclClass* class;
  char* procName;
  int frameCut;               /* set when frame is cutted, 
				 incremented for every 'cutting round' */
} CallStackContent;

typedef struct CallStack {
  CallStackContent content[MAX_NESTING_DEPTH];
  CallStackContent* top;
  int frameCutRound;
} CallStack;

/*
 * Tcl_Obj Types for XOTcl Objects
 */

static int SetXOTclObjectFromAny(Tcl_Interp* interp, Tcl_Obj* objPtr);
static void UpdateStringOfXOTclObject(Tcl_Obj* objPtr);

Tcl_ObjType XOTclObjectType = {
  "XOTclObject",			
  (Tcl_FreeInternalRepProc *) NULL,
  (Tcl_DupInternalRepProc *) NULL,
  UpdateStringOfXOTclObject,	
  SetXOTclObjectFromAny		
};

/*
 * definitions of the main otcl objects
 */

static Namespace* XOTclClassesNS;

static Tcl_HashTable* theObjects = 0;
static Tcl_HashTable* theClasses = 0;
static Tcl_CmdProc* ProcInterpId = 0;
static Tcl_ObjCmdProc* ProcObjInterpId = 0;

/* these are names and contents for global (corresponding) Tcl_Objs
   and Strings - otherwise these "constants" would have to be built
   every time they are used; now they are built once in XOTcl_Init */
typedef enum {
    EMPTY, UNKNOWN, CREATE, DESTROY, ALLOC, INIT, INCR, INSTVAR, AUTONAMES,
    ZERO, ONE, RENAME, MOVE, SELF, CLASS, CALLED_FROM_TCL,
    SELF_CLASS, SELF_PROC, OBJECT, PARAM_CL, DEFAULT, EXIT_HANDLER
} global_names;
static char *global_strings[] = {
  "",
  "unknown",
  "create",
  "destroy",
  "alloc",
  "init",
  "incr",
  "instvar",
  "__autonames",
  "0", "1",
  "rename", "move", "self", "class",
  "CALLED_FROM_TCL",
  "self class", "self proc",
  "Object", "Class::Parameter", "default",
  "::Object::__exitHandler"
};
static Tcl_Obj *global_objects[nr_elements(global_strings)];

static XOTCLINLINE int IsType(XOTclObject* obj, XOTclClass* type);
static XOTCLINLINE XOTclClass* InObject(Tcl_Interp* in);
static XOTCLINLINE XOTclClass* InClass(Tcl_Interp* in);
static XOTCLINLINE XOTclClass* XOTclIsClass(Tcl_Interp* in, ClientData cd);

static int XOTclObjDispatch(ClientData cd,
			    Tcl_Interp* in,
			    int objc,
			    Tcl_Obj *CONST objv[]);
static XOTCLINLINE int ObjDispatch(ClientData cd,
		       Tcl_Interp* in,
		       int objc,
		       Tcl_Obj *CONST objv[],
		       int noFilters);
#ifdef PROCCACHE
static void ProcCacheRemoveCache(Tcl_HashTable* procCacheTable);
#endif

#ifndef NODOUBLECHECK
extern XOTclObject*
XOTclAsObject(Tcl_Interp* in, ClientData cd) {
    return IsType((XOTclObject*)cd, InObject(in)) ? (XOTclObject*)cd : 0;
}
extern XOTCLINLINE XOTclClass*
XOTclAsClass(Tcl_Interp* in, ClientData cd) {
  return IsType((XOTclObject*)cd, InClass(in)) ? (XOTclClass*)cd : 0;
}
#else
#       define XOTclAsObject(in,cd) ((XOTclObject*)cd)
#       define XOTclAsClass(in,cd) ((XOTclClass*)cd)
#endif /* NODOUBLECHECK */

static XOTCLINLINE CallStackContent*
CallStackGetFrame(Tcl_Interp* in);

/*
 *  realize self, class, proc through the [self] command
 */

static XOTCLINLINE char*
GetSelfProc(Tcl_Interp *in) {
  return CallStackGetFrame(in)->procName;
}

static XOTCLINLINE XOTclClass*
GetSelfClass(Tcl_Interp *in) {
  return CallStackGetFrame(in)->class;
}

static XOTCLINLINE XOTclObject*
GetSelfXObj(Tcl_Interp* in) {
  return CallStackGetFrame(in)->self;
}

static XOTCLINLINE char*
className(XOTclClass *cl)
{
  return cl ? ObjStr(cl->object.cmdName) : "";
}

static XOTCLINLINE XOTclObject*
classXObj(XOTclClass *cl)
{
  return cl ? &cl->object : NULL;
}

static XOTCLINLINE char*
xobjName(XOTclObject *obj) {
  if (obj)
    return ObjStr(obj->cmdName);
  else
    return "";
}

/*
 * error return functions
 */

static int
XOTclErrMsg(Tcl_Interp *in, char* msg, Tcl_FreeProc* type) {
    Tcl_SetResult(in, msg, type);
    return TCL_ERROR;
}

static int
XOTclVarErrMsg TCL_VARARGS_DEF (Tcl_Interp *,arg1)
{
    va_list argList;
    char *string;
    Tcl_Interp *in;

    in = TCL_VARARGS_START(Tcl_Interp *,arg1,argList);
    Tcl_ResetResult(in);
    while (1) {
	  string = va_arg(argList, char *);
	   if (string == NULL) {
	      break;
	  }
	  Tcl_AppendResult(in, string, (char*) NULL);
    }
    va_end(argList);
    return TCL_ERROR;
}

static int
XOTclErrInProc (Tcl_Interp *in, XOTclObject* obj, XOTclClass* cl, char* procName) {
    Tcl_DString errMsg;
    char *cName, *space;
    ALLOC_DSTRING(&errMsg, "\n    ");
    if (cl) {
      cName = className(cl);
      space = " ";
    } else {
      cName = "";
      space ="";
    }
    Tcl_DStringAppend(&errMsg, ObjStr(obj->cmdName),-1);
    Tcl_DStringAppend(&errMsg, space, -1);
    Tcl_DStringAppend(&errMsg, cName, -1);
    Tcl_DStringAppend(&errMsg, "->", 2);
    Tcl_DStringAppend(&errMsg, procName, -1);
    Tcl_AddErrorInfo (in, Tcl_DStringValue(&errMsg));
    Tcl_DStringFree(&errMsg);
    return TCL_ERROR;
}

static int
XOTclObjErrArgCnt(Tcl_Interp *in, Tcl_Obj *cmdname, char *arglist) {
  Tcl_ResetResult(in);
  Tcl_AppendResult(in, "wrong # args: should be {", 0);
  if (cmdname) {
    Tcl_AppendResult(in, ObjStr(cmdname), " ", 0);
  }
  if (arglist != 0) Tcl_AppendResult(in, arglist, 0);
  Tcl_AppendResult(in, "}", 0);
  return TCL_ERROR;
}

static int
XOTclErrBadVal(Tcl_Interp *in, char *expected, char *value) {
  Tcl_ResetResult(in);
  Tcl_AppendResult(in, "expected ", expected, " but got", 0);
  Tcl_AppendElement(in, value);
  return TCL_ERROR;
}

static int
XOTclObjErrType(Tcl_Interp *in, Tcl_Obj *nm, char* wt) {
  Tcl_ResetResult(in);
  Tcl_AppendResult(in,"'",ObjStr(nm), "' method should be called on '",
		   wt, "'", 0);
  return TCL_ERROR;
}

/*
 *   Tracing facilities
 */
static void
XOTclStackTrace (Tcl_Interp* in) {
  Interp *iPtr = (Interp *) in;
  CallFrame *f = iPtr->framePtr, *v = iPtr->varFramePtr;
  Tcl_Obj* varCmdObj;

  XOTclNewObj(varCmdObj);
  fprintf (stderr, "     TCL STACK: ");
  if (f == 0) fprintf(stderr, "- ");
  while (f) {
    Tcl_Obj* cmdObj;
    XOTclNewObj(cmdObj);
    if (f && f->isProcCallFrame && f->procPtr && f->procPtr->cmdPtr) {
      Tcl_GetCommandFullName(in, (Tcl_Command)  f->procPtr->cmdPtr, cmdObj);
      if (cmdObj) {
	fprintf(stderr, " %s (%d)", ObjStr(cmdObj), f->level);
      }
      DECR_REF_COUNT(cmdObj);
    } else fprintf(stderr, "- ");

    f = f->callerPtr;
    if (f) fprintf(stderr, ",");
  }

  fprintf (stderr, " VARFRAME: ");
  if (v && v->isProcCallFrame && v->procPtr && v->procPtr->cmdPtr) {
    Tcl_GetCommandFullName(in, (Tcl_Command)  v->procPtr->cmdPtr, varCmdObj);
    if (varCmdObj) {
      fprintf(stderr, " %s (%d)\n", ObjStr(varCmdObj), v->level);
    }
  } else fprintf(stderr, "- \n");
  DECR_REF_COUNT(varCmdObj);
}

static void
XOTclCallStackTrace (Tcl_Interp* in) {
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent *csc;

  fprintf (stderr, "     XOTCL CALLSTACK: \n");
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    fprintf(stderr, "       ");
    if (csc->self)
      fprintf(stderr, "OBJ: %s, ", ObjStr(csc->self->cmdName));
    if (csc->class)
      fprintf(stderr, "CL: %s, ", className(csc->class));
    fprintf(stderr, "PROC: %s, ", csc->procName);

    if (csc->frameCut > 0) 
      fprintf(stderr, "--frame cutted (%d)", csc->frameCut);
    fprintf(stderr, "\n");
  }
}

static void
XOTclFilterTrace (Tcl_Interp* in) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = &XOTclFilterList ?
	  Tcl_FirstHashEntry(&XOTclFilterList, &hSrch) : 0;

  fprintf(stderr, "     Filters: \n");
  if (hPtr == 0)
    fprintf(stderr, "-");

  while (hPtr != 0) {
    Tcl_Obj* cmdObj;
    Command* cmdPtr = (Command*) Tcl_GetHashKey(&XOTclFilterList, hPtr);
    if (cmdPtr && !cmdPtr->cmdEpoch) {
      XOTclNewObj(cmdObj);
      Tcl_GetCommandFullName(in, 
			     (Tcl_Command) cmdPtr, 
			     cmdObj);
      if (cmdObj) {
	char* filterName = ObjStr(cmdObj);
	XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
	if (entry) {
	  ObjList* o = entry->activeObjs;
	  fprintf(stderr, "       %s <%d> active on: ", filterName, entry->count);
	  if (o == 0)
	    fprintf(stderr, "-");
	  else 
	    while (o) {
	      fprintf(stderr, "%s, ", ObjStr(o->objPtr->cmdName));
	      o = o->next;
	    }
	}
      }
      DECR_REF_COUNT(cmdObj);
      fprintf(stderr, "\n");
    }
    hPtr = Tcl_NextHashEntry(&hSrch);
  }
}
/* helper function to print the vars dynamically created on a 
   callframe 
static void printLocalTable (CallFrame* c) { 
  Tcl_HashEntry* entryPtr;
  Tcl_HashTable *localVarTablePtr = c->varTablePtr;
  Tcl_HashSearch search;
  
  fprintf(stderr, "LocalVars:");

  if (localVarTablePtr != NULL) {
    for (entryPtr = Tcl_FirstHashEntry(localVarTablePtr, &search);
	 entryPtr != NULL;
	 entryPtr = Tcl_NextHashEntry(&search)) {
      char* varName = Tcl_GetHashKey(localVarTablePtr, entryPtr);
      fprintf(stderr, " %s,", varName);
    }
  }
  fprintf(stderr,"\n");
}
*/

/*
 *  Tcl_Obj functions for objects
 */

static void RegisterObjTypes () {
    Tcl_RegisterObjType(&XOTclObjectType);
}

static int
SetXOTclObjectFromAny(Tcl_Interp *interp, register Tcl_Obj * objPtr) {
    Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    char *string;

    XOTclObject* obj;
    string = ObjStr(objPtr);

    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    obj = XOTclGetObject (interp, string);
    if (!obj) {
	if (interp != NULL)
	    return XOTclVarErrMsg (interp, "Expected an Object, but got ", 
				   string, (char*) NULL);
	return TCL_ERROR;
    }

    objPtr->internalRep.otherValuePtr = (XOTclObject*) obj;
    objPtr->typePtr = &XOTclObjectType;
    return TCL_OK;
}

static void
UpdateStringOfXOTclObject(objPtr)
    register Tcl_Obj *objPtr;
{

    XOTclObject* obj = (XOTclObject*) objPtr->internalRep.otherValuePtr;
    char* nsFullName = NULL;
    register Command *cmdPtr;

    /* Here we use GetCommandName, because it doesnt need
       Interp*, but Tcl_GetCommandFullName(in,obj->id,ObjName); does*/
    if (obj) {
	cmdPtr = (Command *) obj->id;
    } else {
	cmdPtr = NULL;
    }	
    if (cmdPtr != NULL) {
	Tcl_DString ds, *dsp = &ds;
	unsigned l;

	Tcl_DStringInit(dsp);
	nsFullName = cmdPtr->nsPtr->fullName;
	if (!(*nsFullName==':' && *(nsFullName+1)==':' &&
	    *(nsFullName+2)=='\0')) {
	     Tcl_DStringAppend(dsp, nsFullName, -1);
	}
	Tcl_DStringAppend(dsp, "::", 2);
	Tcl_DStringAppend(dsp, Tcl_GetCommandName(NULL, obj->id), -1);

	l = (unsigned) Tcl_DStringLength(dsp)+1;
	objPtr->bytes = (char*) ckalloc(l);
	memcpy(objPtr->bytes, Tcl_DStringValue(dsp), l);
	objPtr->length = Tcl_DStringLength(dsp);
	Tcl_DStringFree(dsp);
    } else {
	objPtr->bytes = NULL;
	objPtr->length = 0;
    }
}

static Tcl_Obj *
NewXOTclObjectObj(obj)
    register XOTclObject* obj;
{
    register Tcl_Obj *objPtr;

    XOTclNewObj(objPtr);
    objPtr->bytes = NULL;

    objPtr->internalRep.otherValuePtr = obj;
    objPtr->typePtr = &XOTclObjectType;
    return objPtr;
}

/* 
static void
SetXOTclObjectObj(objPtr, obj)
    register Tcl_Obj *objPtr;
    register XOTclObject* obj;
{
    register Tcl_ObjType *oldTypePtr = objPtr->typePtr;

    if (Tcl_IsShared(objPtr)) {
  	panic("Tcl_SetXOTclObjectObj called with shared object");
    }

    Tcl_InvalidateStringRep(objPtr);
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = obj;
    objPtr->typePtr = &XOTclObjectType;
}

static int
GetXOTclObjectfromObj(interp, objPtr, obj)
    register Tcl_Interp *interp; 	
    register Tcl_Obj *objPtr;	
    register XOTclObject **obj;
{
    register int result;

    result = SetXOTclObjectFromAny(interp, objPtr);
    if (result == TCL_OK) {
	*obj = (XOTclObject*) objPtr->internalRep.otherValuePtr;
    }
    return result;
}
*/

/*
 * precedence ordering functions
 */

enum colors { WHITE, GRAY, BLACK };

static int
TopoSort(XOTclClass* cl, XOTclClass* base, XOTclClasses* (*next)(XOTclClass*)) {
  XOTclClasses* sl = (*next)(cl);
  XOTclClasses* pl;

  /*
   * careful to reset the color of unreported classes to
   * white in case we unwind with error, and on final exit
   * reset color of reported classes to white
   */

  cl->color = GRAY;
  for (; sl != 0; sl = sl->next) {
    XOTclClass* sc = sl->cl;
    if (sc->color==GRAY) { cl->color = WHITE; return 0; }
    if (sc->color==WHITE && !TopoSort(sc, base, next)) {
      cl->color=WHITE;
      if (cl == base) {
        XOTclClasses* pc = cl->order;
        while (pc != 0) { pc->cl->color = WHITE; pc = pc->next; }
      }
      return 0;
    }
  }
  cl->color = BLACK;
  pl = (XOTclClasses*)ckalloc(sizeof(XOTclClasses));
  pl->cl = cl;
  pl->next = base->order;
  base->order = pl;
  if (cl == base) {
    XOTclClasses* pc = cl->order;
    while (pc != 0) { pc->cl->color = WHITE; pc = pc->next; }
  }
  return 1;
}

static void
RC(XOTclClasses* sl) {
  while (sl != 0) {
    XOTclClasses* n = sl->next;
    ckfree((char*)sl); sl = n;
  }
}

static XOTCLINLINE void
AC(XOTclClasses** cList, XOTclClass* cl) {
  XOTclClasses *l = *cList, 
    *new = (XOTclClasses*) ckalloc ((unsigned) sizeof (XOTclClasses));
  new->cl = cl;
  new->next = NULL;
  if (l) {
    while (l->next) l = l->next; 
    l->next = new;
  } else
    *cList = new;
}

static XOTclClasses* Super(XOTclClass* cl) { return cl->super; }
static XOTclClasses* Sub(XOTclClass* cl) { return cl->sub; }

static XOTclClasses*
TopoOrder(XOTclClass* cl, XOTclClasses* (*next)(XOTclClass*)) {
     if (TopoSort(cl, cl, next))
	  return cl->order;
     RC(cl->order);
     cl->order = 0;
     return 0;
}

static XOTCLINLINE XOTclClasses*
ComputePrecedence(register XOTclClass* cl) {
     if (!cl->order)
	  cl->order = TopoOrder(cl, Super);
     return cl->order;
}

static XOTCLINLINE XOTclClasses*
ComputeDependents(register XOTclClass* cl) {
     if (!cl->order)
	  cl->order = TopoOrder(cl, Sub);
     return cl->order;
}

static void
FlushPrecedences(XOTclClass* cl) {
  XOTclClasses* pc;
  RC(cl->order); cl->order = 0;
  pc = ComputeDependents(cl);

  /*
   * ordering doesn't matter here - we're just using toposort
   * to find all lower classes so we can flush their caches
   */

  if (pc) pc = pc->next;
  while (pc != 0) {
    RC(pc->cl->order); pc->cl->order = 0;
    pc = pc->next;
  }
  RC(cl->order); cl->order = 0;
}

static void
AddInstance(XOTclObject* obj, XOTclClass* cl) {
  obj->filterinfo.valid = 0;
  obj->cl = cl;
  if (cl != 0) {
    int nw;
    (void) Tcl_CreateHashEntry(&cl->instances, (char*)obj, &nw);
  }
}

static int
RemoveInstance(XOTclObject* obj, XOTclClass* cl) {
  obj->filterinfo.valid = 0;
  if (cl) {
    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&cl->instances, (char*)obj);
    if (hPtr) {
#ifdef PROCCACHE
	if (obj) ProcCacheRemoveCache(&obj->procCache);
#endif
	Tcl_DeleteHashEntry(hPtr);
	return 1;
    }
  }
  return 0;
}

/*
 * superclass/subclass list maintenance
 */

static void
AS(XOTclClass* cl, XOTclClass* s, XOTclClasses** sl) {
  XOTclClasses* l = *sl;
  while (l &&  l->cl != s) l = l->next;
  if (!l) {
    XOTclClasses* sc = (XOTclClasses*)ckalloc(sizeof(XOTclClasses));
    sc->cl = s; sc->next = *sl; *sl = sc;
  }
}

static void
AddSuper(XOTclClass* cl, XOTclClass* super) {
  if (cl && super) {
    /*
     * keep corresponding sub in step with super
     */
    AS(cl, super, &cl->super);
    AS(super, cl, &super->sub);
  }
}

static int
RS(XOTclClass* cl, XOTclClass* s, XOTclClasses** sl) {
  XOTclClasses* l = *sl;
  if (!l) return 0;
  if (l->cl == s) {
    *sl = l->next;
    ckfree((char*)l);
    return 1;
  }
  while (l->next && l->next->cl != s) l = l->next;
  if (l->next) {
    XOTclClasses* n = l->next->next;
    ckfree((char*)(l->next));
    l->next = n;
    return 1;
  }
  return 0;
}

static int
RemoveSuper(XOTclClass* cl, XOTclClass* super) {
  /*
   * keep corresponding sub in step with super
   */
  int sp = RS(cl, super, &cl->super);
  int sb = RS(super, cl, &super->sub);

  return (sp && sb);
}

/*
 * internal type checking
 */

static XOTCLINLINE XOTclClass*
InObject(Tcl_Interp* in) {
  Tcl_HashEntry* hp = Tcl_FindHashEntry(theObjects, (char*)in);
  return hp ? (XOTclClass*)Tcl_GetHashValue(hp) : 0;
}

static XOTCLINLINE XOTclClass*
InClass(Tcl_Interp* in) {
  Tcl_HashEntry* hp = Tcl_FindHashEntry(theClasses, (char*)in);
  return hp ? (XOTclClass*)Tcl_GetHashValue(hp) : 0;
}

static XOTCLINLINE int
IsType(XOTclObject* obj, XOTclClass* type) {
  register XOTclClass* t = obj ? obj->type : 0;
  while (t && t!=type) t = t->parent;
  return (t != 0);
}

/*
 * methods lookup
 */

static int XOTclDispatch(ClientData c, Tcl_Interp* i, int a, char* v[]);

static XOTCLINLINE Command*
FindMethod (char* methodName, Namespace* nsPtr) {
  if (nsPtr) {
    register Tcl_HashEntry* entryPtr =
      Tcl_FindHashEntry(&nsPtr->cmdTable, methodName);
    if (entryPtr) {
      Command *cp = (Command *) Tcl_GetHashValue(entryPtr);
      if (cp->proc != XOTclDispatch) 
	return cp;
    }
  }
  return NULL;
}

#ifdef PROCCACHE
static void
ProcCacheORemoveMethod (XOTclObject* obj, char* name) {
    Tcl_HashEntry* hEntryPtr = Tcl_FindHashEntry(&obj->procCache, name);
    if (hEntryPtr) {
	ProcCacheEntry* entry = (ProcCacheEntry *) Tcl_GetHashValue(hEntryPtr);
	entry->cmdPtr->refCount--;
	Tcl_DeleteHashEntry(hEntryPtr);
	ckfree((char*)entry);
    }
}

static void
ProcCacheCRemoveMethod (XOTclClass* cl, char* name) {
  XOTclClasses *saved = cl->order, *clPtr;

  cl->order = 0;
  clPtr = ComputeDependents(cl);

  while (clPtr != 0) {
      Tcl_HashSearch hSrch;
      Tcl_HashEntry* hPtr = &clPtr->cl->instances ?
	  Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;
      for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
	  XOTclObject* obj = (XOTclObject*)
	    Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
	  ProcCacheORemoveMethod (obj, name);
      }
      clPtr = clPtr->next;
  }

  RC(cl->order); cl->order = saved;
}

static void
ProcCacheRemoveCache(Tcl_HashTable* procCacheTable) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = Tcl_FirstHashEntry(procCacheTable, &hSrch);
  while (hPtr) {
      ProcCacheEntry* entry = (ProcCacheEntry *) Tcl_GetHashValue(hPtr);
      entry->cmdPtr->refCount--;
      Tcl_DeleteHashEntry(hPtr);
      ckfree((char*)entry);
      hPtr = Tcl_FirstHashEntry(procCacheTable, &hSrch);
  }
}

static void
ProcCacheRemoveAllDepending (XOTclClass* cl) {
  /* Remove procs of this class from all dependend classes */
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = Tcl_FirstHashEntry(&cl->nsPtr->cmdTable, &hSrch);
  while (hPtr) {
      ProcCacheCRemoveMethod (cl, Tcl_GetHashKey(&cl->nsPtr->cmdTable, hPtr));
      hPtr = Tcl_NextHashEntry(&hSrch);
  }
}
#endif

static XOTCLINLINE XOTclClass*
SearchCMethod(XOTclObject* obj,
	      XOTclClass* cl,
	      XOTclClasses* pl,
	      char* nm,
	      Command **procInfo) {
    /* Search the local proc cache */
#ifdef PROCCACHE
  if (obj) {
    register Tcl_HashEntry* entryPtr;
    entryPtr = Tcl_FindHashEntry(&obj->procCache,nm);
    if (entryPtr) {
      ProcCacheEntry* entry = (ProcCacheEntry *) Tcl_GetHashValue(entryPtr);
      if (entry->cmdPtr->cmdEpoch) {
	ProcCacheORemoveMethod (obj, nm);
      } else {
	*procInfo = entry->cmdPtr;
	return entry->cl;
      }
    }
  }
#endif

  if (cl){
    if (!cl->order)
      cl->order = ComputePrecedence(cl);
    pl = cl->order;
  }

  /* Search the class hierarchy */
  while (pl) {
    Command *pi = FindMethod (nm, pl->cl->nsPtr);
    if (pi) {
#ifdef PROCCACHE
      if (obj) {
	Tcl_HashEntry* hPtr = NULL;	
	int nw;
	hPtr = Tcl_CreateHashEntry(&obj->procCache, nm, &nw);
	if (nw) {
	  ProcCacheEntry* entry = 
	    (ProcCacheEntry*) ckalloc ((unsigned) sizeof (ProcCacheEntry));
	  
	  entry->cmdPtr = pi;
	  entry->cmdPtr->refCount++;
	  entry->cl = pl->cl;
	  Tcl_SetHashValue(hPtr, (ClientData) entry);
	}
      }
#endif
      *procInfo = pi;
      return pl->cl;
    }
    pl = pl->next;
  }
  *procInfo = NULL;
  return NULL;
}

/*
 * Namespace related commands
 */

static int
NSDeleteCmd (Tcl_Interp* in, Namespace* ns, char* name) {
  /* a simple deletion would delete a global command with
     the same name, if the command is not existing, so
     we use the CmdToken */
  Command* token = FindMethod (name, ns);
  if (token)
    return Tcl_DeleteCommandFromToken(in, (Tcl_Command) token);
  return 0;
}

static Tcl_Command
NSCreateObjCmd (Tcl_Interp* in, char *cmdName, Tcl_ObjCmdProc *proc,
                ClientData cD, Tcl_CmdDeleteProc *dProc, Namespace* ns) {
    Tcl_DString ds;
    Tcl_Command c;

    ALLOC_NAME_NS (&ds, ns->fullName,cmdName);

    c = Tcl_CreateObjCommand(in, Tcl_DStringValue(&ds), *proc,cD,dProc);
    Tcl_DStringFree(&ds);
    return c;
}

static void
NSDeleteChildren (Tcl_Interp* in, Namespace* nsPtr) {
  if (&nsPtr->childTable){
    Tcl_HashEntry* hPtr;
    Tcl_HashSearch hSrch;
    hPtr = Tcl_FirstHashEntry(&nsPtr->childTable, &hSrch);

    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      char *oname = Tcl_GetHashKey(&nsPtr->childTable,hPtr);
      Tcl_DString name;
      XOTclObject* obj;
      Tcl_Command oid;

      ALLOC_NAME_NS(&name,nsPtr->fullName,oname);

      obj = XOTclGetObject (in,  Tcl_DStringValue(&name));
      if (obj) {
	if ((obj->teardown != 0) && obj->id) {
	  if (!obj->destroyCalled) {
	    Tcl_Obj* args[2];
	    char* cmdName = Tcl_DStringValue(&name);
	    args[0] = obj->cmdName;
	    args[1] = global_objects[DESTROY];

	    if (cmdName && Tcl_FindCommand(in, cmdName, NULL, 0)) {
	      /*
	      fprintf(stderr,"NSDeleteChildren %p <%s> destroyCalled=%d\n", obj, oname,
		      obj->destroyCalled);
	      */
	      (void) XOTclObjDispatch((ClientData)obj, in, 2, args);
	      /*	      
	      fprintf(stderr,"NSDeleteChildren %p <%s> destroyCalled=%d after dispatch\n",
		      obj, oname, obj->destroyCalled);
	      */
	      obj->destroyCalled = 1;
	    }

	    /* perhaps that was a user defined destroy and the command
	       is not really destroyed -> make sure by checking the Cmd */
	    if (cmdName && Tcl_FindCommand(in, cmdName, NULL, 0) && obj) {
	      oid = obj->id;
	      obj->id = 0;
	      if ((obj->teardown != 0) && oid) {
		/*	      		
		fprintf(stderr,"NSDeleteChildren %p <%s> destroyCalled=%d calling DCFT\n",
			obj, oname, obj->destroyCalled);
		*/
		(void*) Tcl_DeleteCommandFromToken(in,oid);

		/*	      
		fprintf(stderr,"NSDeleteChildren %p <%s> destroyCalled=%d calling DCFT done\n",
			obj, oname, obj->destroyCalled);
		*/		
	      }
	    }
	  }
	}
      }
    }
  }
}

/* delete a namespace recursively, but check whether the 
   namespace is an object or not */
static void
NSDeleteNamespace (Tcl_Interp* in, Namespace* nsPtr) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;

  NSDeleteChildren(in, nsPtr);

  hPtr = &nsPtr->childTable ?
    Tcl_FirstHashEntry(&nsPtr->childTable, &hSrch) : 0;

  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    Namespace *child = ((Namespace*)Tcl_GetHashValue(hPtr));
    NSDeleteNamespace (in, child);
  } 
  /*	      
  fprintf(stderr, "NSDeleteNamespace deleting %s\n", nsPtr->fullName);
  */
  Tcl_DeleteNamespace((Tcl_Namespace*) nsPtr);
}

/* cuts preceding ::XOTclClasses away */
static char*
NSCutXOTclClasses (char* name) {
  while (*name == ':') name++;
  if (strncmp (name, "XOTclClasses", 12) == 0) name += 12;
  return name;
}

/* initializes and builds a DString with preceding ::XOTclClasses */
static void
NSAddXOTclClasses(char* name, Tcl_DString* nameDS) {
  Tcl_DStringInit(nameDS);
  Tcl_DStringAppend(nameDS, "::XOTclClasses::", -1);
  Tcl_DStringAppend(nameDS, NSCutXOTclClasses(name), -1);
}

static void
NSBuildFullName (Tcl_Interp* in, Namespace* nsPtr){
  Interp *iPtr = (Interp *) in;
  Namespace *ancestorPtr;
  Namespace *globalNsPtr = iPtr->globalNsPtr;
  Tcl_DString buffer1, buffer2;
  char* name;
  int l;
  /*
   * Build the fully qualified name for this namespace.
   * this is from Tcl_CreateNamespace in the Tcl - Source
   */
  Tcl_DStringInit(&buffer1);
  Tcl_DStringInit(&buffer2);
  for (ancestorPtr = nsPtr;  ancestorPtr != NULL;
       ancestorPtr = ancestorPtr->parentPtr) {
    if (ancestorPtr != globalNsPtr) {
      Tcl_DStringAppend(&buffer1, "::", 2);
      Tcl_DStringAppend(&buffer1, ancestorPtr->name, -1);
    }
    Tcl_DStringAppend(&buffer1,
		      Tcl_DStringValue(&buffer2),
		      Tcl_DStringLength(&buffer2));

    Tcl_DStringSetLength(&buffer2, 0);
    Tcl_DStringAppend(&buffer2,
		      Tcl_DStringValue(&buffer1),
		      Tcl_DStringLength(&buffer1));
    Tcl_DStringSetLength(&buffer1, 0);
  }

  l = Tcl_DStringLength(&buffer2)+1;
  name = Tcl_DStringValue(&buffer2);
  if (nsPtr->fullName) ckfree((char*)nsPtr->fullName);
  nsPtr->fullName = (char *) ckalloc((unsigned)l);
  memcpy(nsPtr->fullName, name, l);
  Tcl_DStringFree(&buffer1);
  Tcl_DStringFree(&buffer2);
}

static void
NSTreeBuildFullName (Tcl_Interp* in, Namespace* nsPtr) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = &nsPtr->childTable ?
    Tcl_FirstHashEntry(&nsPtr->childTable, &hSrch) : 0;
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    Namespace *child = ((Namespace*)Tcl_GetHashValue(hPtr));
    NSTreeBuildFullName(in,child);
  }
  NSBuildFullName(in,nsPtr);
}
static int
XOTcl_RenameObjCmd(ClientData d, Tcl_Interp *i, int c, Tcl_Obj *CONST v[]);

static void
NSCopy(Tcl_Interp* in, Namespace* nsPtr, char* newName, ClientData cd) {
  Var *varPtr=0;
  Tcl_Command cmd;
  char* varName, *cmdName;
  Tcl_Obj *cmdObj;
  Tcl_DString ds;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;
  Tcl_Obj* ov[3];
  /* rename all procs & commands in the ns */
  hPtr = &nsPtr->cmdTable ?
    Tcl_FirstHashEntry(&nsPtr->cmdTable, &hSrch) : 0;
  ov[0] = global_objects[RENAME];
  while (hPtr != NULL) {
    cmdName = Tcl_GetHashKey(&nsPtr->cmdTable, hPtr);
    cmd = (Tcl_Command) Tcl_GetHashValue(hPtr);
    XOTclNewObj(cmdObj);
    Tcl_GetCommandFullName(in, cmd, cmdObj);
    ov[1] = cmdObj;

    ov[2] = Tcl_NewStringObj(newName,-1);
    INCR_REF_COUNT(ov[2]);
    Tcl_AppendStringsToObj(ov[2] , "::", cmdName, (char*)NULL);
    /*
      fprintf(stderr, "+++Command Rename: %s to %s\n", ObjStr(cmdObj), ObjStr(ov[2]));
    */
    /* no direct call to TclRenameCommand in order to rename XOTcl
       Objects using the overloaded ::rename and to filter the
       object renames */
    if (XOTcl_RenameObjCmd((ClientData) 0, in, 3, ov) != TCL_OK) {
      /* if a rename fails -> ignore it, because these are already defined
	 commands, e.g. self & next if we copy to a newly created object */
      Tcl_ResetResult(in);
    }
    DECR_REF_COUNT (ov[2]);
    DECR_REF_COUNT(cmdObj);
    hPtr = Tcl_NextHashEntry(&hSrch);
  }

  /* rename all vars in the ns */
  hPtr = &nsPtr->varTable ?
    Tcl_FirstHashEntry(&nsPtr->varTable, &hSrch) : 0;
  while (hPtr != NULL) {
    varPtr = (Var *) Tcl_GetHashValue(hPtr);
    if (!TclIsVarUndefined(varPtr) && !TclIsVarLink(varPtr)) {
      varName = Tcl_GetHashKey(&nsPtr->varTable, hPtr);
      ALLOC_NAME_NS(&ds, newName, varName);

      if (TclIsVarScalar(varPtr)) {
	Tcl_SetVar2(in, Tcl_DStringValue(&ds), 0,
		    ObjStr(varPtr->value.objPtr), TCL_NAMESPACE_ONLY);	
      } else {
	if (TclIsVarArray(varPtr)) {
	  Tcl_HashTable* aTable = varPtr->value.tablePtr;
	  Tcl_HashSearch ahSrch;
	  Tcl_HashEntry* ahPtr = aTable ? Tcl_FirstHashEntry(aTable, &ahSrch) : 0;
	
	  for (; ahPtr != 0; ahPtr = Tcl_NextHashEntry(&ahSrch)) {
	    char *eltName = Tcl_GetHashKey(aTable, ahPtr);
	    Var  *eltVar  = (Var*) Tcl_GetHashValue(ahPtr);
	
	    if (TclIsVarScalar(eltVar)) {
	      Tcl_SetVar2(in, Tcl_DStringValue(&ds), eltName,
			  ObjStr(eltVar->value.objPtr), TCL_NAMESPACE_ONLY);
	    }
	  }
	}
      }
      Tcl_DStringFree(&ds);
    }
    hPtr = Tcl_NextHashEntry(&hSrch);
  }
}

static Namespace*
NSCreateAndCopy(Tcl_Interp* in, Namespace* nsPtr, char* newName, ClientData cd) {
  Namespace* newNsPtr = (Namespace*)
    Tcl_CreateNamespace (in, newName,
			 cd ,(Tcl_NamespaceDeleteProc*)NULL);
  /*
  fprintf(stderr, "+++Copy -- Creating NS: %s\n", newName);
  */
  NSCopy(in, nsPtr, newName, cd); 
  return newNsPtr;
}

static int
NSRenameCommand(Tcl_Interp* in, char* newName, char* self) {
  int selfLen = strlen(self);
  /*
   * rename does not succeed if the rename-destination overwrites an existiting 
   * object or if it nests inside of the (to be deleted) origin
   */
  if (strncmp(newName, self, selfLen) == 0) {
    if (((int) (strlen(newName)) > selfLen + 2) &&
	*(newName+selfLen) == ':' &&
	*(newName+selfLen+1) == ':') {
      return (XOTclVarErrMsg(in, "Can't rename: '",self, "' to '",
			     newName, "'; Destination nested inside of origin",
			     (char*) NULL));
    }
  }
  if (Tcl_FindCommand(in, newName, NULL, 0))
      return (XOTclVarErrMsg(in, "Can't rename: '",self, "' to '",
				 newName, "'; Object allready exists",
				 (char*) NULL));

  return TclRenameCommand (in, self, newName);
}

static void
NSCorrectCmdName (Tcl_Interp* in, XOTclObject* obj) {
  /* correct Info in cmdName */
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = &obj->nsPtr->childTable ?
    Tcl_FirstHashEntry(&obj->nsPtr->childTable, &hSrch) : 0;

  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    Namespace *child = ((Namespace*)Tcl_GetHashValue(hPtr));
    NSCorrectCmdName(in, (XOTclObject *)child->clientData);
  }
  DECR_REF_COUNT(obj->cmdName);
  obj->cmdName = NewXOTclObjectObj(obj);
  INCR_REF_COUNT(obj->cmdName);
}

/*
 *   XOTcl CallStack
 */

static XOTCLINLINE int
CallStackPush(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl, char* procName,
	      int isFilterEntry, int isMixinEntry) {
  CallStack *cs;
  CallStackContent *csc;

  cs = (CallStack *)XOTclClassesNS->clientData;
  if (cs->top >= &cs->content[MAX_NESTING_DEPTH-1]) {
    Tcl_ResetResult(in);
    Tcl_AppendToObj(Tcl_GetObjResult(in),
		    "too many nested calls to Tcl_EvalObj (infinite loop?)",
		    -1);
    return TCL_ERROR;
  }
  csc = ++cs->top;
  csc->self          = obj;
  csc->class         = cl;
  csc->procName      = procName;
  csc->destroyed     = 0;
  csc->isFilterEntry = isFilterEntry;
  csc->isMixinEntry  = isMixinEntry;
  csc->frameCut      = 0;
  return TCL_OK;
}

static XOTCLINLINE void
CallStackDestroyObject(Tcl_Interp* in, XOTclObject* obj) {
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent *csc;
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (csc->self == obj) {
      csc->destroyed = 1;
    }
  }
}

static XOTCLINLINE int
CallStackIsDestroyed() {
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  return cs->top->destroyed;
}

static XOTCLINLINE void
CallStackPop() {
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent *csc;
  assert(cs->top > cs->content);
  csc = cs->top;
  cs->top--;
}

static XOTCLINLINE CallStackContent*
CallStackGetFrame(Tcl_Interp* in) {
  Interp *iPtr= (Interp *) in;
  CallFrame *v = iPtr->varFramePtr, *f = iPtr->framePtr;
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent
    *bot = cs->content,
    *top = cs->top;
  int i;

  /* the xotcl call stack contains entries for filters, these are cut out
     of the tcl stack -> determine the entry in the xotcl stack that
     corresponds to the varFramePtr of the tcl stack */
  if (f && v) {
    int diff = f->level - v->level;
    CallFrame* d = iPtr->framePtr;

    for (i = diff; i > 0; i--, d = d->callerPtr) {
      if (d && d->nsPtr && !XOTclGetObject(in, d->nsPtr->fullName)
	  && !XOTclGetClass(in, NSCutXOTclClasses(d->nsPtr->fullName))) {
	diff--;
      }
    }

    /* step through the diff, overstep filters */
    while (diff > 0 && top > bot) {
      if (!top->isFilterEntry)
        diff--;
      do
	top--;
      while (top->isFilterEntry);
    }
  }

  if (top->destroyed)
    top = bot;

  return top;
}

static XOTCLINLINE int FilterActiveOnObj (XOTclObject* o, Command* c);

static XOTCLINLINE CallStackContent*
CallStackGetFrameWithoutInactiveFilters(Tcl_Interp* in) {
  CallStackContent *top;
  top = CallStackGetFrame(in); 
  while (top->isFilterEntry &&
	 top->self->filterinfo.chains &&
	 (!(top->self->filterinfo.chains->filterPtr) || 
	 (top->self->filterinfo.chains->filterPtr &&
	 !FilterActiveOnObj(top->self,
		       top->self->filterinfo.chains->filterPtr->cmdPtr)))) {
    top--;
  }
  return top;
}

static XOTCLINLINE CallStackContent*
CallStackGetFrameWithoutFilters(Tcl_Interp* in) {
  CallStackContent *top;
  for (top = CallStackGetFrame(in); top->isFilterEntry; top--);
  return top;
}

static XOTCLINLINE int
AssertionCheckPre(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl, char* m);
static XOTCLINLINE int
AssertionCheckPost(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl, char* m);
static XOTCLINLINE int
AssertionCheckInvars(Tcl_Interp* in, XOTclObject* obj);

static XOTCLINLINE int
callProcCheck(ClientData cp, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[],
	      Command *procInfo,
	      XOTclObject* obj, XOTclClass *cl, char *methodName,
	      int isFilterEntry , int isMixinEntry, int* isDestroyed)
{
  int result;
  if (isDestroyed != 0)
    *isDestroyed = 0;
  if (!TclIsProc(procInfo)) {
    /*if (obj && obj->teardown &&
	AssertionCheckInvars(in,obj) == TCL_ERROR)
	return TCL_ERROR;*/
    result = (*procInfo->objProc)(cp, in, objc, objv);
    /*if (obj && obj->teardown &&
	AssertionCheckInvars(in,obj) == TCL_ERROR)
	return TCL_ERROR;*/
  } else 
    if (CallStackPush(in, obj, cl, methodName,
		      isFilterEntry, isMixinEntry) == TCL_OK) {  
       if (obj && obj->teardown &&
	  AssertionCheckPre(in, obj, cl, methodName) == TCL_ERROR)
	  return TCL_ERROR;
      result = (*procInfo->objProc)(cp, in, objc, objv);
      if (obj && obj->teardown &&
	  AssertionCheckPost(in, obj, cl, methodName) == TCL_ERROR)
	return TCL_ERROR;
      if (isDestroyed != 0)
	*isDestroyed = CallStackIsDestroyed();
      CallStackPop();
    }
    else
      result = TCL_ERROR;

  return result;
}

static XOTCLINLINE int
callProc(ClientData cp, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[],
	 Command *procInfo,
	 XOTclObject* obj, XOTclClass *cl, char *methodName,
	 int isFilterEntry , int isMixinEntry)
{
  int result;
  if (CallStackPush(in, obj, cl, methodName,
		    isFilterEntry, isMixinEntry) == TCL_OK) {
    /*if (obj && obj->teardown &&
	AssertionCheckInvars(in,obj) == TCL_ERROR)
	return TCL_ERROR;*/
    result = (*procInfo->objProc)(cp, in, objc, objv);
    /*if (obj && obj->teardown &&
	AssertionCheckInvars(in,obj) == TCL_ERROR)
	return TCL_ERROR;*/
    CallStackPop();
  }
  else
    result = TCL_ERROR;

  return result;
}

static int
XOTclGetSelfObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int rc = TCL_OK;
  XOTclObject* obj;

  if (objc == 1) {
    if ((obj = GetSelfXObj(in))) {
      /*
	fprintf(stderr,"obj=%p, cd=%p <%s> <%s>\n", obj,cd,
	  ObjStr((obj->cmdName)),
	  ObjStr(((XOTclObject*)cd)->cmdName));
      */
      Tcl_SetObjResult(in,obj->cmdName);
    }
    else
      rc = XOTclVarErrMsg (in,"Cant find self", (char*) NULL);
  } else if (objc == 2) {
    char *option = ObjStr(objv[1]);
    if (isProcString(option)) { /* proc subcommand */
      char* procName;
      if ((procName = GetSelfProc(in)))
	Tcl_SetResult(in, procName, TCL_VOLATILE);
      else
	rc = XOTclVarErrMsg (in,"Cant find proc", (char*) NULL);
    } else if (isClassString(option)) { /* class subcommand */
      if ((obj = classXObj(GetSelfClass(in))))
	Tcl_SetObjResult(in, obj->cmdName);
      else
	Tcl_SetObjResult(in, global_objects[EMPTY]);
    } else
      rc = XOTclVarErrMsg(in,"unknown option: self <class|proc>",
			  (char*)NULL);
  } else /* objc != 2 */ {
    rc = XOTclVarErrMsg (in,"wrong # of args: self <class|proc>",
			 (char*)NULL);
  }

  return rc;
}

/*
 * Default values specified through 'parameters'
 */

static XOTCLINLINE int
SearchDefaultValues(Tcl_Interp* in,
		    XOTclObject* obj, XOTclClass* cmdCl) {
  Tcl_CallFrame frame;	
  XOTclClass*   cl = obj->cl;
  XOTclClasses* pl = 0;

  if (cl){
    if (!cl->order)
      cl->order = ComputePrecedence(cl);
    pl = cl->order;
  }

  Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);

  while (pl) {

    if (pl->cl->object.nsPtr) {
      register Tcl_HashEntry* entryPtr =
	Tcl_FindHashEntry(&pl->cl->object.nsPtr->varTable, "__defaults");
      Var* defaults;
      if (entryPtr) {
	defaults = (Var*) Tcl_GetHashValue(entryPtr);
	/*fprintf(stderr,"+++ we have defaults for <%s>\n", className(pl->cl));*/
	if (TclIsVarArray(defaults)) {
	  Tcl_HashTable* table = defaults->value.tablePtr;
	  Tcl_HashSearch hSrch;
	  Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
	
	  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
	    char *varName = Tcl_GetHashKey(table, hPtr);
	    Var  *val     = (Var*) Tcl_GetHashValue(hPtr);
	    if (TclIsVarScalar(val)) {
	      char *oldValue = Tcl_GetVar2(in, varName, NULL, TCL_NAMESPACE_ONLY);
	      /*fprintf(stderr,"+++ oldValue <%s>\n",oldValue);*/
	      if (oldValue == NULL) {
		char* value = ObjStr(val->value.objPtr), *v = value;
		int isCmd = 0;
		while (*v) {
		  if (isCmd == 0 && *v == '[')
		    isCmd = 1;
		  if (isCmd == 1 && *v == ']') {
		    isCmd = 2;
		    break;
		  }
		  v++;
		}
		if (isCmd != 2) {
		  /*fprintf(stderr,"+++ setting %s to <%s>\n", varName, value);*/
		  Tcl_SetVar2(in, varName, 0, value, TCL_NAMESPACE_ONLY);
		} else {
		  int r,s;

		  r = CallStackPush(in, obj, cmdCl, "init", 0, 0);
		  /*fprintf(stderr,"+++ executing set %s %s\n", varName, value);*/
		  s = Tcl_VarEval(in, "set ", varName, " ", value, (char*) 0);
		  CallStackPop();

		  if (s != TCL_OK) {
		    Tcl_ResetResult(in);
		    Tcl_AppendResult(in, "Default value '", value, "' for '",
				     varName, "' failed", (char*) NULL);
		    r = s;
		  }	
		  if (r != TCL_OK) {
		    Tcl_PopCallFrame(in);
		    return s;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    pl = pl->next;
  }
  Tcl_PopCallFrame(in);
  return TCL_OK;
}

static int
ParameterDefaultMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  Tcl_Obj *classVar,*val;
  char* setResult;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 3) 
    return XOTclObjErrArgCnt(in, 
				  cl->object.cmdName,
				  "default param values");

  classVar = Tcl_NewStringObj(className(cl),-1);
  INCR_REF_COUNT(classVar);
  Tcl_AppendStringsToObj(classVar, "::__defaults", (char*) 0);
  val = Tcl_ConcatObj(objc-2, objv+2);
  INCR_REF_COUNT(val);
  setResult = Tcl_SetVar2(in, ObjStr(classVar), ObjStr(objv[1]), ObjStr(val), 0);

  DECR_REF_COUNT(val);
  DECR_REF_COUNT(classVar);
  if (setResult == NULL)
    return XOTclVarErrMsg(in, "Could not set default value: '",
			       ObjStr(objv[2]), "'", (char*) 0);
  return TCL_OK;
}

static XOTCLINLINE int isDigitParam(char* c) {
  if (!c) 
    return 0;
  for (; *c; c++) 
    if (*c != '.' && (*c < '0' || *c > '9')) 
      return 0;
  return 1;
}

/*
 * Conversion from CmdPtr to Class
 */

extern XOTclClass*
GetClassFromCmdPtr(Tcl_Interp* in, Command* cmdPtr) {
  XOTclClass* cl;

  if (cmdPtr && !cmdPtr->cmdEpoch &&
      cmdPtr->proc == XOTclDispatch)
      cl = XOTclAsClass(in, cmdPtr->clientData);
  else
      cl = 0;
  return cl;
}

/*
 * Autonaming
 */

static Tcl_Obj*
AutonameIncr (Tcl_Interp* in, Tcl_Obj* Name, XOTclObject* obj) {
  Tcl_Obj* valueObject, *result;
  char* name = ObjStr(Name);

  Tcl_Obj* autonames = Tcl_NewStringObj(ObjStr(obj->cmdName), -1);
  Tcl_AppendToObj(autonames, "::", 2);
  Tcl_AppendToObj(autonames, global_strings[AUTONAMES], -1);
  INCR_REF_COUNT(autonames);

  valueObject = TclIncrVar2(in, autonames, Name, 1, 0);
  if (valueObject == NULL) {
    valueObject =
#if V81
            Tcl_SetObjVar2(in, ObjStr(autonames), name,
#else
	    Tcl_ObjSetVar2(in, autonames, Name,
#endif
			   global_objects[ZERO], 0);
  }
  result = Tcl_NewStringObj(name, -1);
  Tcl_AppendToObj(result, ObjStr(valueObject), -1);
  return result;
}

/*
 *  Filter-Commands
 */

#define FilterIsLast(obj) (obj->filterinfo.chains && \
                           obj->filterinfo.chains->filterPtr == 0 && \
                           obj->filterinfo.doFilters)

static XOTCLINLINE int
FilterIsFilterFrame (Tcl_Interp* in, CallFrame* framePtr) {
  int result = 0;

  if (framePtr && framePtr->isProcCallFrame && framePtr->procPtr) {
    Command* cmdPtr = framePtr->procPtr->cmdPtr;
    if (cmdPtr && !cmdPtr->cmdEpoch) {
      result = (Tcl_FindHashEntry(&XOTclFilterList, 
				  (char*) cmdPtr) != NULL);
    }
  }
  return result;
}

static XOTclClass*
SearchPrecedenceForFullName (char* cmdName, XOTclClass* cl) {
  XOTclClasses* pl=0;

  if (cl){
    if (!cl->order)
      cl->order = ComputePrecedence(cl);
    pl = cl->order;
  }
  while (pl && cl) {
    char* clName = pl->cl->nsPtr->fullName;
    if (strcmp(cmdName, clName) == 0) {
      return pl->cl;
    }
    pl = pl->next;
  }
  return NULL;
}

static Command*
FilterSearch (Tcl_Interp* in, char* name, XOTclClass* cl) {
  /* the search order is class/superclass - meta-class */
  Command* cmd;
  Tcl_CallFrame frame;	
  
  while (*name == ':') name++;
  Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)XOTclClassesNS, 0);
  
  SearchCMethod((XOTclObject*) NULL, cl, NULL, name, &cmd);
  
  /* If no filter is found -> search the meta-class */
  if (!cmd) {
    SearchCMethod((XOTclObject*) NULL, cl->object.cl, NULL, name, &cmd);
    if (!cmd) {
      cmd = (Command *) Tcl_FindCommand(in, name,(Tcl_Namespace *) NULL, 0);	
      if (cmd) {
	/* it is neither a meta-class nor found by SearchCMethod,
	   perhaps it is a full-qualified name on the precedence order of the
	   cl or meta-class */	
	char *cmdName = cmd->nsPtr->fullName;
	XOTclClass* mcl = SearchPrecedenceForFullName(cmdName, cl);

	if (!mcl)
	  mcl = SearchPrecedenceForFullName(cmdName, cl->object.cl);
	if (!mcl) 
	  cmd = 0;
      }
    }
  }   
  Tcl_PopCallFrame(in);
  
  return cmd;
}

static void
FilterInvalidateFlags (XOTclClass* cl) {
  XOTclClasses *saved = cl->order, *clPtr;
  cl->order = 0;
  clPtr = ComputeDependents(cl);

  while (clPtr != 0) {
      Tcl_HashSearch hSrch;
      Tcl_HashEntry* hPtr = &clPtr->cl->instances ? 
	Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;
      
      for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
	  XOTclObject* obj = (XOTclObject*)
	    Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
	  obj->filterinfo.valid = 0;
      }
      clPtr = clPtr->next;
  }

  RC(cl->order); cl->order = saved;
}

static void
FilterSetFlags (Tcl_Interp* in, XOTclObject* obj) {
  XOTclClasses *clPtr;
  assert(obj);

  if (!obj->cl->order)
    obj->cl->order = ComputePrecedence(obj->cl);
  clPtr = obj->cl->order;
  obj->filterinfo.doFilters = 0;

  if (clPtr) {
    while (clPtr != 0) {
      if (!obj->filterinfo.doFilters && clPtr->cl->filters)
	obj->filterinfo.doFilters = 1;
      if (obj->filterinfo.doFilters) break;
      clPtr = clPtr->next;
    }
  }
  obj->filterinfo.valid = 1;
}

static void
FilterRemoveEntry (Filter* fPtr, XOTclClass* cl) {
  Filter* filters = cl->filters;
  if (fPtr) {
    Tcl_HashEntry* hPtr;
    hPtr = Tcl_FindHashEntry(&XOTclFilterList,
			     (char*) fPtr->cmdPtr);
    if (hPtr) {	
      XOTclFilterListEntry* entry = 
	(XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
      if (entry) {
	if (entry->count > 0) {
	  entry->count--;
	} else {
	  ObjList* o;
	  while (entry->activeObjs) {
	    o = entry->activeObjs;
	    entry->activeObjs = entry->activeObjs->next;
	    ckfree ((char*) o);
	  }
	  ckfree((char*) entry);
	}
      }
      Tcl_DeleteHashEntry(hPtr);
    }

    if (fPtr == filters) {
      cl->filters = fPtr->next;
    } else {
      while (filters->next != fPtr) {
	filters = filters->next;
      }
      filters->next = fPtr->next;
    }
    fPtr->cmdPtr->refCount--;
    DECR_REF_COUNT (fPtr->simpleName);
    ckfree((char*) fPtr);
  }
  FilterInvalidateFlags(cl);
}

static void
FilterRemoveList (XOTclClass* cl) {
  while (cl->filters) {
    FilterRemoveEntry(cl->filters, cl);
  }
}

static int
FilterAdd (Tcl_Interp* in, XOTclClass* cl, char* name) {
  Filter *fPtr = 0;
  Command* cmdPtr = FilterSearch(in, name, cl);
  int nw = 0;
  Tcl_HashEntry* hPtr;
  char* simpleName;
  if (!cmdPtr)
    return TCL_ERROR;
  
  simpleName = Tcl_GetCommandName (in, (Tcl_Command)cmdPtr);
  
  fPtr = (Filter*) ckalloc ((unsigned) sizeof (Filter));
  
  fPtr->next = NULL;
  
  fPtr->cmdPtr = cmdPtr;
  fPtr->simpleName = Tcl_NewStringObj (simpleName, -1);
  INCR_REF_COUNT (fPtr->simpleName);
  fPtr->cmdPtr->refCount++;

  fPtr->filterClass =
    XOTclGetClass(in, NSCutXOTclClasses(cmdPtr->nsPtr->fullName));

  if (!fPtr->filterClass)
    return TCL_ERROR;
  
  hPtr = Tcl_FindHashEntry(&XOTclFilterList, (char*) fPtr->cmdPtr);
  
  if (hPtr) {
    XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
    if (entry) {
      entry->count++;
    }
  } else {	
    hPtr = Tcl_CreateHashEntry(&XOTclFilterList, (char*) fPtr->cmdPtr, &nw);
    if (nw) {
      XOTclFilterListEntry* entry = 
	(XOTclFilterListEntry*) ckalloc (sizeof(XOTclFilterListEntry));
      entry->count = 0;
      entry->activeObjs = 0;
      Tcl_SetHashValue(hPtr, (ClientData) entry);
    } else {
      return XOTclVarErrMsg(in,"Can't create filter list entry", (char*) NULL);
    }
  }
  
  if (cl->filters) {
    Filter* f = cl->filters;
    while (f->next != NULL) {
      f = f->next;
    }
    f->next = fPtr;
  } else {
    cl->filters = fPtr;
  }
  
  FilterInvalidateFlags(cl);
  return TCL_OK;
}

static void
FilterList (Tcl_Interp* in, Filter* fPtr) {
  Tcl_DString fullName;
  Tcl_ResetResult(in);
  for (; fPtr != NULL; fPtr = fPtr->next) {
    ALLOC_NAME_NS (&fullName, fPtr->cmdPtr->nsPtr->fullName,
		   ObjStr(fPtr->simpleName));
    Tcl_AppendElement(in, Tcl_DStringValue(&fullName));
    Tcl_DStringFree (&fullName);
  }
}

static XOTCLINLINE int
FilterDoFilters (Tcl_Interp* in, XOTclObject* obj) {
  int result;

  if (obj->filterinfo.valid == 0)
    FilterSetFlags(in, obj);
  result = obj->filterinfo.doFilters;
  return result;
}

static XOTCLINLINE int
FilterCallActualProc (Tcl_Interp* in, XOTclObject* obj, int objc, 
		      Tcl_Obj*  CONST objv[]) {
  int result = TCL_OK;
  int oc = objc+1;	
  /* call the actual proc, but not if it is destroy -> this handled after the
     filter is closed in DoDispatch */

  if (!isDestroyString(ObjStr(obj->filterinfo.chains->calledProc))) {
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov);

    ov[0] = obj->cmdName;
    ov[1] = obj->filterinfo.chains->calledProc;
    memcpy(ov+2, objv+1, sizeof(Tcl_Obj *)*(objc-1));

    /* call the actual method */
    result = ObjDispatch((ClientData) obj, in, oc, ov, 1);
    FREE_TCL_OBJECTS_ON_STACK(ov);
  } else {
    obj->filterinfo.chains->destroyCalled = 1;
  }
  if (obj && !obj->destroyCalled)
    obj->filterinfo.chains->actualProcCalled = 1;
  return result;
}

/* FilterInfoStack */
static XOTCLINLINE void
FilterChainInfoPush (XOTclObject* obj) {
  FilterChainInfo* h = (FilterChainInfo *) ckalloc(sizeof(FilterChainInfo));
  h->calledProc = 0;
  h->filterPtr = 0;
  h->callingProc = 0;
  h->callingClass = 0;
  h->callingObject = 0;
  h->next = obj->filterinfo.chains;
  obj->filterinfo.chains = h;
}

static XOTCLINLINE void
FilterChainInfoPop (XOTclObject* obj){
  FilterChainInfo *h = obj->filterinfo.chains;
  obj->filterinfo.chains = h->next;
  ckfree((char*)h);
}

/* List to maintain the filters currently active on an obj */

static XOTCLINLINE void
FilterActiveListPrepend (XOTclObject* obj, Command* c) {
  Tcl_HashEntry* hPtr;
  hPtr = Tcl_FindHashEntry(&XOTclFilterList,
			   (char*) c);
  if (hPtr) {	
    XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
    if (entry) {
      ObjList* o = (ObjList*) ckalloc (sizeof(ObjList));
      o->next = entry->activeObjs;
      o->objPtr = obj;
      entry->activeObjs = o;
    }
  }
}

static XOTCLINLINE void
FilterActiveListDelete (XOTclObject* obj, Command* cmdPtr) {
  Tcl_HashEntry* hPtr;
  hPtr = Tcl_FindHashEntry(&XOTclFilterList,
			   (char*) cmdPtr);
  if (hPtr) {
    XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
    if (entry) {
      ObjList* c = entry->activeObjs, *del = c;
      if (!c)
	return;
      if (c->objPtr == obj) {
	entry->activeObjs = c->next;
      } else {
	while (c->next && (obj != c->next->objPtr)) {
	  c = c->next;
	}
	del = c->next;
	if (del) {
	  c->next = del->next;
	}
      }
      if (del) {
	ckfree((char*)del);}
    }
  }
}

static XOTCLINLINE void
FilterActiveDeleteObj(XOTclObject* obj) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = &XOTclFilterList ?
	  Tcl_FirstHashEntry(&XOTclFilterList, &hSrch) : 0;

  while (hPtr != 0) {
    XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
    if (entry) {
      ObjList* o = entry->activeObjs, *h = 0;
      while (o) {
	if (o->objPtr == obj) {
	  Command* cmdPtr = (Command*) Tcl_GetHashKey(&XOTclFilterList, hPtr);
	  h = o->next; 
	  if (cmdPtr) {
	    FilterActiveListDelete(obj, cmdPtr);
	  }
	}
	o = h;
      }
    }
    hPtr = Tcl_NextHashEntry(&hSrch);
  }
}

static XOTCLINLINE int
FilterActive (Command* cmdPtr) {
  Tcl_HashEntry* hPtr;
  hPtr = Tcl_FindHashEntry(&XOTclFilterList,
			   (char*) cmdPtr);
  if (hPtr) {
    XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
    if (entry) {
      if (entry->activeObjs) {
	return 1;
      } 
    }
  }
  return 0;
}

static XOTCLINLINE int
FilterActiveOnObj (XOTclObject* obj, Command* cmdPtr) {
  Tcl_HashEntry* hPtr;
  hPtr = Tcl_FindHashEntry(&XOTclFilterList,
			   (char*) cmdPtr);
  if (hPtr) {
    XOTclFilterListEntry* entry = (XOTclFilterListEntry*) Tcl_GetHashValue (hPtr);
    if (entry) {
      ObjList* o = entry->activeObjs;
      while (o) {
	if (o->objPtr == obj)
	  return 1;
	o = o->next;
      }
    }
  }
  return 0;
}

static XOTCLINLINE void
FilterComputeNext (XOTclObject* obj) {
  Filter* fPtr = 0;
  XOTclClasses* pl;

  /* compute the next filter, if one is there ... overstep a found entry
     if it belongs to a regClass with already deregistered filters */
  do {
    if (!obj->cl->order)
      obj->cl->order = ComputePrecedence(obj->cl);
    pl = obj->cl->order;
    if (obj->filterinfo.chains) {
      /* there is a chain on the object */
      if (obj->filterinfo.chains->filterPtr) {
	fPtr = obj->filterinfo.chains->filterPtr->next;
	if (fPtr && obj->filterinfo.chains->regClass->filters) {
	  obj->filterinfo.chains->filterPtr = fPtr;
	  return;
	} else {
	  while (pl && pl->cl != obj->filterinfo.chains->regClass)
	    pl = pl->next;
	  if (pl) {
	    pl = pl->next;
	  } else {
	    obj->filterinfo.chains->filterPtr = 0;
	    return;
	  }
	}
      }

      while (pl != 0) {
	fPtr = pl->cl->filters;
	if (fPtr) break;
	pl = pl->next;
      }
      obj->filterinfo.chains->filterPtr = fPtr;
      if (pl) {
	obj->filterinfo.chains->regClass = pl->cl;
      }
    }
  } while (obj->filterinfo.chains->filterPtr != 0 &&
	   (obj->filterinfo.chains && obj->filterinfo.chains->regClass &&
	    obj->filterinfo.chains->regClass->filters == 0));
}

static XOTCLINLINE void
FilterSkipActiveFilters (XOTclObject* obj) {
  while (obj->filterinfo.chains->filterPtr &&
	 FilterActiveOnObj (obj, obj->filterinfo.chains->filterPtr->cmdPtr)) {
    FilterComputeNext(obj);
  }
}

static XOTCLINLINE void
FilterCutInactiveFrames (Tcl_Interp* in, CallFrame **savedFramePtr,
                 CallFrame **savedVarFramePtr) {
  Interp *iPtr = (Interp *) in;
  CallFrame *l, *oldFrame, *lastFrame, *lastVarFrame;
  int firstFrame = 1;
  int isObj;
  int level = (iPtr->framePtr) ? iPtr->framePtr->level : 0;
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent
    *bot = cs->content,
    *csc = cs->top;

  /* we increment the frameCutRound in order to be able to determine
     which frames have been cut in this cutting round */
  cs->frameCutRound++;

  /* for the time of the actual proc we create an enviroment
     without the filter calls (e.g. for the reason that uplevel/upvar
     function correctly) */

  *savedFramePtr = oldFrame = iPtr->framePtr;
  *savedVarFramePtr = iPtr->varFramePtr;
  iPtr->framePtr = 0;
  iPtr->varFramePtr = 0;
  lastFrame = 0; lastVarFrame = 0;
  /*
   * copy all frames that aren't inactive filters
   */
  while (oldFrame != 0) {
    while (csc > bot && csc->frameCut > 0) {
      csc--;
    }

    isObj = (oldFrame->nsPtr && (XOTclGetObject(in, oldFrame->nsPtr->fullName)
				 || XOTclGetClass(in, NSCutXOTclClasses(oldFrame->nsPtr->fullName))));
  
    if (!isObj ||
	!FilterIsFilterFrame(in, oldFrame) ||
	(oldFrame->procPtr && 
	 FilterActiveOnObj(csc->self, oldFrame->procPtr->cmdPtr))) {
      CallFrame *f;

      f = (CallFrame *) ckalloc(sizeof(CallFrame));
      memcpy(f, oldFrame, sizeof(CallFrame));
      if (lastFrame)
        lastFrame->callerPtr = f;
      lastFrame = f;

      if (*savedVarFramePtr == oldFrame) {
        iPtr->varFramePtr = f;
      }
      if (firstFrame) {
        iPtr->framePtr = f;
        firstFrame = 0;
      }
    } else {
      level--;
      csc->frameCut = cs->frameCutRound;
    }

    /* determine corresponding callstack entry */
    if (isObj)
      csc--;

    oldFrame = oldFrame->callerPtr;
  }

  if (lastFrame)
    lastFrame->callerPtr = 0;

  if(iPtr->varFramePtr == 0)
    iPtr->varFramePtr = iPtr->framePtr;

  /*
   * correct callerVarPtr information
   */

  csc = cs->top;

  oldFrame = *savedFramePtr;
  l = iPtr->framePtr;
  while (oldFrame != 0) {

    /* overstep cutted frames, except frames cutted in this round */
    while (csc > bot && csc->frameCut > 0 && 
	   csc->frameCut != cs->frameCutRound)
      csc--;
    if (csc <= bot)
      break;

    if (!FilterIsFilterFrame(in, oldFrame) || csc <= bot ||
	(oldFrame->procPtr && 
	 FilterActiveOnObj(csc->self, oldFrame->procPtr->cmdPtr))) {
      CallFrame *h1 = oldFrame->callerPtr, *h2 = l->callerPtr;
      while (h1 && oldFrame->callerVarPtr != h1) {
	h1 = h1->callerPtr;
	if (h2 && (!FilterIsFilterFrame(in, h1) ||
	    (h1->procPtr &&
	     FilterActiveOnObj(csc->self, h1->procPtr->cmdPtr)))) {
	  h2 = h2->callerPtr;
	}
      }
      if (h1 && h2) {
        l->callerVarPtr = h2;
      } else {
        l->callerVarPtr = l->callerPtr;
      }
      l = l->callerPtr;
    }
    if (oldFrame->nsPtr && (XOTclGetObject(in, oldFrame->nsPtr->fullName)
	|| XOTclGetClass(in, NSCutXOTclClasses(oldFrame->nsPtr->fullName)))) {
      csc--;
    }
    oldFrame = oldFrame->callerPtr;
  }

  /*
   * correct level information
   */
 
  level = 0;

  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr) 
    level++;
  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr) {
    l->level = level;
    level--;
  }
}

static XOTCLINLINE void
FilterRestoreFrames (Tcl_Interp* in, CallFrame *savedFramePtr,
                     CallFrame *savedVarFramePtr) {
  Interp *iPtr = (Interp *) in;
  CallFrame* save = iPtr->framePtr, *copy = save, *original, *l;
  int level = 0;
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent *csc = cs->top;

  /* reset to old callframes */
  iPtr->framePtr = savedFramePtr;
  original = savedFramePtr;
  iPtr->varFramePtr = savedVarFramePtr;

  while (original != 0 && copy != 0) {
    int isObj = (original->nsPtr && (XOTclGetObject(in, original->nsPtr->fullName)
			  || XOTclGetClass(in, NSCutXOTclClasses(original->nsPtr->fullName))));
  
    if (!(isObj && csc->frameCut)) {
      if (copy->procPtr && copy->procPtr->cmdPtr
	  && original->procPtr && original->procPtr->cmdPtr
	  && copy->procPtr->cmdPtr == original->procPtr->cmdPtr) {
	/* if local var table is created on copy -> let the original
	   use the created one */
	if (original->varTablePtr == 0)
	  original->varTablePtr = copy->varTablePtr;
	copy = copy->callerPtr;
      }
    }
    original = original->callerPtr;
    if (isObj)
      csc--;
  }
  /*
   * correct level information in original frames
   */
  level = 0;
  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr) 
    level++;
  
  for(l = iPtr->framePtr; l != 0; l = l->callerVarPtr) {
    l->level = level;
    level--;
  }

  /* free memory */
  while (save != 0) {
    CallFrame *f= save;
    save = save->callerPtr;
    ckfree((char*)f);
  }

  /* Update the XOTcl CallStack cutted frames of this round */
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (csc->frameCut == cs->frameCutRound) {
      csc->frameCut = 0;
    }
  }
  cs->frameCutRound--;
}

/* determine whether a filter is active on the current callstack on not */
static int
FilterEntryOnCallStack () {
  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent *csc;
  for (csc = &cs->content[1]; csc <= cs->top; csc++) {
    if (csc->isFilterEntry) {
      return 1;
    }
  }
  return 0;
}

/*
 * Assertions
 */

static void
AssertionDeleteList (Assertion* alist) {
    Assertion* delAss;
    while (alist) {
	delAss = alist;
	alist = alist->next;
	DECR_REF_COUNT (delAss->content);
	ckfree((char*)delAss);
    }
}

static Assertion*
AssertionNewElement (Tcl_Interp* in, Tcl_Obj* ov) {
    Assertion* elt = (Assertion *)ckalloc (sizeof(Assertion));
    INCR_REF_COUNT (ov);
    elt->content = ov;
    return elt;
}

static Assertion*
AssertionNewList (Tcl_Interp* in, Tcl_Obj* aObj){
    Tcl_Obj **ov; int oc;
    int i;
    Assertion *last=NULL, *result=NULL;

    if (Tcl_ListObjGetElements(in, aObj, &oc, &ov)==TCL_OK) {
	if (oc == 0) return NULL;
	result = last = AssertionNewElement (in, ov[0]);
	for (i=1; i<oc; i++) {
	    Assertion* elt = AssertionNewElement (in, ov[i]);
	    if (elt != NULL) {
		last->next = elt;
		last = elt;
		
	    }
	}
    }
    if (last) last->next = NULL;

    return result;
}

static void
AssertionList (Tcl_Interp* in, Assertion* alist) {
  for (; alist!=NULL; alist = alist->next) {
    Tcl_AppendResult(in,"{",ObjStr(alist->content),"}", (char *) NULL);
    if (alist->next != NULL) Tcl_AppendResult(in, " ", (char*) NULL);
  }
}

static void
AssertionListCheckOption (Tcl_Interp* in, XOTclObject* obj) {
  if (obj->checkoptions & CHECK_OBJINVAR)
    Tcl_AppendElement(in, "invar");
  if (obj->checkoptions & CHECK_CLINVAR)
    Tcl_AppendElement(in, "instinvar");
  if (obj->checkoptions & CHECK_PRE)
    Tcl_AppendElement(in, "pre");
  if (obj->checkoptions & CHECK_POST)
    Tcl_AppendElement(in, "post");
}

static ProcAssertion*
AssertionFindProcs (AssertionStore* aStore, char* name) {
  Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&aStore->procs, name);
  if (hPtr == NULL) return NULL;
  return (ProcAssertion*) Tcl_GetHashValue (hPtr);
}

static void
AssertionRemoveProc(AssertionStore* aStore, char* name) {
    Tcl_HashEntry *hPtr = Tcl_FindHashEntry(&aStore->procs, name);
    if (hPtr) {
      ProcAssertion* procAss = (ProcAssertion*) Tcl_GetHashValue (hPtr);
      AssertionDeleteList (procAss->pre);
      AssertionDeleteList (procAss->post);
      ckfree((char*)procAss);
      Tcl_DeleteHashEntry(hPtr);
    }
}

static void
AssertionAddProc(Tcl_Interp* in,char* name, AssertionStore* aStore, 
		 Tcl_Obj* pre, Tcl_Obj* post) {
  int nw = 0; Tcl_HashEntry* hPtr = NULL;
  ProcAssertion* procs = (ProcAssertion *) ckalloc (sizeof(ProcAssertion));

  AssertionRemoveProc (aStore,name);
  procs->pre = AssertionNewList (in, pre);
  procs->post = AssertionNewList (in, post);
  hPtr = Tcl_CreateHashEntry(&aStore->procs,name, &nw);
  if (nw) Tcl_SetHashValue(hPtr, (ClientData) procs);
}

static void
AssertionRemoveStore(AssertionStore* aStore) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = Tcl_FirstHashEntry(&aStore->procs, &hSrch);

  while (hPtr) {
    AssertionRemoveProc (aStore,Tcl_GetHashKey(&aStore->procs,hPtr));
    hPtr = Tcl_FirstHashEntry(&aStore->procs, &hSrch);
  }
  Tcl_DeleteHashTable(&aStore->procs);
  AssertionDeleteList (aStore->invariants);
}

static XOTCLINLINE int
AssertionCheckList (Tcl_Interp* in, XOTclObject* obj, Assertion* alist) {
  char* strResult;
  int result;
  Assertion* checkFailed = NULL;
  Tcl_Obj* savedObjResult = Tcl_GetObjResult(in);

  INCR_REF_COUNT(savedObjResult);
  Tcl_ResetResult(in);

  while ((alist != NULL) && (checkFailed == NULL)) {
    /* Eval instead of IfObjCmd => the substitutions in the 
       conditions will be done by Tcl */
    char* assStr = ObjStr(alist->content), *c = assStr;
    int comment = 0;
    Tcl_Obj* ifCmd;

    while (c != 0 && *c != '\0') {
      if (*c == '#') {
	comment = 1; break;
      }
      c++;
    }
    if (!comment) {
      Tcl_CallFrame frame;
      Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);
      ifCmd = Tcl_NewStringObj("if {",-1); INCR_REF_COUNT (ifCmd);
      Tcl_AppendStringsToObj(ifCmd, assStr,
			     "} {return 1} else {return 0}", (char*)NULL);

      /*
      fprintf(stderr, "Checking Assertion %s ", assStr); 
      */
      result = EvalObj (in, ifCmd);

      DECR_REF_COUNT (ifCmd);
      strResult = Tcl_GetStringResult(in);
      if ((result == TCL_ERROR) || (*strResult != '1')) checkFailed = alist;
      /*
      if (!checkFailed) 
	fprintf(stderr, "...ok\n");
      else 
	fprintf(stderr, "...failed\n");
      */
      Tcl_PopCallFrame(in);
    }
    alist = alist->next;
  }

  if (checkFailed != NULL) {
    DECR_REF_COUNT(savedObjResult);
    return XOTclVarErrMsg (in,
			   "Error - Assertion failed check: {",
			   ObjStr(checkFailed->content), "} in proc '",
			   GetSelfProc(in),"'",(char*)NULL);
  }
  
  Tcl_SetObjResult(in, savedObjResult);
  DECR_REF_COUNT(savedObjResult);
  return TCL_OK;
}

static XOTCLINLINE int
AssertionCheckInvars(Tcl_Interp* in, XOTclObject* obj) {
  int result=TCL_OK;
 if (obj->checkoptions & CHECK_OBJINVAR)
   result = AssertionCheckList (in, obj, obj->assertions.invariants);
 if  ((result != TCL_ERROR) && (obj->checkoptions & CHECK_CLINVAR))
   result = AssertionCheckList (in, obj, obj->cl->assertions.invariants);
 return result;
}

static XOTCLINLINE int
AssertionCheckPre(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl, char* method) {
  ProcAssertion* procs;
  int result = TCL_OK;
  if (!(obj->checkoptions & CHECK_NONE)) {   
    if (cl) {
      procs = AssertionFindProcs (&cl->assertions, method);
    } else {
      procs = AssertionFindProcs (&obj->assertions, method);
    }
    if (procs && (obj->checkoptions & CHECK_PRE)) 
      result = AssertionCheckList (in, obj, procs->pre);
    if (result != TCL_ERROR)
      result = AssertionCheckInvars(in, obj);
  }
  return result;
}

static XOTCLINLINE int
AssertionCheckPost(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl, char* method) {
  ProcAssertion* procs;
  int result = TCL_OK;
  if (!(obj->checkoptions & CHECK_NONE)) {
    if (cl) {
      procs = AssertionFindProcs (&cl->assertions, method);
    } else {
      procs = AssertionFindProcs (&obj->assertions, method);
    }
    if (procs && (obj->checkoptions & CHECK_POST)) 
      result = AssertionCheckList (in, obj, procs->post);
    if (result != TCL_ERROR)
      result = AssertionCheckInvars(in, obj);
  }
  return result;
}

/*
 * Meta Data
 */

static void
MetaDataRemoveDepending(XOTclClass* cl, char* name) {
  XOTclClasses *saved = cl->order, *clPtr;
  cl->order = 0;
  clPtr = ComputeDependents(cl);

  while (clPtr != 0) {
      Tcl_HashSearch hSrch;
      Tcl_HashEntry* hPtr = &clPtr->cl->instances ? 
	Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;
      for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
	  XOTclObject* obj = (XOTclObject*)
	    Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
	  Tcl_HashEntry* h1Ptr = Tcl_FindHashEntry(&obj->metaData, name);
	  if (h1Ptr) {
	      Tcl_DeleteHashEntry(h1Ptr);
	  }
      }
      clPtr = clPtr->next;
  }

  RC(cl->order); cl->order = saved;
}

static int
MetaDataInheritance (XOTclObject* obj, char* name) {
    XOTclClasses *clPtr;
    assert(obj);

    if (!obj->cl->order)
	obj->cl->order = ComputePrecedence(obj->cl);
    clPtr = obj->cl->order;
    if (clPtr) {
      while (clPtr != 0) {
	  if (Tcl_FindHashEntry(&clPtr->cl->object.metaData, name)) {
	      return 1;
	  }
	  clPtr = clPtr->next;
      }
    }
    return 0;
}

/*
 * Per-Object-Mixins
 */

/* FilterInfoStack */
static XOTCLINLINE
void MixinStackPush (XOTclObject* obj) {
  MixinStack* h = (MixinStack *) ckalloc(sizeof(MixinStack));
  h->currentCmdPtr = 0;
  h->mixinChainOn = 0;
  h->mixinInit = 0;
  h->next = obj->mixinStack;
  obj->mixinStack = h;
}

static XOTCLINLINE
void MixinStackPop (XOTclObject* obj){
  MixinStack *h = obj->mixinStack;
  obj->mixinStack = h->next;
  ckfree((char*)h);
}

/* Cmd List Add/Remove */

static void
RemoveCmdList(CmdList* cl) {
  while (cl != 0) {
    CmdList* n = cl->next;
    ckfree((char*)cl); cl = n;
  }
}

static XOTCLINLINE void
AddCmdList(CmdList** cList, Command* c) {
  CmdList *l = *cList, 
    *new = (CmdList*) ckalloc ((unsigned) sizeof (CmdList));
  new->cmdPtr = c;
  new->next = NULL;
  if (l) {
    while (l->next) l = l->next; 
    l->next = new;
  } else
    *cList = new;
}

/* 
 * Computes a linearized order from the mixins, their inheritance
 * chains and the normal class hierarchy of the object each class is
 * used exactly once and the last occurence stays 
 */

static XOTCLINLINE void
MixinComputeOrder(Tcl_Interp* in, XOTclObject* obj) {
  Mixin* m = obj->mixins;
  XOTclClasses* mixinList = 0, *pl, *checker;
  RemoveCmdList(obj->mixinOrder);
  obj->mixinOrder = 0;

  /*fprintf(stderr, "Mixin Order:\n First List: ");*/

  while (m) {
    XOTclClass* mCl = GetClassFromCmdPtr(in, m->cmdPtr);
    if (mCl) {
      if (!mCl->order) 
	mCl->order = ComputePrecedence(mCl);
      pl = mCl->order;
      while (pl) {
	/*fprintf(stderr, " %s,", pl->cl->object.nsPtr->fullName);*/
	AC(&mixinList, pl->cl);
	pl = pl->next;
      }
    }   
    m = m->next;
  }
  
  /* use no duplicates & no classes of the precedence order 
     on the resulting list */
  while (mixinList) {
    checker = mixinList->next;
    while (checker) {
      if (checker->cl == mixinList->cl) break;
      checker = checker->next;
    }
    /* if no duplicate -> check obj->cl hierachy */
    if (checker == 0) { 
      if (!obj->cl->order) 
	obj->cl->order = ComputePrecedence(obj->cl);
      checker = obj->cl->order;
      while (checker) {
	if (checker->cl == mixinList->cl) break;
	checker = checker->next;
      }
      /* if class is also not found on precedence order -> add it*/
      if (checker == 0) {
	AddCmdList(&obj->mixinOrder, (Command*) mixinList->cl->object.id);
	/*fprintf(stderr, "Adding %s,\n", mixinList->cl->object.nsPtr->fullName);*/
      }   
    } 
    mixinList = mixinList->next;
  }
}

static void
MixinAdd(Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl) {
  Mixin* nm = (Mixin*)ckalloc(sizeof(Mixin));
  nm->cmdPtr = (Command *) cl->object.id;
  nm->cmdPtr->refCount++;
  nm->next = 0;

  if (obj->mixins) {
    Mixin* m = obj->mixins;
    while (m->next != NULL) {
      m = m->next;
    }
    m->next = nm;
  } else {
    obj->mixins = nm;
  }
}

static void
MixinRemoveList (XOTclObject* obj) {
  Mixin *del;

  while (obj->mixins) {
    del = obj->mixins;
    obj->mixins = obj->mixins->next;
    del->cmdPtr->refCount--;
    ckfree((char*)del);
  }
  if (obj->mixinStack) {
    obj->mixinStack->currentCmdPtr = 0;
  }
  while (obj->mixinStack)
    MixinStackPop(obj);
}

static void
MixinRemove(XOTclObject* obj, Mixin* dm) {
  Mixin* m = obj->mixins, *del = 0;
  if (m == 0)
    return;
  if (m == dm) {
    obj->mixins = m->next;
    del = m;
  } else {
    while (m->next && m->next != dm) {
      m = m->next;
    }
    if (m->next == dm) {
      del = dm;
      m->next = dm->next;
    }
  }
  if (del) {
    MixinStack* h = obj->mixinStack;
    if (obj->mixins == 0 && h) {
      obj->mixinStack->currentCmdPtr = 0;
    }

    /* if the current mixin is to be deleted, we make its predecessor
       current mixin */
    if (obj->mixinStack->currentCmdPtr == del->cmdPtr) {
      if (obj->mixinOrder && del->cmdPtr != obj->mixinOrder->cmdPtr) {
	CmdList* predecessor = obj->mixinOrder;
	while (predecessor->next  && 
	       predecessor->next->cmdPtr != del->cmdPtr)
	  predecessor = predecessor->next;
	if (predecessor->next)
	  obj->mixinStack->currentCmdPtr = predecessor->next->cmdPtr;
      } else {
	obj->mixinStack->currentCmdPtr = 0;
      }
    }

    /* actually delete the mixin entry */
    del->cmdPtr->refCount--;
    ckfree((char*)del);
  }
}

static void MixinRemoveFromCmdPtr(XOTclObject* obj, Command* c) {
  Mixin* m = obj->mixins;
  while (m) {
    if (m->cmdPtr == c) {
      MixinRemove(obj, m);
      break;
    }
    m = m->next;
  }
}

static XOTCLINLINE MixinSeekCurrent(Tcl_Interp* in, XOTclObject* obj, 
				    CmdList** cmdList) {
  Command* currentCmdPtr = obj->mixinStack->currentCmdPtr;
  if (!obj->mixinOrder) 
    MixinComputeOrder(in, obj);

  *cmdList = obj->mixinOrder;

  /* go forward to current class */
  while (*cmdList && currentCmdPtr) {
    if ((*cmdList)->cmdPtr == currentCmdPtr) currentCmdPtr = 0;
    *cmdList = (*cmdList)->next;
  }
}

static XOTCLINLINE Command*
MixinSearchProc(Tcl_Interp* in, XOTclObject* obj, char* methodName,
		   int* mixinFoundVar, XOTclClass** cl,
		   Tcl_ObjCmdProc** proc, ClientData* cp) {
  Command* procInfo = NULL;
  CmdList *cmdList;
  XOTclClass* class;
  MixinSeekCurrent(in, obj, &cmdList);

  while (cmdList) { 
    if(cmdList->cmdPtr->cmdEpoch) {
      MixinRemoveFromCmdPtr(obj, cmdList->cmdPtr);
      /* invalidate the order & seek start class */
      obj->mixinOrder = 0;
      MixinSeekCurrent(in, obj, &cmdList);
    } else {
      class = GetClassFromCmdPtr(in, cmdList->cmdPtr);
      if (class) {
	procInfo = FindMethod (methodName, class->nsPtr);
	if (procInfo) {
	  *cl = class;
	  *proc = procInfo->objProc;
	  *cp   = procInfo->objClientData;
	  *mixinFoundVar = 1;
	  obj->mixinStack->currentCmdPtr = cmdList->cmdPtr;
	  break;
	} else {
	  cmdList = cmdList->next;
	}
      }
    }
  }
  if (!procInfo) 
    obj->mixinStack->currentCmdPtr = 0;
   return procInfo;
}

static void
MixinInfo (Tcl_Interp* in, XOTclObject* obj, char* pattern) {
  Mixin* m = obj->mixins;
  if (pattern == 0) {
    Tcl_ResetResult(in);
    while (m) {
      XOTclClass* mixinClass = GetClassFromCmdPtr(in, m->cmdPtr);
      if (mixinClass) {
	Tcl_AppendElement(in, className(mixinClass));
	m = m->next;
      } else {
	Mixin* del = m;
	m = m->next;
	MixinRemove(obj, del);
      }
    }
  } else {
    XOTclClass* isc = XOTclGetClass(in, pattern);
    if (isc == 0)
      m = 0;
    while (m != 0) {
      if (GetClassFromCmdPtr(in, m->cmdPtr) == isc) {
	Tcl_SetObjResult(in, global_objects[ONE]);
	break;
      }
      m = m->next;
    }
    if (m == 0)
      Tcl_SetObjResult(in, global_objects[ZERO]);
  }
}

static void
MixinInvalidateObjOrders (XOTclClass* cl) {
  XOTclClasses *saved = cl->order, *clPtr;
  cl->order = 0;
  clPtr = ComputeDependents(cl);
  
  while (clPtr != 0) {
    Tcl_HashSearch hSrch;
    Tcl_HashEntry* hPtr = &clPtr->cl->instances ? 
      Tcl_FirstHashEntry(&clPtr->cl->instances, &hSrch) : 0;
    
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      XOTclObject* obj = (XOTclObject*)
	Tcl_GetHashKey(&clPtr->cl->instances, hPtr);
      obj->mixinOrder= 0;
    }
    clPtr = clPtr->next;
  }
  
  RC(cl->order); cl->order = saved;
}

/*
 * Short Form for class/obj nesting
 * Make nested Obj/Cl --- like:   objName Class aggObj
 *                                clName MetaClass nestCl
 */

static int
IsMetaClass (Tcl_Interp* in, XOTclClass* cl) {
  /* check if cl is a meta-class by looking up the location of alloc*/
  XOTclClasses* pl  = ComputePrecedence(cl);
  XOTclClass* thecls = XOTclGetClass (in, "Class");
  if(pl){
    while (pl != 0) {
      if (FindMethod ("alloc",pl->cl->object.nsPtr)) break;
      pl = pl->next;
    }
    if (!pl) return 0;
    if(pl->cl == thecls)
      return 1;
  }
  return 0;
}

static int
MakeObject(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[],
	   XOTclClass* Class) {
   XOTclObject* obj = XOTclAsObject(in, cd);
  int result, i;
  DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc,ov);

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 3) return XOTCL_UNKNOWN;

  for (i=3; i< objc; i++)
    ov[i] = objv[i];

  ov[0] = objv[1];
  ov[1] = global_objects[CREATE];
  ov[2] = Tcl_NewStringObj(obj->nsPtr->fullName,-1);
  INCR_REF_COUNT(ov[2]);
  Tcl_AppendStringsToObj(ov[2], "::", ObjStr(objv[2]), (char*)NULL);

  result =  XOTclObjDispatch((ClientData) Class, in, objc, ov);

  DECR_REF_COUNT(ov[2]);
  FREE_TCL_OBJECTS_ON_STACK(ov);
  return result;
}

static int
MakeClass(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[], 
	  XOTclClass* Class) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  int result, i;
  DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc,ov);

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 3) return XOTCL_UNKNOWN;

  for (i=3; i< objc; i++)
    ov[i] = objv[i];

  ov[0] = objv[1];
  ov[1] = global_objects[CREATE];
  ov[2] = Tcl_NewStringObj(cl->object.nsPtr->fullName,-1);
  INCR_REF_COUNT(ov[2]);
  Tcl_AppendStringsToObj(ov[2], "::", ObjStr(objv[2]), (char*)NULL);

  result =  XOTclObjDispatch((ClientData) Class, in, objc, ov);

  DECR_REF_COUNT(ov[2]);
  FREE_TCL_OBJECTS_ON_STACK(ov);
  return result;
}

/*
 * method dispacth
 */

static XOTCLINLINE int
FilterDispatch (Tcl_Interp* in, XOTclObject* obj, int objc, Tcl_Obj* CONST objv[]) {
  int result;
  /*Tcl_ObjCmdProc *proc;*/
  ClientData cd;
  Filter* filterEntry = 0;

  if (!obj->filterinfo.chains) {
    return XOTclVarErrMsg(in, 
			       "Tried to dispatch Filter without a chain on the stack",
			       (char*)NULL);
  }

  filterEntry = obj->filterinfo.chains->filterPtr;

  if (filterEntry->cmdPtr->cmdEpoch) {
      FilterRemoveEntry (obj->filterinfo.chains->filterPtr, 
			 obj->filterinfo.chains->regClass);
      return XOTCL_NOT_FILTERED;
    }

  cd = filterEntry->cmdPtr->objClientData;

  FilterActiveListPrepend (obj, filterEntry->cmdPtr);

  result = callProcCheck(cd, in, objc, objv, filterEntry->cmdPtr,
			 obj, filterEntry->filterClass,
			 Tcl_GetCommandName(in,(Tcl_Command)filterEntry->cmdPtr),
			 1, 0, 0);

  FilterActiveListDelete (obj, filterEntry->cmdPtr);

  if (result != TCL_OK)
    result = TCL_ERROR;

  return result;
}

static XOTCLINLINE int
FilterStartChain (Tcl_Interp* in, XOTclObject* obj, 
		  int objc, Tcl_Obj* CONST objv[]) {
  int destroyCalled;
  Tcl_Obj* savedName;
  int result;
  FilterChainInfo* chains;
  CallStackContent *csc;

  FilterChainInfoPush(obj);
  obj->filterinfo.chains->filterPtr = 0;

  FilterComputeNext (obj);
  FilterSkipActiveFilters(obj);
  if (!obj->filterinfo.chains->filterPtr) {
    FilterChainInfoPop(obj);
    return XOTCL_NOT_FILTERED;
  }
  chains = obj->filterinfo.chains;

  chains->calledProc = objv[1];	
  chains->procVarFramePtr = ((Interp *)in)->varFramePtr;
  csc = CallStackGetFrame(in);

  chains->callingProc = Tcl_NewStringObj(csc->procName,-1);
  INCR_REF_COUNT (chains->callingProc);

  chains->callingClass =
    csc->class ? csc->class->object.cmdName : global_objects[EMPTY];
  INCR_REF_COUNT (chains->callingClass);

  chains->callingObject =
    csc->self ? csc->self->cmdName : global_objects[EMPTY];
  INCR_REF_COUNT (chains->callingObject);

  chains->destroyCalled = 0;
  chains->actualProcCalled = 0;

  savedName = obj->cmdName;	
  INCR_REF_COUNT(savedName);

  result = FilterDispatch (in, obj, objc-1, objv+1);

  /* the filtered call may have been a destroy on the obj */
  if (obj && !obj->destroyCalled &&
      XOTclGetObject(in, ObjStr(savedName))) {
    DECR_REF_COUNT (chains->callingObject);
    DECR_REF_COUNT (chains->callingClass);
    DECR_REF_COUNT (chains->callingProc);
    destroyCalled = chains->destroyCalled;
    FilterChainInfoPop(obj);
  } else {
    destroyCalled = 0;
  }

  if (destroyCalled) {
    CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
    CallStackContent *csc;
    int oc = objc;
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov);

    ov[0] = obj->cmdName;
    ov[1] = global_objects[DESTROY];

    obj->destroyCalled = 1;
    for (csc = &cs->content[1]; csc <= cs->top; csc++) {
      if (obj == csc->self)
	csc->self = 0;
    }

    memcpy(ov+2, objv+2, sizeof(Tcl_Obj *)*(objc-2));

    result = ObjDispatch((ClientData) obj, in, oc, ov, 1);
    FREE_TCL_OBJECTS_ON_STACK(ov);
  }
  return result;
}

static XOTCLINLINE int
DoDispatch (ClientData cd, Tcl_Interp* in, XOTclObject* obj,
	    int objc, Tcl_Obj *CONST objv[], int noFilters) {
  int result = TCL_OK, filtered = 0;
  int mixinStackPushed = 0;
  char *methodName = ObjStr(objv[1]);
  int isDestroyed = 0;

  /* if filters are registered for the object's classes we create a new filter
     chain and then we start a filter chain  */
  if (!noFilters && !isNextString(methodName) &&
	FilterDoFilters(in, obj)) {
    result = FilterStartChain(in, obj, objc, objv);
    if (result != XOTCL_NOT_FILTERED)
      filtered = 1;
  }

   /*
    *  not filtered =>
    *  normal dispatch
    */
  if (!filtered) {
    XOTclClass     *cl = 0;
    ClientData     cp = 0;
    Tcl_ObjCmdProc *proc = 0;
    Command        *procInfo;
    Tcl_Obj        *objName    = obj->cmdName;
    int            isMixinEntry = 0;

    INCR_REF_COUNT(objName);

    /* check if a per object mixin is to be called */
    if (obj->mixins != 0) {
      if (!isNextString(methodName)) {
	XOTclClass     *mcl=0;
	Tcl_ObjCmdProc *mproc = 0;
	Command        *mProcInfo =0;
	ClientData     mcp;
	MixinStackPush(obj);
	
        mixinStackPushed = 1;
	mProcInfo = MixinSearchProc(in, obj, methodName,
				    &obj->mixinStack->mixinChainOn,
				    &mcl, &mproc, &mcp);
	if (mproc) {
	  isMixinEntry = 1;
	  cl = mcl;
	  proc = mproc;
	  cp = mcp;
	  procInfo = mProcInfo;
	}
      }
    }

    /* if no mixin is found => do ordinary method lookup */
    if (proc == 0) {
      procInfo = FindMethod(methodName, obj->nsPtr);
      if (procInfo) {
	cl = 0;
      } else {
	cl = SearchCMethod(obj, obj->cl, NULL, methodName, &procInfo);
      }
      if (procInfo) {
	proc = procInfo->objProc;
	cp   = procInfo->objClientData;
      } else {
	proc = 0;
	cp   = 0;
      }
    }

    if (proc) {
      if (cp == 0)
	cp = cd;
      result = callProcCheck(cp, in, objc-1, objv+1, procInfo,
			     obj, cl, methodName, 0, isMixinEntry, &isDestroyed);

      if (result == TCL_ERROR) {
	if (!isDestroyed) {
	  XOTclErrInProc(in, obj, cl, methodName);
	} else {
	  XOTclVarErrMsg(in, 
			 "destroy failed (probably wrong # args?) or 'self' called after destroy",
			 (char*) 0);
	}
      }
    } else
      result = XOTCL_UNKNOWN;
	
    DECR_REF_COUNT(objName);
  }

  if (!isDestroyed && obj && mixinStackPushed
      && obj->mixins != 0 && obj->mixinStack) {
    MixinStackPop(obj);
  }
  
  
  return result;
}

static XOTCLINLINE int
ObjDispatch (ClientData cd, 
	     Tcl_Interp* in, 
	     int objc, 
	     Tcl_Obj *CONST objv[], 
	     int noFilters) {
  XOTclObject *obj = (XOTclObject*)cd;
  char       *methodName = 0;
  XOTclClass  *mNclass;
  int result;

  if (objc < 2)
      return XOTclObjErrArgCnt(in, objv[0], "message ?args...?");

#ifdef DISPATCH_TRACE
  fprintf (stderr, "     DISPATCH:");
  {
    int i, j = 0;
    if (objc == 2){ j=2; } else j=3;
    for (i=0;i<j;i++) fprintf(stderr, " %s", ObjStr(objv[i]));
    fprintf(stderr, " objc == %d \n",objc);
  }
#endif

#ifdef STACK_TRACE
  XOTclStackTrace(in);
#endif

#ifdef CALLSTACK_TRACE
  XOTclCallStackTrace(in);
#endif

  /* try normal dispatch */
  if ((result = DoDispatch(cd, in, obj, objc, objv, noFilters)) != XOTCL_UNKNOWN) {
    return result;
  }

  /*
   * if methodName is a class/meta-class => create an instance
   */
  if (objc > 2) {
    char* flag = ObjStr(objv[2]);
    if (*flag != '-') {
      methodName = ObjStr(objv[1]);
      mNclass = XOTclGetClass(in, methodName);
      if (XOTclIsClass(in, (ClientData) mNclass)) {
	if (IsMetaClass(in, mNclass)) {
	  result = MakeClass(cd,in,objc,objv, mNclass);
	} else {
	  result = MakeObject(cd,in,objc,objv, mNclass);
	}
	if (result != XOTCL_UNKNOWN)
	  return result;
      }
    }
  }

  /*
   * back off and try unknown
   */
  {
      int oc = objc +1;
      DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov);

      ov[0] = objv[0];
      ov[1] = global_objects[UNKNOWN];
      memcpy(ov+2, objv+1, sizeof(Tcl_Obj *)*(objc-1));

      result=DoDispatch(cd, in, obj, oc, ov, noFilters);

      FREE_TCL_OBJECTS_ON_STACK(ov);

      if (result != XOTCL_UNKNOWN)
	return result;
  }

  /*
   * and if that fails too, error out
   */

  Tcl_ResetResult(in);

  if (methodName == 0 && objc > 0) 
    methodName = ObjStr(objv[1]);

  Tcl_AppendResult(in, ObjStr(objv[0]),
                   ": unable to dispatch method '", methodName, "'", 0);
  return TCL_ERROR;
}

static int
XOTclObjDispatch(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[]) {
  return ObjDispatch (cd, in, objc, objv, 0);
}

static int
XOTclDispatch(ClientData cd, Tcl_Interp* in, int argc, char* argv[]) {
  /* dummy */
  return TCL_ERROR;
}

/*
 * autoloading
 */

static void
AutoLoaderDP(ClientData cd) {
  ckfree((char*)cd);
}

static int
AutoLoader(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *CONST objv[]) {
  /*
   * cd is a script to evaluate; object context reconstructed from argv
   */
  CallStackContent *csc = CallStackGetFrame(in);
  fprintf(stderr,"Autoloader -- self %s\n",xobjName(csc->self));
  fprintf(stderr,"Autoloader -- c %s\n",className(csc->class));
  fprintf(stderr,"Autoloader -- p %s\n",csc->procName);

  if (Tcl_GlobalEval(in, (char*)cd) != TCL_OK) {
    Tcl_AppendResult(in, " (during XOTcl Autoloading in proc '",
		     ObjStr(objv[0]), "')", 0);
    return TCL_ERROR;
  }
  return TCL_OK;						
}

static int
MakeAuto(Tcl_Interp* in, char* loader, char* methodName, Namespace* ns) {
  ClientData cd;
  int l;
  if (!*loader)
    return 0;
  l = strlen(loader)+1;
  if (!(cd = (ClientData)ckalloc(l)))
    return 0;
  memcpy(cd,loader,l);
  if(!NSCreateObjCmd (in, methodName, AutoLoader, cd, AutoLoaderDP, ns)) return 0;
  return 1;
}

/*
 *  Proc-Creation
 */

static int
MakeProc(Namespace* ns, AssertionStore* aStore, 
	 Tcl_Interp* in, int objc, Tcl_Obj* objv[]) {
    int result, oc = objc;
    Tcl_CallFrame frame;
#ifdef AUTOVARS
    Tcl_Obj *oldBody;
    char *body;
    char *p;
    oldBody = objv[3];
    body = ObjStr(oldBody);

    objv[3] = Tcl_NewStringObj("", 0);
    INCR_REF_COUNT(objv[3]);

    if ((p = strstr(body,"self")) && p != body && *(p-1) != '[')
      Tcl_AppendStringsToObj(objv[3], "::set self [self]\n", NULL);
    if (strstr(body,"proc"))
      Tcl_AppendStringsToObj(objv[3], "::set proc [self proc]\n", NULL);
    if (strstr(body,"class"))
      Tcl_AppendStringsToObj(objv[3], "::set class [self class]\n", NULL);

    Tcl_AppendStringsToObj(objv[3], body, NULL);
#endif
    Tcl_PushCallFrame(in,&frame,(Tcl_Namespace*)ns,0);

    if (objc > 4) oc = 4;
    result = Tcl_ProcObjCmd(0, in, oc, objv) != TCL_OK;

    Tcl_PopCallFrame(in);

    if (objc == 6) 
      AssertionAddProc (in, ObjStr(objv[1]), aStore, objv[4], objv[5]);

#ifdef AUTOVARS
    DECR_REF_COUNT(objv[3]);
    objv[3] = oldBody;
#endif

    return result;
}

/*
 * List-Functions for Info
 */

static void
ListInfo(Tcl_Interp* in, char* key) {
  XOTclClass* cl = XOTclGetClass(in, key);
  if (!XOTclIsClass(in, (ClientData) cl))
    cl =0;
  Tcl_ResetResult(in);
  Tcl_AppendElement(in, "vars"); Tcl_AppendElement(in, "body");
  Tcl_AppendElement(in, "default"); Tcl_AppendElement(in, "args");
  Tcl_AppendElement(in, "procs"); Tcl_AppendElement(in, "commands");
  Tcl_AppendElement(in, "class"); Tcl_AppendElement(in, "children");
  Tcl_AppendElement(in, "parent"); Tcl_AppendElement(in, "invar");
  Tcl_AppendElement(in, "pre"); Tcl_AppendElement(in, "post");
  Tcl_AppendElement(in, "mixins"); Tcl_AppendElement(in, "calledproc");
  Tcl_AppendElement(in, "callingproc"); Tcl_AppendElement(in, "callingobject");
  Tcl_AppendElement(in, "regclass"); Tcl_AppendElement(in, "callingclass");
  Tcl_AppendElement(in, "metadata");  Tcl_AppendElement(in, "info");

  if (cl) {
    Tcl_AppendElement(in, "superclass"); Tcl_AppendElement(in, "subclass");
    Tcl_AppendElement(in, "heritage"); Tcl_AppendElement(in, "instances");
    Tcl_AppendElement(in, "instcommands"); Tcl_AppendElement(in, "instprocs");
    Tcl_AppendElement(in, "instdefault"); Tcl_AppendElement(in, "instbody");
    Tcl_AppendElement(in, "classchildren"); Tcl_AppendElement(in, "classparent");
    Tcl_AppendElement(in, "filters"); Tcl_AppendElement(in, "instinvar");
    Tcl_AppendElement(in, "instpre"); Tcl_AppendElement(in, "instpost");
  }
}

static void
ListKeys(Tcl_Interp* in, Tcl_HashTable* table, char* pattern) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
  Tcl_ResetResult(in);
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    char* key = Tcl_GetHashKey(table, hPtr);
    if (!pattern || Tcl_StringMatch(key, pattern))
      Tcl_AppendElement(in, key);
  }
}

static void
ListInstanceKeys(Tcl_Interp* in, Tcl_HashTable* table, char* pattern) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
  Tcl_ResetResult(in);
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    XOTclObject* obj = (XOTclObject*)Tcl_GetHashKey(table, hPtr);
    char* name = ObjStr(obj->cmdName); 
    if (!pattern || Tcl_StringMatch(name, pattern)) {
      Tcl_AppendElement(in, name);
    }
  }
}

static void
ListProcKeys(Tcl_Interp* in, Tcl_HashTable* table, char* pattern) {
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
  Tcl_ResetResult(in);
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    char* key = Tcl_GetHashKey(table, hPtr);
    Tcl_ObjCmdProc* proc = ((Command*)Tcl_GetHashValue(hPtr))->objProc;
    if (pattern && !Tcl_StringMatch(key, pattern)) continue;

    /*
     * also counts anything to be autoloaded as a proc
     */

    if (proc!=AutoLoader && proc!=ProcObjInterpId) continue;
    Tcl_AppendElement(in, key);
  }
}

static Proc*
FindProc(Tcl_HashTable* table, char* name) {
  Tcl_HashEntry* hPtr = table ? Tcl_FindHashEntry(table, name) : 0;
  if (hPtr) {
       Command* co = (Command*)Tcl_GetHashValue(hPtr);
       if (co->objProc == ProcObjInterpId)
	    return (Proc*) co->objClientData;
       else if (co->proc == ProcInterpId)
	    return (Proc*) co->clientData;
  }
  return 0;
}

static char *
FindAutoLoadBody(Tcl_HashTable* table, char* name) {
    Tcl_HashEntry* hPtr = table ? Tcl_FindHashEntry(table,name) : 0;
    if (hPtr) {
        Command* co = (Command*) Tcl_GetHashValue(hPtr);
        if (co->objProc == AutoLoader)
            return (char *) co->objClientData;
    }
    return (char*) NULL;
}

static int
ListProcArgs(Tcl_Interp* in, Tcl_HashTable* table, char* name) {
  Proc* proc = FindProc(table, name);
  if (proc) {
      CompiledLocal *args = proc->firstLocalPtr;
      Tcl_ResetResult(in);
      for (;args != NULL; args = args->nextPtr) {
          if TclIsVarArgument(args)
	      Tcl_AppendElement(in, args->name);

      }
      return TCL_OK;
  }
  return XOTclErrBadVal(in, "a tcl method name", name);
}

#if defined(VISUAL_CC)
#define XOTclInitStringRep(objPtr, bytePtr, len) \
    if ((len) == 0) { \
        (objPtr)->bytes  = '\0'; \
	(objPtr)->length = 0; \
    } else { \
	(objPtr)->bytes = (char *) ckalloc((unsigned) ((len) + 1)); \
	memcpy((VOID *) (objPtr)->bytes, (VOID *) (bytePtr), \
	        (unsigned) (len)); \
	(objPtr)->bytes[len] = '\0'; \
	(objPtr)->length = (len); \
    }
#endif

static int
ListProcDefault(Tcl_Interp* in, Tcl_HashTable* table,
                char* name, char* arg, char* var)
{

  /*
   * code snarfed from tcl info default
   */
  Proc* proc = FindProc(table, name);
  if (proc) {
      CompiledLocal *ap;
#if !defined(V81)
        Tcl_Obj *namePtr;
#endif

      for (ap = proc->firstLocalPtr; ap != 0; ap = ap->nextPtr) {
          if (!TclIsVarArgument(ap)) continue;
          if (strcmp(arg, ap->name) != 0) continue;

#if !defined(V81)
          XOTclNewObj(namePtr);
          INCR_REF_COUNT(namePtr);
#if defined(VISUAL_CC)
          XOTclInitStringRep(namePtr, var, strlen(var));
#else
          TclInitStringRep(namePtr, var, strlen(var));
#endif
          if (ap->defValuePtr != NULL) {
              if (Tcl_ObjSetVar2(in, namePtr, NULL, ap->defValuePtr,0)
                  == NULL) {
#else
          if (ap->defValuePtr != NULL) {
              if (Tcl_SetObjVar2(in, var, NULL, ap->defValuePtr,0)
                  == NULL) {
#endif /* V81 */
                defStoreError:
                  Tcl_ResetResult(in);
                  Tcl_AppendResult
                      (in, "couldn't store default value in variable '",
                       var, "'", (char *) 0);
#if !defined(V81)
                  DECR_REF_COUNT(namePtr);
#endif
                  return TCL_ERROR;
              }
              Tcl_SetObjResult(in,global_objects[ONE]);
          } else {
              Tcl_Obj *nullObjPtr, *resultObjPtr;
              XOTclNewObj(nullObjPtr);
              INCR_REF_COUNT(nullObjPtr);
              resultObjPtr =
#if defined(V81)
                  Tcl_SetObjVar2(in, var, NULL, nullObjPtr, 0);
#else
                  Tcl_ObjSetVar2(in, namePtr, NULL, nullObjPtr, 0);
#endif
              if (resultObjPtr == NULL) {
                  DECR_REF_COUNT(nullObjPtr);
                  goto defStoreError;
              }
              if (resultObjPtr == nullObjPtr) {
                  DECR_REF_COUNT(nullObjPtr);
              }
	      Tcl_SetObjResult(in,global_objects[ZERO]);
          }
#if !defined(V81)
          DECR_REF_COUNT(namePtr);
#endif
          return TCL_OK;
      }

      Tcl_ResetResult(in);
      Tcl_AppendResult(in, "procedure '", name,
                "' doesn't have an argument '", arg, "'", (char *) 0);
      return TCL_ERROR;
  }

  return XOTclErrBadVal(in, "a tcl method name", name);
}

static int
ListProcBody(Tcl_Interp* in, Tcl_HashTable* table, char* name) {
  Proc* proc = FindProc(table, name);

  if (proc) {
    Tcl_SetObjResult(in, proc->bodyPtr);
    return TCL_OK;
  } else {
    Tcl_SetResult(in, FindAutoLoadBody(table, name), TCL_VOLATILE);
    return TCL_OK;
  }
  return XOTclErrBadVal(in, "a tcl method name", name);
}

static void
ListObjChildren(Tcl_Interp* in, XOTclObject* obj, char* pattern) {
  Tcl_CallFrame frame;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = &obj->nsPtr->cmdTable ?
    Tcl_FirstHashEntry(&obj->nsPtr->cmdTable, &hSrch) :
    0;
  Tcl_ResetResult(in);
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    char* key = Tcl_GetHashKey(&obj->nsPtr->cmdTable, hPtr);
    if (!pattern || Tcl_StringMatch(key, pattern)){
	Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);
	if (XOTclGetObject(in,key))
	  Tcl_AppendElement(in, key);
	Tcl_PopCallFrame(in);
    }
  }
}

static void
ListClassChildren(Tcl_Interp* in, XOTclClass* cl, char* pattern) {
  XOTclObject* obj = &cl->object;
  Tcl_CallFrame frame;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr = &obj->nsPtr->cmdTable ?
    Tcl_FirstHashEntry(&obj->nsPtr->cmdTable, &hSrch) :
    0;
  Tcl_ResetResult(in);
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    char* key = Tcl_GetHashKey(&obj->nsPtr->cmdTable, hPtr);
    if (!pattern || Tcl_StringMatch(key, pattern)){
	Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);
	if (IsType(XOTclGetObject(in,key), InClass(in)) ? 1 : 0)
	  Tcl_AppendElement(in, key);
	Tcl_PopCallFrame(in);
    }
  }
}

XOTCLINLINE static void
ListParent (Tcl_Interp* in, Namespace* nsPtr) {
  Tcl_SetResult(in, nsPtr->parentPtr->fullName, TCL_VOLATILE);
}

static void
FindInCallStack(Tcl_Interp* in, Tcl_Obj* searchElt) {
  Interp *iPtr = (Interp *) in;
  int i, result = TCL_ERROR;
  Tcl_Obj* ov[2];
  ov[0] = Tcl_NewIntObj(1);
  ov[1] = searchElt;
  INCR_REF_COUNT(ov[0]);

  for (i=1; i < iPtr->numLevels && result != TCL_OK; i++) {
    Tcl_SetIntObj(ov[0],i);
    result = Tcl_UplevelObjCmd(NULL, in, 2, ov);
  }
  DECR_REF_COUNT(ov[0]);

  if (result != TCL_OK)
    Tcl_SetObjResult(in, global_objects[EMPTY]);
}

static XOTclClass *
FindCalledClass(Tcl_Interp* in, XOTclObject* obj) {
  char* methodName;
  XOTclClass      *cl;
  Command        *procInfo = NULL;
  if (!obj->filterinfo.chains && !obj->mixinStack)
    return GetSelfClass(in);

  if (obj->filterinfo.chains && obj->filterinfo.chains->filterPtr)
    methodName = ObjStr(obj->filterinfo.chains->calledProc);
  else if (obj->mixinStack && obj->mixinStack->mixinChainOn)
    methodName = GetSelfProc(in);
  else
    methodName = "";

  procInfo = FindMethod(methodName, obj->nsPtr);
  if (procInfo) {
    cl = 0;
  } else {
    cl = SearchCMethod(obj, obj->cl, NULL, methodName, &procInfo);
  }
  return cl;
}

/*
 * Next Primitive Handling
 */

static int
XOTclNextMethod(XOTclObject* obj, Tcl_Interp* in, XOTclClass *cl, char *method,
               int objc, Tcl_Obj * CONST objv[]);	
static XOTCLINLINE int
NextComputeStdArgs (Tcl_Interp* in, XOTclObject* obj, XOTclClass* cl,
		char *method, int objc, Tcl_Obj* CONST objv[]) {
  int result = XOTCL_STDARGS_COMPUTED;
  Tcl_Obj* args = 0;
  int ocArgs; Tcl_Obj** ovArgs;
  int ocLast=0; Tcl_Obj** ovLast;
  Interp* iPtr = (Interp*) in;
  int i;

  /*
   * retrieve the args of the current proc
   */
  if (iPtr->framePtr && iPtr->framePtr->nsPtr) {
    Tcl_HashTable* table = &iPtr->framePtr->nsPtr->cmdTable;
    Proc* proc = FindProc(table, method);
    if (proc) {
      CompiledLocal *compLocal = proc->firstLocalPtr;
      int oc = 0, ocMax = proc->numCompiledLocals;
      DEFINE_NEW_TCL_OBJECTS_ON_STACK(ocMax,ov);

      for (;compLocal != NULL; compLocal = compLocal->nextPtr) {
	if (TclIsVarArgument(compLocal))
	  ov[oc++] = Tcl_NewStringObj(compLocal->name, -1);	
      }

      for (i = 0; i < oc; i++) INCR_REF_COUNT(ov[i]);	
      args = Tcl_NewListObj(oc,ov);
      for (i = 0; i < oc; i++) DECR_REF_COUNT(ov[i]);

      FREE_TCL_OBJECTS_ON_STACK(ov);
    } else {
      return XOTclVarErrMsg(in,"StdArgs: Can't find args", (char*) NULL);
    }
    INCR_REF_COUNT(args);
  }
  /* now we break up the args; only if there are args we need to
     compute the std args, otherwise we can call without args */
  if ((Tcl_ListObjGetElements(in, args, &ocArgs, &ovArgs) == TCL_OK)
      && (ocArgs > 0)) {
    int i, oc = ocArgs + objc;

    Tcl_Obj *lastArgObj = ovArgs[ocArgs-1], *lastArgValue;
    char* lastArgStr = ObjStr(lastArgObj), *lastArgValueStr=NULL;

    /* if the last Argument is "args" -> we must split this too*/
    lastArgValue =
#if defined(V81)
      Tcl_GetObjVar2(in,ObjStr(lastArgObj),NULL,0);
#else
      Tcl_ObjGetVar2(in,lastArgObj,NULL,0);
#endif
    if (lastArgValue) {
      lastArgValueStr = ObjStr(lastArgValue);
    }
    if (isArgsString(lastArgStr) &&
        lastArgValueStr != NULL  && *lastArgValueStr != '\0' &&
        (Tcl_ListObjGetElements(in,lastArgValue,
                                &ocLast, &ovLast) == TCL_OK)
        && (ocLast > 0)) {
      oc += ocLast-1;
    }

    {
      int j;
      DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov);

      /* Now we have oc Args, spread over ocArgs and ocLast =>
	 all their values will be copied into the array ov */
      assert(objc <= oc);
      assert(ocArgs+objc-1 <= oc);

      i = objc;
      memcpy(ov,objv,sizeof(Tcl_Obj*)*objc);

      /* get the values of the ovArgs and put them to ov =>
	 the first ocArgs' of ov are filled */
      for (j = 0; j < ocArgs-1 ; j++, i++) {
#if defined(V81)
	ov[i] = Tcl_GetObjVar2(in,ObjStr(ovArgs[j]),NULL,0);
#else
	ov[i] = Tcl_ObjGetVar2(in,ovArgs[j],NULL,0);
#endif /* V81 */
      }

      /* it may be that the last value was empty
	 => we don't need the empty Arg -- this may only happen when the lastArg
         is "args" otherwise Tcl produces an error */
      if ((lastArgValueStr == NULL || *lastArgValueStr== '\0') &&
	  isArgsString(lastArgStr)) {
	oc = oc-1;
      } else {
	/* the last arg was not empty => it contains at least
	   one arg; if ocLast == 0 append this,
	   else: append all the last args in the for-statement */
	if (ocLast == 0) {
#if defined(V81)
	  ov[i] = Tcl_GetObjVar2(in,ObjStr(ovArgs[j]),NULL,0);
#else
	  ov[i] = Tcl_ObjGetVar2(in,ovArgs[j],NULL,0);
#endif /* V81 */

	}
	
	for (j=0; j<ocLast; j++, i++) {
	  ov[i] = ovLast[j];
	}
      }

      /* if there are std-arguments computed -> call next with these args*/
      if (oc > 1) {
	result = XOTclNextMethod (obj, in, cl, method, oc, ov);
      }
      FREE_TCL_OBJECTS_ON_STACK(ov);
    }
  }
  DECR_REF_COUNT(args);

  return result;
}

static int
XOTclNextMethod(XOTclObject* obj, Tcl_Interp* in, XOTclClass *cl, char *method,
               int objc, Tcl_Obj * CONST objv[]) {
  XOTclClass* ncl = 0; XOTclClass* mcl = 0;
  XOTclClasses* pl = 0;
  Command* procInfo;  Command* mprocInfo=0;
  Tcl_ObjCmdProc* proc = 0; Tcl_ObjCmdProc* mproc = 0;
  ClientData cp = 0;  ClientData mcp = 0;
  int result = TCL_OK;
  int mixinFound = 0;
  int isMixinEntry = 0;
  int isFilterEntry = 0;

  CallStack *cs = (CallStack *)XOTclClassesNS->clientData;
  CallStackContent *csc = cs->top;

  /* if no args are given => call itself with standardargs */
  if (objc == 1) {
    int stdArgsResult = NextComputeStdArgs (in, obj, cl, method, objc, objv);
    if (stdArgsResult != XOTCL_STDARGS_COMPUTED)
      return stdArgsResult;
  }

  /* if mixins are removed from within  a mixin method ->
     reset the callstack class, otherwise cl is unnecessary,
     since mixins do their own search
     we can reset to null, since mixins are always placed at the
     beginning of the precedence order
  */
  if (csc->isMixinEntry == 1) {
    cl = NULL;
  }

  /*
   * if we are already in the precedence ordering, then advance
   * past our last point; otherwise (if cl==0) begin from the start
   */

  pl = ComputePrecedence(obj->cl);
  while (pl && cl) {
    if (pl->cl == cl) cl = 0;
    pl = pl->next;
  }

  /*
   * search for a further class method
   */

  ncl = SearchCMethod((XOTclObject*) NULL, NULL, pl, method, &procInfo);

  if (procInfo) {
    proc = procInfo->objProc;
    cp   = procInfo->objClientData;
  }

  /*
   *  Next in filters
   */

  if (obj->filterinfo.chains && obj->filterinfo.chains->filterPtr != 0) {
    isFilterEntry = 1;
    /* if proc is already found the filter is specialized ->
       if not there are two possibilities: 
       1. The filter is on the meta-class and may be specialized there.
          Then cl is not set to 0 in the above 'while (pl && cl)' loop, 
	  because cl isn't on the precendence path
       2. if afterwards proc remains == 0 then we compute the next filter
    */
    if (proc == 0 && cl) {
      XOTclClasses* metaPl = ComputePrecedence(obj->cl->object.cl);
      while (metaPl && cl) {
	if (metaPl->cl == cl) cl = 0;
	metaPl = metaPl->next;
      }
      ncl = SearchCMethod((XOTclObject*) NULL, NULL, metaPl, method, &procInfo);

      if (procInfo) {
	proc = procInfo->objProc;
	cp   = procInfo->objClientData;
      }
    }
	
    if(proc == 0) {
      Command* lastFilter = obj->filterinfo.chains->filterPtr->cmdPtr;

      FilterComputeNext(obj);

      /* Deactivate the last filter for the next call */
      FilterActiveListDelete (obj, lastFilter);

      FilterSkipActiveFilters(obj);

      if (obj->filterinfo.chains->filterPtr != 0) {
	result = FilterDispatch (in, obj, objc, objv);
      } else {	
	char* filterMethod = ObjStr(obj->filterinfo.chains->calledProc);
	if (isNextString(filterMethod)) {
	  result = XOTclNextMethod (obj, in,
				    obj->filterinfo.chains->regClass,
				    ObjStr(obj->filterinfo.chains->callingProc),
				    objc, objv);	
	} else {
	  result = FilterCallActualProc (in, obj, objc, objv);
	}
      }

      /* Reactivate the last filter for it's post part */
      FilterActiveListPrepend (obj, lastFilter);

      return result;
    }
  }

  /*
   *  Next in Mixins
   */

  if (obj->mixins && obj->mixinStack && obj->mixinStack->currentCmdPtr) {
    mprocInfo = MixinSearchProc(in, obj, method, &mixinFound,
				&mcl, &mproc, &mcp);
  }

  if (mixinFound == 0) {
    if (obj->mixinStack && obj->mixinStack->mixinChainOn) {
      obj->mixinStack->mixinChainOn = 0;
      ncl = 0;

      /* if this is the last call in a "mixin init"-call, that means
	 OMixinMethod was invoked in "init" of the object's class and
	 "init" on the mixin was called, don't call the classes constrctor
	 again */
      if (obj->mixinStack->mixinInit)
	return TCL_OK;

      procInfo = FindMethod(method, obj->nsPtr);
      if (procInfo) {
	ncl = 0;
      } else {
	ncl = SearchCMethod(obj, obj->cl, NULL, method, &procInfo);
      }
      if (procInfo) {
	proc = procInfo->objProc;
	cp   = procInfo->objClientData;
      } else {
	proc = 0;
	cp   = 0;
      }
    }
  } else {
    ncl = mcl;
    procInfo = mprocInfo;
    proc = mproc;
    cp = mcp;
    isMixinEntry = 1;
  }

  /*
   * ordinary next
   */

  if (proc != 0) {
    int oc = objc;
    cp = (cp != 0) ? cp : (ClientData)obj;

    /* cut the flag, that no stdargs should be used, if it is there*/
    if ((objc > 1) && (strcmp(ObjStr (objv[1]),"--noArgs") == 0)) {
      oc = 1;
    }
    result = callProcCheck(cp, in, oc, objv, procInfo,
			   obj, ncl, method, isFilterEntry, isMixinEntry, 0);
  }
  return result;
}

static int
XOTclNextObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  CallStackContent *csc = CallStackGetFrame(in);

  if (!csc->self)
    return XOTclVarErrMsg(in, "next: can't find self", (char*) NULL);

  if (!csc->procName)
    return XOTclErrMsg(in,"next: no executing proc", TCL_STATIC);

  return XOTclNextMethod(csc->self, in, csc->class, csc->procName,
			      objc, objv);
}

/*
 * object creation
 */

static void
PrimitiveODestroy(ClientData cd) {
  XOTclObject* obj = (XOTclObject*)cd;
  Tcl_Interp* in;

  /*
   * check and latch against recurrent calls with obj->teardown
   */
  /*fprintf(stderr,"PrimitiveODestroy obj=%p\n",obj);*/

  if (!obj || !obj->teardown) return;
  in = obj->teardown; obj->teardown = 0;
  NSDeleteChildren(in, obj->nsPtr);
  /*
   * call and latch user destroy with obj->id if we haven't
   */
  /*fprintf(stderr,"PrimitiveODestroy after deleteChildren destroyCalled=%d\n",obj->destroyCalled);*/

  if (!obj->destroyCalled) {
    Tcl_Obj* args[2];
    char* cmdName = ObjStr(obj->cmdName);
    int result;
    args[0] = obj->cmdName;
    args[1] = global_objects[DESTROY];

    /*fprintf(stderr,"PrimitiveODestroy %p <%s> calling method=%d\n",
       obj,cmdName, obj->destroyCalled);*/
    if (cmdName && Tcl_FindCommand(in, cmdName, NULL, 0)) {
      result = XOTclObjDispatch(cd, in, 2, args);
      if (result != TCL_OK) {
	panic("Can't call destroy on %s. Probably neverending loop?", cmdName);
      }
    }
     /*fprintf(stderr,"PrimitiveODestroy %p <%s> calling method=%d DONE\n",
       obj,cmdName, obj->destroyCalled);*/
    obj->id = 0;
    obj->destroyCalled = 1;
  }
  (void)RemoveInstance(obj, obj->cl);
  AssertionRemoveStore (&obj->assertions);

#ifdef PROCCACHE
  ProcCacheRemoveCache(&obj->procCache);
  Tcl_DeleteHashTable(&obj->procCache);
#endif
  Tcl_DeleteHashTable(&obj->metaData);
  FilterActiveDeleteObj(obj);
  while (obj->filterinfo.chains != NULL)
    FilterChainInfoPop(obj);
  while (obj->mixinStack != NULL)
    MixinStackPop(obj);
  MixinRemoveList(obj);
  RemoveCmdList(obj->mixinOrder);
  obj->mixinOrder = 0;
  Tcl_DeleteNamespace((Tcl_Namespace *)obj->nsPtr);
  DECR_REF_COUNT(obj->cmdName);
  obj->cmdName = NULL;

  CallStackDestroyObject(in, obj);
  ckfree((char*)obj);
  return;
}

static void
PrimitiveOInit(void* mem, Tcl_Interp* in, char* name, XOTclClass* cl) {
  XOTclObject* obj = (XOTclObject*)mem;
  Tcl_DString cmd;
  Namespace *namespacePtr;

  obj->teardown = in;
  AddInstance(obj, cl);
  obj->type = InObject(in);
  obj->destroyCalled = 0;

#ifdef PROCCACHE
  Tcl_InitHashTable(&obj->procCache, TCL_STRING_KEYS);
#endif
  Tcl_InitHashTable(&obj->metaData, TCL_STRING_KEYS);

  obj->assertions.invariants = NULL;
  Tcl_InitHashTable(&obj->assertions.procs, TCL_STRING_KEYS);
  obj->checkoptions = CHECK_NONE;

  obj->filterinfo.chains = NULL;
  obj->filterinfo.doFilters = 0;
  obj->filterinfo.valid = 0;

  obj->mixins = 0;
  obj->mixinStack = 0;
  obj->mixinOrder = 0;

  namespacePtr = (Namespace*) Tcl_FindNamespace(in, name,
				   (Tcl_Namespace *) NULL, 0);
  if(namespacePtr) {
    NSDeleteNamespace(in, namespacePtr);
  }

  obj->nsPtr = (Namespace*) Tcl_CreateNamespace (in,name,
                 (ClientData)obj,(Tcl_NamespaceDeleteProc*)NULL);

  ALLOC_NAME_NS(&cmd, obj->nsPtr->fullName, "self");
  Tcl_CreateObjCommand(in, Tcl_DStringValue(&cmd), XOTclGetSelfObjCmd, 0, 0);
  Tcl_DStringFree(&cmd);

  ALLOC_NAME_NS(&cmd, obj->nsPtr->fullName, "next");
  Tcl_CreateObjCommand(in, Tcl_DStringValue(&cmd), XOTclNextObjCmd, 0, 0);
  Tcl_DStringFree(&cmd);

  obj->cmdName = NewXOTclObjectObj(obj);

  INCR_REF_COUNT(obj->cmdName);
  return;
}

static XOTclObject*
PrimitiveOCreate(Tcl_Interp* in, char* name, XOTclClass* cl) {
  XOTclObject* obj = (XOTclObject*)ckalloc(sizeof(XOTclObject));
  Tcl_DString fullName;
  char* fn;

  if (*name == ':' && *(name+1) == ':') {
    ALLOC_DSTRING (&fullName,name);
  } else {
    Tcl_Namespace* currNsPtr = Tcl_GetCurrentNamespace(in);
    if (currNsPtr != Tcl_GetGlobalNamespace(in) && 
	!XOTclGetObject(in,currNsPtr->fullName)
	&& !XOTclGetClass(in, NSCutXOTclClasses(currNsPtr->fullName))) {
      ALLOC_NAME_NS(&fullName, currNsPtr->fullName, name);
    } else {
      ALLOC_TOP_NS(&fullName, name);
    }
  }
  fn = Tcl_DStringValue(&fullName);
  if (obj != 0) {
    Tcl_CreateCommand(in, fn, XOTclDispatch,
                                (ClientData)obj, PrimitiveODestroy);
    obj->id = Tcl_CreateObjCommand(in, fn, XOTclObjDispatch,
                         (ClientData)obj, PrimitiveODestroy);
    PrimitiveOInit(obj, in, fn, cl);
  }
  Tcl_DStringFree(&fullName);
  return obj;
}

static void
PrimitiveCDestroy(ClientData cd) {
  XOTclClass* cl = (XOTclClass*)cd,
    *objectClass = XOTclGetClass(cl->object.teardown, "::Object");
  XOTclObject* obj = (XOTclObject*)cd;
  Tcl_HashSearch hSrch;
  Tcl_HashEntry* hPtr;
  Tcl_Interp* in;
  Namespace* saved;

  /*
   * check and latch against recurrent calls with obj->teardown
   */
  if (!obj || !obj->teardown) return;
  in = obj->teardown; obj->teardown = 0;

  /*
   * call and latch user destroy with obj->id if we haven't
   */

  if (obj->id) {
    Tcl_Obj *args[2];
    char* cmdName = ObjStr(obj->cmdName);
    args[0] = obj->cmdName;
    args[1] = global_objects[DESTROY];

    if (cmdName && Tcl_FindCommand(in, cmdName, NULL, 0))
      (void) XOTclObjDispatch(cd, in, 2, args);
    obj->id = 0;
    obj->destroyCalled = 1;
  }

  FilterRemoveList(cl);

  hPtr = &cl->instances ? Tcl_FirstHashEntry(&cl->instances, &hSrch) : 0;
  for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
    XOTclObject* inst = (XOTclObject*)Tcl_GetHashKey(&cl->instances, hPtr);
    if (inst && (inst != (XOTclObject*)cl) && inst->id) {
      /* we remove the instance to be sure the proc cache
	 entry of the obj is removed */
      (void)RemoveInstance(inst, obj->cl);
      AddInstance(inst, objectClass);
    }
  }

  Tcl_DeleteHashTable(&cl->instances);
#ifdef PROCCACHE
  ProcCacheRemoveAllDepending (cl);
#endif

  if (cl->objectdata) {
    Tcl_DeleteHashTable(cl->objectdata);
    ckfree((char*)cl->objectdata); cl->objectdata = 0;
  }

  NSDeleteChildren(in, cl->object.nsPtr);

  /*
   * flush all caches, unlink superclasses
   */

  FlushPrecedences(cl);
  while (cl->sub) {
    XOTclClass* subClass = cl->sub->cl;
    (void)RemoveSuper(subClass, cl);
    /* if there are no more super classes add the Object
       class as superclasses */
    if (subClass->super == 0)
      AddSuper(subClass, XOTclGetClass(in, "::Object"));
  }
  while (cl->super) (void)RemoveSuper(cl, cl->super->cl);

  /*
   * handoff the primitive teardown
   */

  saved = cl->nsPtr;
  obj->teardown = in;

  AssertionRemoveStore (&cl->assertions);

  PrimitiveODestroy(cd);

  Tcl_DeleteNamespace((Tcl_Namespace *)saved);

  Tcl_ResetResult(in);
  return;
}

static void
PrimitiveCInit(void* mem, Tcl_Interp* in, char* name, XOTclClass* class) {
  XOTclObject* obj = (XOTclObject *)mem;
  XOTclClass* cl = (XOTclClass*)mem;
  Tcl_CallFrame frame;
  Tcl_DString cmd;
  Namespace* namespacePtr;

  obj->type = InClass(in);
  cl->super = 0;
  cl->sub = 0;
  AddSuper(cl, InObject(in));
  cl->parent = InObject(in);
  cl->color = WHITE;
  cl->order = 0;

  cl->filters = 0;

  cl->parameterClass = 0;

  cl->assertions.invariants = NULL;
  Tcl_InitHashTable(&cl->assertions.procs, TCL_STRING_KEYS);

  Tcl_InitHashTable(&cl->instances, TCL_ONE_WORD_KEYS);
  cl->objectdata = 0;
  if (Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)XOTclClassesNS,
            /*isProcCallFrame*/ 0)!= TCL_OK) return;

  namespacePtr = (Namespace*) Tcl_FindNamespace(in, name,
				   (Tcl_Namespace *) NULL, 0);
  if(namespacePtr) {
    NSDeleteNamespace(in, namespacePtr);
  }

  cl->nsPtr = (Namespace*) Tcl_CreateNamespace (in,name,
                 (ClientData)cl,(Tcl_NamespaceDeleteProc*)NULL);

  Tcl_PopCallFrame(in);

  ALLOC_NAME_NS(&cmd, cl->nsPtr->fullName, "self");
  Tcl_CreateObjCommand(in, Tcl_DStringValue(&cmd), XOTclGetSelfObjCmd, 0, 0);
  Tcl_DStringFree(&cmd);

  ALLOC_NAME_NS(&cmd, cl->nsPtr->fullName, "next");
  Tcl_CreateObjCommand(in, Tcl_DStringValue(&cmd), XOTclNextObjCmd, 0, 0);
  Tcl_DStringFree(&cmd);

  return;
}

static XOTclClass*
PrimitiveCCreate(Tcl_Interp* in, char* name, XOTclClass* class){
  XOTclClass* cl = (XOTclClass*)ckalloc(sizeof(XOTclClass));

  if (cl != 0) {
    Tcl_DString fullName;
    char* nsName, *fn;
    XOTclObject* obj = (XOTclObject*)cl;

    if (*name == ':' && *(name+1) == ':') {
      nsName = name+2;
      ALLOC_DSTRING (&fullName,name);
      fn = Tcl_DStringValue(&fullName);
    } else {
      Tcl_Namespace * currNsPtr = Tcl_GetCurrentNamespace(in);
      if (currNsPtr != Tcl_GetGlobalNamespace(in) && 
	  !XOTclGetObject(in,currNsPtr->fullName)
	  && !XOTclGetClass(in,NSCutXOTclClasses(currNsPtr->fullName))) {
	ALLOC_NAME_NS(&fullName, currNsPtr->fullName, name);
      } else {
	ALLOC_TOP_NS(&fullName, name);
      }
      fn = Tcl_DStringValue(&fullName);
      nsName = fn+2;
    }

    obj->id = Tcl_CreateCommand(in, fn,
				XOTclDispatch,
				(ClientData)cl, PrimitiveCDestroy);
    Tcl_CreateObjCommand(in, fn, XOTclObjDispatch,
			 (ClientData)cl, PrimitiveCDestroy);

    PrimitiveOInit(obj, in, fn, class);
    PrimitiveCInit(cl, in, nsName, class);
  }

  return cl;
}

/*
 * object method implementations
 */

static int
XOTclOAllocMethod(ClientData cd, Tcl_Interp* in, 
		  int objc, Tcl_Obj *CONST objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  XOTclObject* newobj;
  int i;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) 
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "alloc <obj> ?args?");
  newobj = PrimitiveOCreate(in, ObjStr(objv[1]), cl);
  if (newobj == 0) return XOTclErrMsg(in,"Object alloc failed", TCL_STATIC);

  /*
   * this alloc doesn't process any extra args, so return them all
   */

  Tcl_ResetResult(in);
  for (i = 2; i < objc; i++)
       Tcl_AppendElement(in, ObjStr(objv[i]));

  return TCL_OK;
}

static int
XOTclODestroyMethod(ClientData cd, Tcl_Interp* in, 
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  Tcl_Command oid;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 1) return XOTclObjErrArgCnt(in, obj->cmdName, "destroy");
  /*
   * latch, and call delete command if not already in progress
   */
  oid = obj->id;
  obj->id = 0;
  obj->destroyCalled = 1;
  if ((obj->teardown != 0) && oid) {
    return (Tcl_DeleteCommandFromToken(in, oid)==0) ? TCL_OK : TCL_ERROR;
  }

  Tcl_ResetResult(in);
  return TCL_OK;
}

static int
XOTclOIsClassMethod(ClientData cd, Tcl_Interp* in, 
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 1) return XOTclObjErrArgCnt(in, obj->cmdName, "isclass");

  if (XOTclIsClass(in, cd)) {
    Tcl_SetObjResult(in, global_objects[ONE]);
  } else {
    Tcl_SetObjResult(in, global_objects[ZERO]);
  }
  return TCL_OK;
}

static int
XOTclOIsObjectMethod(ClientData cd, Tcl_Interp* in, 
		     int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2) return XOTclObjErrArgCnt(in, 
					       obj->cmdName, 
					       "isobject objName");

  if (XOTclGetObject(in, ObjStr(objv[1]))) {
    Tcl_SetObjResult(in, global_objects[ONE]);
  } else {
    Tcl_SetObjResult(in, global_objects[ZERO]);
  }
  return TCL_OK;
}

static int
XOTclOIsMetaClassMethod(ClientData cd, Tcl_Interp* in, 
			int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  XOTclClass* cl;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 1) return XOTclObjErrArgCnt(in, obj->cmdName, "ismetaclass");
  
  if (XOTclIsClass(in,cd)) {
    cl = XOTclGetClass(in, ObjStr(obj->cmdName));
    if (cl && IsMetaClass(in, cl)) {
      Tcl_SetObjResult(in, global_objects[ONE]);
      return TCL_OK;
    } 
  }
  Tcl_SetObjResult(in, global_objects[ZERO]);
  return TCL_OK;
}

static int
XOTclOClassMethod(ClientData cd, Tcl_Interp* in, 
		  int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  XOTclClass* cl;
  char *className;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2) return XOTclObjErrArgCnt(in, obj->cmdName, "class <class>");

  /*
   * allow a change to any class; type system enforces safety later
   */

  className = ObjStr(objv[1]);
  cl = XOTclGetClass(in, className);
  if (!cl)
    return XOTclErrBadVal(in, "a class", className);

  (void)RemoveInstance(obj, obj->cl);
  AddInstance(obj, cl);

  obj->mixinOrder = 0;

  return TCL_OK;
}

static int XOTclCInfoMethod(ClientData d, Tcl_Interp* h, int i, Tcl_Obj *CONST v[]);

static int
XOTclOInfoMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj* CONST objv[]) {
    XOTclObject* obj = XOTclAsObject(in, cd);
    char *cmd, *pattern;
    Interp *iPtr = (Interp *) in;
    CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;

    if (!obj) return XOTclObjErrType(in, objv[0], "Object");
    if (objc < 2) 
      return XOTclObjErrArgCnt(in,obj->cmdName,"info <opt> ?args?");

    cmd = ObjStr(objv[1]);
    pattern = (objc > 2) ? ObjStr(objv[2]) : 0;

    if (isClassString(cmd)) {
	if (objc > 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info class ?class?");
	if (objc == 2) {
	    Tcl_SetObjResult(in, obj->cl->object.cmdName);
	} else {
	    int result;
	    DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc,ov);

	    memcpy(ov, objv, sizeof(Tcl_Obj *)*objc);
	    ov[1] = Tcl_NewStringObj("superclass", 10);
	    INCR_REF_COUNT(ov[1]);
	    result = XOTclCInfoMethod((ClientData)obj->cl, in, objc, ov);
	    DECR_REF_COUNT(ov[1]);
	    FREE_TCL_OBJECTS_ON_STACK(ov);

	    return result;
	}
    } else if (!strcmp(cmd, "commands")) {
	if (objc > 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info commands ?pat?");
	ListKeys(in, &obj->nsPtr->cmdTable, pattern);
    } else if (!strcmp(cmd, "procs")) {
	if (objc > 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info procs ?pat?");
	ListProcKeys(in, &obj->nsPtr->cmdTable, pattern);
    } else if (isArgsString(cmd)) {
	if (objc != 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info args <proc>");
	return ListProcArgs(in,&obj->nsPtr->cmdTable, pattern);
    } else if (!strcmp(cmd, "default")) {
	if (objc != 5)
	    return XOTclObjErrArgCnt(in,
					  obj->cmdName,
					  "info default <proc> <arg> <var>");
	return ListProcDefault(in,  &obj->nsPtr->cmdTable, pattern,
			       ObjStr(objv[3]), ObjStr(objv[4]));
    } else if (!strcmp(cmd, "body")) {
	if (objc != 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info body <proc>");
	return ListProcBody(in, &obj->nsPtr->cmdTable, pattern);
    } else if (!strcmp(cmd, "vars")) {
	if (objc > 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info vars ?pat?");
	ListKeys(in, &obj->nsPtr->varTable, pattern);
    } else if (!strcmp(cmd, "callingobject")) {
	if (objc > 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info callingobject");
	FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
	if (obj->filterinfo.chains && obj->filterinfo.chains->callingObject &&
	    FilterIsFilterFrame(in, iPtr->varFramePtr)) {
	  Tcl_SetObjResult(in, obj->filterinfo.chains->callingObject);
	} else {
	  FindInCallStack(in, global_objects[SELF]);
	}
	FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);
    } else if (!strcmp(cmd, "calledproc")) {
	if (objc > 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info calledproc");
	if (!obj->filterinfo.chains || !obj->filterinfo.chains->calledProc)
	    return XOTclVarErrMsg(in,
				       "info calledproc called from outside of a filter",
				       (char*) NULL);
	Tcl_SetObjResult(in, obj->filterinfo.chains->calledProc);
    } else if (!strcmp(cmd, "calledclass")) {
	if (objc > 2)
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info calledclass");
	Tcl_ResetResult(in);
	Tcl_AppendResult(in, className(FindCalledClass(in,obj)), (char*) NULL);
    } else if (!strcmp(cmd, "callingproc")) {
        if (objc > 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info callingproc");
	FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
        if (obj->filterinfo.chains && obj->filterinfo.chains->callingProc &&
	    FilterIsFilterFrame(in, iPtr->varFramePtr)) {
	  Tcl_SetObjResult(in, obj->filterinfo.chains->callingProc);
	} else {
	  FindInCallStack(in, global_objects[SELF_PROC]);
	}
	FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);
    } else if (!strcmp(cmd, "regclass")) {
	if (objc > 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info regclass");
	if (!obj->filterinfo.chains || !obj->filterinfo.chains->regClass)
	  return XOTclVarErrMsg(in,
				     "info regclass called from outside of a filter",
				     (char*) NULL);
	Tcl_SetObjResult(in, obj->filterinfo.chains->regClass->object.cmdName);
    } else if (!strcmp(cmd, "callingclass")) {
	if (objc > 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info callingclass");
	FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
	if (obj->filterinfo.chains && obj->filterinfo.chains->callingClass &&
	    FilterIsFilterFrame(in, iPtr->varFramePtr)) {
	  Tcl_SetObjResult(in, obj->filterinfo.chains->callingClass);
	} else {
	  FindInCallStack(in, global_objects[SELF_CLASS]);
	}
	FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);
    } else if (!strcmp(cmd, "children")) {
	if (objc > 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info children ?pat?");
	ListObjChildren(in, obj, pattern);
    } else if (!strcmp(cmd, "parent")) {
	if (objc > 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info parent");
	ListParent(in, obj->nsPtr);
    } else if (!strcmp(cmd, "invar")) {
	if (objc != 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info invar");
	AssertionList(in, obj->assertions.invariants);
    } else if (!strcmp(cmd, "pre")) {
	ProcAssertion* procs;
	if (objc != 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info pre <proc>");
	procs = AssertionFindProcs (&obj->assertions, ObjStr(objv[2]));
	if (procs) AssertionList(in,procs->pre);
    } else if (!strcmp(cmd, "post")) {
	ProcAssertion* procs;
	if (objc != 3) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info post <proc>");
	procs = AssertionFindProcs (&obj->assertions, ObjStr(objv[2]));
	if (procs) AssertionList(in,procs->post);
    } else if (!strcmp(cmd, "check")) {
      	if (objc != 2) 
	  return XOTclObjErrArgCnt(in,obj->cmdName,"info check");
	AssertionListCheckOption(in, obj);
    } else if (!strcmp(cmd, "metadata")) {
      if (objc > 3)  
	return XOTclObjErrArgCnt(in,obj->cmdName,"info metadata ?pat?");
      ListKeys(in, &obj->metaData, pattern);
    } else if (!strcmp(cmd, "info")) {
      if (objc > 2) 
	return XOTclObjErrArgCnt(in,obj->cmdName,"info info");
      ListInfo(in, ObjStr(obj->cmdName));
    } else if (!strcmp(cmd, "mixins")) {
      if (objc > 3)
        return XOTclObjErrArgCnt(in, obj->cmdName, "info mixin ?class?");
      MixinInfo(in, obj, pattern);
    } else {
	return XOTclErrBadVal
          (in, "an info option (use 'info info' to list all info options)", cmd);
    }
    return TCL_OK;
}

static int
XOTclMakeFailure(Tcl_Interp* in, Tcl_Obj * CONST objv[], char *msg) {
  int i;

  Tcl_ResetResult(in);
  Tcl_AppendResult(in, msg, 0);
  for(i=0;i<=4;i++) Tcl_AppendResult(in, " ", ObjStr(objv[i]), 0 );
  for(;i<=6;i++) Tcl_AppendResult(in, " {", ObjStr(objv[i]), "}", 0 );
  Tcl_AppendResult(in, "\n", 0);
  return TCL_ERROR;
}

static int
XOTclOProcMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  char *argStr, *bdyStr, *name;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 4 && objc != 6)
      return XOTclObjErrArgCnt(in,
				    obj->cmdName,
				    "proc name args body ?preAssertion postAssertion?");

  /*
   * if the args list is "auto", the body is a script to load the proc
   */
  argStr = ObjStr(objv[2]);
  bdyStr = ObjStr(objv[3]);
  name = ObjStr(objv[1]);

#ifdef PROCCACHE
  ProcCacheORemoveMethod (obj, name);
#endif

  if (isAutoString(argStr)) {
      if (!MakeAuto(in, bdyStr, name, obj->nsPtr))
	  return XOTclMakeFailure(in, 
				       objv,
				       "Tcl_CreateObjCommand for autoloading failed in XOTclOProcMethod\n");
  } else if (*argStr == 0 && *bdyStr == 0) {
      AssertionRemoveProc (&obj->assertions,name);
      NSDeleteCmd(in,obj->nsPtr,name);
  } else
      MakeProc (obj->nsPtr, &obj->assertions, in, objc, (Tcl_Obj **) objv);

  return TCL_OK;
}

static int
XOTclOIncrMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
#if defined(PRE82)
  DEFINE_NEW_CHAR_ARRAY(objc, argv);
  Tcl_DString varName;
#else
  Tcl_CallFrame frame;
#endif
  int result;

  if (!obj) 
    return (XOTclObjErrType(in, objv[0], "Object"));
  if (objc < 2) 
    return (XOTclObjErrArgCnt(in, obj->cmdName, "incr <varName> ?increment?"));

#if !defined(PRE82)
  Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);
  result = Tcl_IncrObjCmd(cd, in, objc, objv);
  Tcl_PopCallFrame(in);
#else
  argv[0] = global_strings[INCR];
  
  ALLOC_NAME_NS (&varName, obj->nsPtr->fullName, ObjStr(objv[1]));
  argv[1] = Tcl_DStringValue(&varName);

  if (objc == 3) 
    argv[2] = ObjStr(objv[2]);

  result = Tcl_IncrCmd(cd, in, objc, argv); 
  Tcl_DStringFree (&varName);
  FREE_CHAR_ARRAY(argv);
#endif 
  return result; 
}   
Tcl_Obj*
XOTclOSetInstVar(XOTclObject* obj,Tcl_Interp* in,
		 Tcl_Obj* name,Tcl_Obj* value,int flgs) {
  return XOTclOSetInstVar2(obj, in, name, (Tcl_Obj *) NULL, value, flgs);
}

Tcl_Obj*
XOTclOGetInstVar(XOTclObject* obj, Tcl_Interp* in, Tcl_Obj* name, int flgs) {
    return XOTclOGetInstVar2(obj,in,name,(Tcl_Obj *)NULL,flgs);
}

int
XOTclUnsetInstVar(XOTclObject* obj, Tcl_Interp* in, char* name, int flgs) {
    return XOTclUnsetInstVar2 (obj, in, name,(char*)NULL, flgs);
}

static int
XOTclOSetMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  Tcl_Obj *result;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");

  switch (objc) {
  case 2:	
    result = XOTclOGetInstVar2(obj, in, objv[1], NULL, TCL_LEAVE_ERR_MSG);
    break;
  case 3:
    {
      CallFrame frame, *oldFptr = ((Interp *)in)->varFramePtr;

      frame.nsPtr = obj->nsPtr;
      frame.isProcCallFrame = 0;

      ((Interp *)in)->varFramePtr = &frame;

#if defined(V81)
      result = Tcl_SetObjVar2(in, ObjStr(objv[1]), NULL, objv[2],
			      TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1|
			      TCL_NAMESPACE_ONLY);
#else
      result = Tcl_ObjSetVar2(in, objv[1], NULL, objv[2],
			      TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1|
			      TCL_NAMESPACE_ONLY);
#endif
      ((Interp *)in)->varFramePtr = oldFptr;
      break;
    }
  default:
    return XOTclObjErrArgCnt(in, obj->cmdName, "set var ?value?");
  }

  if (result) {
    Tcl_SetObjResult(in, result);
    return TCL_OK;
  } else {
    return TCL_ERROR;
  }
}

static int
XOTclOUnsetMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  int result = TCL_ERROR;
  int i;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2) return XOTclObjErrArgCnt(in, obj->cmdName, "unset ?vars?");
  for (i=1; i<objc; i++) {
    Tcl_CallFrame frame;
    Tcl_PushCallFrame(in, &frame, (Tcl_Namespace*)obj->nsPtr,0);

    result =  Tcl_UnsetVar2(in, ObjStr(objv[i]), (char *) NULL,
			    (TCL_LEAVE_ERR_MSG | TCL_PARSE_PART1));
    Tcl_PopCallFrame(in);
    if (result != TCL_OK) break;
  }
  return result;
}

static XOTCLINLINE int
XOTclInstVarDS(XOTclObject* obj, Tcl_Interp* in, Tcl_DString* ds, char *destName) {
  int result;
  Tcl_Obj* ov[2];
  char* name = Tcl_DStringValue(ds); 
  int filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;

  if (obj  && obj->filterinfo.chains) { 
    filterFrameCutOff = 1;
    FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
   }

  ov[1] =  Tcl_NewStringObj(name,
                            Tcl_DStringLength(ds));
  INCR_REF_COUNT(ov[1]);
  result = Tcl_VariableObjCmd ((ClientData) NULL,in, 2, ov);
  /* UpVar for compatibility to OTcl (alias names)*/
  if ((result == TCL_OK) && (destName !=NULL))
    result = Tcl_UpVar(in, "0", name, destName, 0);

  if (filterFrameCutOff)
    FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);

  DECR_REF_COUNT(ov[1]);
  return result;
}

static int
XOTclOInstVarMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject  *obj   = XOTclAsObject(in, cd);
  Namespace    *nsPtr;
  Tcl_Obj      **ov;
  int          i, oc, result = TCL_OK;
  Interp       *iPtr = (Interp *) in;
  int filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;

  if (!obj)
    return XOTclObjErrType(in, objv[0], "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, obj->cmdName, "instvar ?vars?");

  if (obj  && obj->filterinfo.chains) { 
    filterFrameCutOff = 1;
    FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
   }

  nsPtr = iPtr->varFramePtr->nsPtr;
  iPtr->varFramePtr->nsPtr = obj->nsPtr;

  for (i=1; i<objc; i++) {
    if ((result = Tcl_ListObjGetElements(in, objv[i], &oc, &ov)) == TCL_OK) {
       result = Tcl_VariableObjCmd((ClientData)NULL, in, 2, &ov[-1]);
       if (result != TCL_OK)
	 break;
       if (oc == 2) {
	 result = Tcl_UpVar(in, "0", ObjStr(ov[0]), ObjStr(ov[1]), 0);
       }
       if (result != TCL_OK)
	 break;
    } else
      break;
  }

  iPtr->varFramePtr->nsPtr=nsPtr;

  if (filterFrameCutOff)
    FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);

  return result;
}

static int
XOTclOMoveMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]){
  XOTclObject* obj = XOTclAsObject(in, cd);
  char *self, *newName, *newCmdName;
  Tcl_DString newCmdNameDS;
  int result;
  Namespace* oldNS = obj->nsPtr;

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,obj->cmdName,"move newName");

  self = ObjStr(obj->cmdName);
  newName = ObjStr(objv[1]);

  if ((newName == NULL) || (*newName == '\0')) {
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(2,ov);
    ov[0] = obj->cmdName;
    ov[1] = global_objects[DESTROY];
    obj->destroyCalled = 1;

    result = ObjDispatch((ClientData) obj, in, 2, ov, 1);
    FREE_TCL_OBJECTS_ON_STACK(ov);

    return result;
  }
  
  Tcl_DStringInit(&newCmdNameDS);
  Tcl_DStringAppend(&newCmdNameDS, "::", 2);
  while (newName && *newName == ':')
    newName++;
  Tcl_DStringAppend(&newCmdNameDS, newName, -1);
  newCmdName = Tcl_DStringValue(&newCmdNameDS);

  if ((result=NSRenameCommand(in, newCmdName, self)) != TCL_OK) 
    return TCL_ERROR;

  obj->nsPtr = NSCreateAndCopy (in, obj->nsPtr, newCmdName, (ClientData)obj);
  Tcl_DeleteNamespace((Tcl_Namespace *) oldNS);

  NSCorrectCmdName(in, obj);

  Tcl_DStringFree (&newCmdNameDS);

  return result;
}

static int
XOTclOInvariantsMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);

  if (!obj) return XOTclObjErrType(in, objv[0], "Object");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,obj->cmdName,"invar <invariantList>");

  AssertionDeleteList (obj->assertions.invariants);
  obj->assertions.invariants = AssertionNewList (in, objv[1]);

  return TCL_OK;
}

static int
XOTclOAutonameMethod(ClientData cd, Tcl_Interp* in, 
		     int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  Tcl_Obj* autoname;

  if (!obj)
    return XOTclObjErrType(in, obj->cmdName, "Object");
  if ((objc != 2 && objc != 3) ||
      (objc == 3 && strcmp(ObjStr(objv[1]), "-instance") != 0))
    return XOTclObjErrArgCnt(in,obj->cmdName,
				  "autoname [-instance] name");

  if (objc == 3) {
    char* a = ObjStr(objv[objc-1]);
    if (isupper(UCHAR(*a)))
      *a = (char) tolower(UCHAR(*a));
  }

  autoname = AutonameIncr(in, objv[objc-1], obj);

  Tcl_SetObjResult(in, autoname);

  return TCL_OK;
}

static int
XOTclOCheckMethod (ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  int ocArgs; Tcl_Obj** ovArgs;
  int i, result=TCL_OK;

  if (!obj) return XOTclObjErrType(in, obj->cmdName, "Object");

  if (objc != 2)
    return XOTclObjErrArgCnt(in,obj->cmdName,
				  "check (?all? ?pre? ?post? ?invar? ?instinvar?)");
  obj->checkoptions = CHECK_NONE;
  if ((Tcl_ListObjGetElements(in, objv[1], &ocArgs, &ovArgs) == TCL_OK)
      && (ocArgs > 0)) {
    for (i = 0; i < ocArgs; i++) {
      char* option = ObjStr(ovArgs[i]);
      result = TCL_ERROR;
      if (option != 0 && *option != '\0') {
	switch (*option) {
	case 'i':
	  if (strcmp(option, "instinvar") == 0){
	    setOption(1,obj->checkoptions,CHECK_CLINVAR);
	    result = TCL_OK;
	  } else if (strcmp(option, "invar") == 0) {
	    setOption(1,obj->checkoptions,CHECK_OBJINVAR);
	    result = TCL_OK;
	  }
	  break;
	case 'p':
	  if (strcmp(option, "pre") == 0){
	    setOption(1,obj->checkoptions,CHECK_PRE);
	    result = TCL_OK;
	  } else if (strcmp(option, "post") == 0) {
	    setOption(1,obj->checkoptions,CHECK_POST);
	    result = TCL_OK;
	  }
	  break;
	case 'a':
	  if (strcmp(option, "all") == 0){
	    setOption(1,obj->checkoptions,CHECK_ALL);
	    result = TCL_OK;
	  }
	  break;	
	}
      }
    }
  }
  if (result == TCL_ERROR) {
    return XOTclVarErrMsg(in,"Check configuration option unknown: ",
			       ObjStr(obj->cmdName),
			       ", valid: all pre post invar instinvar", 
			       (char*) NULL);
  }

  Tcl_ResetResult(in);
  return result;
}

static int
XOTclOMetaDataMethod (ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  XOTclClass* cl = IsType((XOTclObject*)cd, InClass(in)) ? (XOTclClass*)cd : 0;
  char* option; int add = -1;
  int result = TCL_OK;
  int oc; Tcl_Obj** ov; int i;

  if (!obj) return XOTclObjErrType(in, obj->cmdName, "Object");
  if (objc < 2)
    return XOTclObjErrArgCnt(in,obj->cmdName,
				  "metadata ?(add|remove)? metaDataList");

  option = ObjStr(objv[1]);
  switch (*option) {
  case 'a':
    if (strcmp(option,"add") == 0) add = 1; break;
  case 'r':
    if (strcmp(option,"remove") == 0) add = 0; break;
  }
  if (add == -1) {
    if (objc == 2) {
      Tcl_HashEntry* hPtr = Tcl_FindHashEntry(&obj->metaData, option);
      if (hPtr) {
	Tcl_Obj* entry = (Tcl_Obj *) Tcl_GetHashValue(hPtr);
	if (entry) {
	  Tcl_SetObjResult(in, (Tcl_Obj *) Tcl_GetHashValue(hPtr));
	} else {
	  Tcl_ResetResult(in);
	}
	return TCL_OK;	
      }
    }
    if (objc == 3) {
      Tcl_HashEntry* hPtr = Tcl_FindHashEntry(&obj->metaData, option);
      if (!hPtr) {
	int nw;
	if (MetaDataInheritance(obj, option)) {
	  hPtr = Tcl_CreateHashEntry(&obj->metaData, option, &nw);
	  if (!nw) 
	    return XOTclVarErrMsg(in, 
				       "MetaData: Cant create MetaData Entry: ",
				       option, (char*) NULL);
	}
      }	
      if (hPtr) {
	Tcl_Obj* entry = (Tcl_Obj *) Tcl_GetHashValue(hPtr);
	if (entry)
	  DECR_REF_COUNT(entry);
	INCR_REF_COUNT(objv[2]);
	Tcl_SetHashValue(hPtr, (ClientData) objv[2]);	
	return TCL_OK;
      }
    }
    return XOTclVarErrMsg(in,"MetaData: Unknown option;  given Option: ",
			  option, (char*) NULL);
  }

  if (Tcl_ListObjGetElements(in, objv[2], &oc, &ov) == TCL_OK) {
    for (i = 0; i < oc; i ++) {
      char* value = ObjStr (ov[i]);
      if (add) {	
	int nw;
	Tcl_HashEntry* hPtr = Tcl_FindHashEntry(&obj->metaData, value);
	if (hPtr) 
	  return XOTclVarErrMsg(in, 
				     "Can't add MetaData, MetaData exists: ",
				     value,
				     (char*) NULL);	
	hPtr = Tcl_CreateHashEntry(&obj->metaData, value, &nw);
	if (!nw) 
	  return XOTclVarErrMsg(in, 
				     "MetaData: Cant create MetaData Entry: ",
				     value,(char*) NULL);
      } else {
	Tcl_HashEntry* hPtr = Tcl_FindHashEntry(&obj->metaData, value);
	if (hPtr) {		
	  Tcl_DeleteHashEntry(hPtr);
	  if (cl) {
	    MetaDataRemoveDepending(cl, value);
	  }
	}
      }
    }
  }

  return result;
}

static int
XOTclOCopyMethod (ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject *newobj, *obj = XOTclAsObject(in, cd);

  if (!obj) return XOTclObjErrType(in, obj->cmdName, "Object");
  if (objc != 2) 
    return XOTclObjErrArgCnt(in, obj->cmdName, "copy <destination>");

  newobj = PrimitiveOCreate(in, ObjStr(objv[1]), obj->cl);
  if (newobj == 0) 
    return XOTclErrMsg(in,"Object copy failed", TCL_STATIC);
  NSCopy(in, obj->nsPtr, ObjStr(objv[1]), cd);
  return TCL_OK;
}

static int
XOTclOMixinMethod (ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  int oc; Tcl_Obj** ov;
  XOTclObject* obj = XOTclAsObject(in, cd);
  int i, result = TCL_OK;
  char* p = GetSelfProc(in);

  if (!obj) return XOTclObjErrType(in, obj->cmdName, "Object");
  if (objc != 2) 
    return XOTclObjErrArgCnt(in, obj->cmdName, "mixin <classes>");

  if (Tcl_ListObjGetElements(in,objv[1], &oc, &ov)!= TCL_OK) return TCL_ERROR;

  MixinRemoveList(obj);

  for (i = 0; i < oc; i++) {
    char* av = ObjStr(ov[i]);
    XOTclClass* mixin = XOTclGetClass(in, av);
    if (!mixin)
      return XOTclErrBadVal(in, "a list of classes", ObjStr(objv[1]));
    if (!XOTclIsClass(in, (ClientData) mixin)) 
      return XOTclVarErrMsg(in, "Mixin for ", ObjStr(obj->cmdName), 
				 " must be class, got: ", av, (char*)0);
    MixinAdd(in, obj, mixin);
  }
 
  MixinComputeOrder(in, obj);
  /*
  {
    CmdList* p;
    p = obj->mixinOrder;
    fprintf(stderr, "Mixin Order: ");
    while (p) {
      fprintf(stderr, " %s,", GetClassFromCmdPtr(in, p->cmdPtr)->object.nsPtr->fullName);
      p=p->next;
    }
    fprintf(stderr, "\n");
  }
  */

  if (p && isInitString(p)) { 
    XOTclClass     *mcl=0;
    Tcl_ObjCmdProc *mproc = 0;
    Command        *mProcInfo =0;
    ClientData     mcp;
    int isDestroyed;

    MixinStackPush(obj);
    mProcInfo = MixinSearchProc(in, obj, "init",
				&obj->mixinStack->mixinChainOn,
				&mcl, &mproc, &mcp);
    if (mproc) {
      Tcl_Obj** ov = &global_objects[INIT];
      /* set flag to prevent calling the class constructor twice */
      obj->mixinStack->mixinInit = 1;
      result = callProcCheck(mcp, in, 1, ov, mProcInfo,
			     obj, mcl, "init", 0, 1 /*isMixinEntry*/, &isDestroyed);
    }

    if (!isDestroyed && obj->mixins != 0 && obj->mixinStack)
      MixinStackPop(obj);
  }
 
  return result;
}

static int
XOTclOProcSearchMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclObject* obj = XOTclAsObject(in, cd);
  XOTclClass* cl=0;
  Command *cmdPtr;
  char *simpleName, *fullName, *methodName;
  Tcl_DString fullNameDS;

  if (!obj) return XOTclObjErrType(in, obj->cmdName, "Object");
  if (objc < 2) return XOTclObjErrArgCnt(in,cl->object.cmdName,
    "procsearch name");

  Tcl_ResetResult(in);

  methodName = ObjStr(objv[1]);

  cmdPtr = FindMethod(methodName, obj->nsPtr);
  if (!cmdPtr) {
    cl = SearchCMethod(obj, obj->cl, NULL, methodName, &cmdPtr);
  }
  if (cmdPtr) {
    simpleName = Tcl_GetCommandName (in, (Tcl_Command)cmdPtr);
    ALLOC_NAME_NS (&fullNameDS, cmdPtr->nsPtr->fullName, simpleName);
    fullName = Tcl_DStringValue(&fullNameDS);

    Tcl_AppendElement(in, fullName);

    Tcl_DStringFree(&fullNameDS);
  }
  return TCL_OK;
}

/*
 * class method implementations
 */

static int
XOTclCAllocMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  XOTclClass* newcl;
  int i;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2)
      return XOTclObjErrArgCnt(in, cl->object.cmdName, "alloc <cl> ?args?");

  newcl = PrimitiveCCreate(in, ObjStr(objv[1]), cl);
  if (newcl == 0) return XOTclErrMsg(in,"Class alloc failed", TCL_STATIC);

  /*
   * this alloc doesn't process any extra args, so return them all
   */

  Tcl_ResetResult(in);
  for (i = 2; i < objc; i++) Tcl_AppendElement(in, ObjStr(objv[i]));

  return TCL_OK;
}

static int
XOTclCCreateMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  XOTclObject* callingobj = XOTclAsObject(in, cd);
  XOTclObject* obj;
  Command* procInfo;
  Tcl_ObjCmdProc* proc = 0;
  ClientData cp = 0;
  XOTclClasses* pl;
  Tcl_Obj *ov0;

  int i, result, initCalled = 0;
  char *objName;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2)
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "create <obj> ?args?");
  for (pl = ComputePrecedence(cl); pl != 0; pl = pl->next) {
    procInfo = FindMethod("alloc",pl->cl->object.nsPtr);
    if (procInfo) {
      proc = procInfo->objProc;
      cp   = procInfo->objClientData;
      break;
    }
  }
  if (!pl)
    return XOTclErrMsg(in, "no reachable alloc", TCL_STATIC);
  if (!proc)
    return XOTclErrMsg(in, "create can't reach alloc", TCL_STATIC);
  ov0 = objv[0];
  objv[0] = global_objects[ALLOC];
  cp = (cp != 0) ? cp : (ClientData)pl->cl;

  result = callProc(cp, in, objc, objv, procInfo,
		    callingobj, cl, "alloc", 0, 0);
  objv[0] = ov0;

  if (result != TCL_OK)
    return result;

  objName = ObjStr(objv[1]);
  obj = XOTclGetObject(in, objName);

  if (obj == 0)
    return XOTclErrMsg(in, "couldn't find result of alloc", TCL_STATIC);

  (void)RemoveInstance(obj, obj->cl);
  AddInstance(obj, cl);

  /*
   *  Search for default values for vars on superclasses
   */
  result = SearchDefaultValues(in, obj, cl);
  if (result != TCL_OK)
    return result;

  /*
   * Check Args for '-' -- call these args as methods
   */

  for (i=2; i < objc; i++) {
    char *flag = ObjStr(objv[i]);
    if (*flag == '-') {
      int j, nrArgs;
      for (j = i+1, nrArgs = 0; j < objc; j++, nrArgs++) {
	char* arg = ObjStr(objv[j]);
	/*
	 * digits are an exception from the '-'-evaluation
	 * they are always treated as arguments (to let -1 etc. pass)
	 */
	if (*arg == '-' && !isDigitParam(arg+1)) {
	  break;
	}
      }

      {
	Tcl_Obj *tmp;
	int k;
	int ac = 2+nrArgs;
	DEFINE_NEW_TCL_OBJECTS_ON_STACK(ac,args);
	char* methodName = flag+1;

	args[0] = obj->cmdName;
	XOTclNewObj(tmp);
	INCR_REF_COUNT(tmp);
	args[1] = tmp;
        Tcl_SetStringObj(args[1], methodName, -1);

	for (k = 0; k < nrArgs; k++) {
	  char* argStr = ObjStr(objv[i+1+k]);
	  if (*argStr == '\\') 
	    argStr++;
	  args[2+k] = Tcl_NewStringObj(argStr,-1); 
	  INCR_REF_COUNT(args[2+k]);
	}

	if (isInitString(methodName))
	  initCalled = 1;

	result = XOTclObjDispatch((ClientData) obj, in, ac, args);

	for (k = 0; k < nrArgs; k++) {
	  DECR_REF_COUNT(args[2+k]);
	}

	FREE_TCL_OBJECTS_ON_STACK(args);

	if (result != TCL_OK) {
	  Tcl_AppendResult(in, " during '", ObjStr(args[0]), " ",
			   flag, "'", 0);
	  DECR_REF_COUNT(tmp);
	  return result;
	}
	DECR_REF_COUNT(tmp);
      }

      i += nrArgs;
    }
  }

  /*
   *  Call the user-defined constructor init, if it exists
   *  in the precedence order
   *  (only if is not called during the '-' args)
   */

  if (!initCalled) {
    Tcl_ResetResult(in);
    result =TCL_OK;

    procInfo = FindMethod("init",obj->nsPtr);
    if (procInfo) {
      cl = 0;
    } else {
      cl = SearchCMethod(obj, obj->cl, NULL, "init", &procInfo);
    }
    if (procInfo) {
      DEFINE_NEW_TCL_OBJECTS_ON_STACK(objc,ov);

      ov[0] = obj->cmdName;
      ov[1] = global_objects[INIT];
      memcpy(ov+2, objv+2, sizeof(Tcl_Obj *)*(objc-2));

      result = ObjDispatch((ClientData) obj, in, objc, ov, 0);
      FREE_TCL_OBJECTS_ON_STACK(ov);
    }
  }

  if(result == TCL_OK) {
    Tcl_SetObjResult(in, objv[1]);
  }

  return result;
}

static int
XOTclCSuperClassMethod(ClientData cd, Tcl_Interp* in,
                      int objc, Tcl_Obj *objv[])
{
  XOTclClass* cl = XOTclAsClass(in, cd);
  XOTclClasses* osl = 0;
  int oc; Tcl_Obj** ov;
  XOTclClass** scl = 0;
  int reversed = 0;
  int i, j;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2) 
    return XOTclObjErrArgCnt(in, cl->object.cmdName, "superclass <classes>");

  if (Tcl_ListObjGetElements(in,objv[1], &oc, &ov)!= TCL_OK) return TCL_ERROR;

#ifdef PROCCACHE
  ProcCacheRemoveAllDepending (cl);
#endif

  scl = (XOTclClass**)ckalloc(oc*sizeof(XOTclClass*));
  for (i = 0; i < oc; i++) {
    char* av = ObjStr(ov[i]);
    scl[i] = XOTclGetClass(in, av);
    if (!scl[i]) {

      /*
       * try to force autoloading if we can't resolve a class name
       */

      int loaded = 0;
      char* args = (char*)ckalloc(strlen("auto_load ")+strlen(av)+1);
      (void)strcpy(args, "auto_load ");
      (void) strcat(args, av);
      if (Tcl_GlobalEval(in, args) == TCL_OK) {
        scl[i] = XOTclGetClass(in, av);
        loaded = (scl[i] != 0);
      }
      ckfree((char*)args);
      if (!loaded) {
        ckfree((char*)scl);
        return XOTclErrBadVal(in, "a list of classes", ObjStr(objv[1]));
      }
    }
  }

  /*
   * check that superclasses don't precede their classes
   */

  for (i = 0; i < oc; i++) {
    if (reversed != 0) break;
    for (j = i+1; j < oc; j++) {
      XOTclClasses* dl = ComputePrecedence(scl[j]);
      if (reversed != 0) break;
      while (dl != 0) {
        if (dl->cl == scl[i]) break;
        dl = dl->next;
      }
      if (dl != 0) reversed = 1;
    }
  }

  if (reversed != 0) {
    ckfree((char*)scl);
    return XOTclErrBadVal(in, "classes in dependence order", ObjStr(objv[1]));
  }

  while (cl->super != 0) {

    /*
     * build up an old superclass list in case we need to revert
     */

    XOTclClass* sc = cl->super->cl;
    XOTclClasses* l = osl;
    osl = (XOTclClasses*)ckalloc(sizeof(XOTclClasses));
    osl->cl = sc;
    osl->next = l;
    (void)RemoveSuper(cl, cl->super->cl);
  }
  for (i = 0; i < oc; i++)
    AddSuper(cl, scl[i]);
  ckfree((char*)scl);
  FlushPrecedences(cl);

  if (!ComputePrecedence(cl)) {

    /*
     * cycle in the superclass graph, backtrack
     */

    XOTclClasses* l;
    while (cl->super != 0) (void)RemoveSuper(cl, cl->super->cl);
    for (l = osl; l != 0; l = l->next) AddSuper(cl, l->cl);
    RC(osl);
    return XOTclErrBadVal(in, "a cycle-free graph", ObjStr(objv[1]));
  }
  RC(osl);

  /* if there are no more super classes add the Object
     class as superclasses */
  if (cl->super == 0)
    AddSuper(cl, XOTclGetClass(in, "::Object"));

  /* invalidate all mixin order of instances of this 
     and of all depended classes */
  MixinInvalidateObjOrders(cl);

  Tcl_ResetResult(in);
  return TCL_OK;
}

static int
XOTclCInfoMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj * CONST objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  char *pattern, *cmd;

  if (!cl) 
    return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) 
    return XOTclObjErrArgCnt(in,cl->object.cmdName,"info <opt> ?args?");

  cmd = ObjStr(objv[1]);
  pattern = (objc > 2) ? ObjStr(objv[2]) : 0;

  if (!strcmp(cmd, "superclass")) {
    if (objc > 3)
        return XOTclObjErrArgCnt(in,
				      cl->object.cmdName,
				      "info superclass ?class?");
    if (objc == 2) {
      XOTclClasses* sl = cl->super;
      XOTclClasses* sc = 0;

      /*
       * reverse the list to obtain presentation order
       */

      Tcl_ResetResult(in);
      while (sc != sl) {
        XOTclClasses* nl = sl;
        while (nl->next != sc) nl = nl->next;
        Tcl_AppendElement(in, className(nl->cl));
        sc = nl;
      }
    } else {
      XOTclClass* isc = XOTclGetClass(in, pattern);
      XOTclClasses* pl;
      if (isc == 0) return XOTclErrBadVal(in, "a class", pattern);
      pl = ComputePrecedence(cl);

      /*
       * search precedence to see if we're related or not
       */

      while (pl != 0) {
        if (pl->cl == isc) {
	  Tcl_SetObjResult(in,global_objects[ONE]);
          break;
        }
        pl = pl->next;
      }
      if (pl == 0)
	Tcl_SetObjResult(in,global_objects[ZERO]);
    }
  } else if (!strcmp(cmd, "subclass")) {
    if (objc > 3) 
      return XOTclObjErrArgCnt(in,
				    cl->object.cmdName,
				    "info subclass ?class?");
    if (objc == 2) {
      XOTclClasses* sl = cl->sub;
      XOTclClasses* sc = 0;

      /*
       * order unimportant
       */

      Tcl_ResetResult(in);
      for (sc = sl; sc != 0; sc = sc->next)
        Tcl_AppendElement(in, className(sc->cl));
    } else {
      XOTclClass* isc = XOTclGetClass(in, pattern);
      XOTclClasses* pl;
      XOTclClasses* saved;
      if (isc == 0) return XOTclErrBadVal(in, "a class", pattern);
      saved = cl->order; cl->order = 0;
      pl = ComputeDependents(cl);

      /*
       * search precedence to see if we're related or not
       */

      while (pl != 0) {
        if (pl->cl == isc) {
	  Tcl_SetObjResult(in,global_objects[ONE]);
          break;
        }
        pl = pl->next;
      }
      if (pl == 0)
	Tcl_SetObjResult(in,global_objects[ZERO]);
      RC(cl->order); cl->order = saved;
    }
  } else if (!strcmp(cmd, "heritage")) {
    XOTclClasses* pl = ComputePrecedence(cl);

    if (objc > 3) 
      return XOTclObjErrArgCnt(in,
				    cl->object.cmdName,
				    "info heritage ?pat?");
    if (pl) pl = pl->next;
    Tcl_ResetResult(in);
    for (; pl != 0; pl = pl->next) {
      char* name = className(pl->cl);
      if (pattern && !Tcl_StringMatch(name, pattern)) continue;
      Tcl_AppendElement(in, name);
    }
  } else if (!strcmp(cmd, "instances")) {
    if (objc > 3)
        return XOTclObjErrArgCnt(in,cl->object.cmdName,
				      "info instances ?pat?");
    ListInstanceKeys(in, &cl->instances, pattern);
  } else if (!strcmp(cmd, "instcommands")) {
    if (objc > 3)
        return XOTclObjErrArgCnt(in,cl->object.cmdName,
				      "info instcommands ?pat?");
    ListKeys(in, &cl->nsPtr->cmdTable, pattern);
  } else if (!strcmp(cmd, "instprocs")) {
    if (objc > 3) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						"info instprocs ?pat?");
    ListProcKeys(in, &cl->nsPtr->cmdTable, pattern);
  } else if (!strcmp(cmd, "instargs")) {
    if (objc != 3)
        return XOTclObjErrArgCnt(in,cl->object.cmdName,
				      "info instargs <instproc>");
    return ListProcArgs(in, &cl->nsPtr->cmdTable, pattern);
  } else if (!strcmp(cmd, "instdefault")) {
    if (objc != 5)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,
                           "info instdefault <instproc> <arg> <var>");
    return ListProcDefault(in, &cl->nsPtr->cmdTable, pattern,
                           ObjStr(objv[3]), ObjStr(objv[4]));
  } else if (!strcmp(cmd, "instbody")) {
    if (objc != 3)
        return XOTclObjErrArgCnt(in,cl->object.cmdName,
				      "info instbody <instproc>");
    return ListProcBody(in, &cl->nsPtr->cmdTable, pattern);
  } else if (!strcmp(cmd, "classchildren")) {
    if (objc > 3) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						"info classchildren ?pat?");
    ListClassChildren(in, cl, pattern);
  } else if (!strcmp(cmd, "classparent")) {
    if (objc > 2) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						"info classparent");
    ListParent(in, cl->nsPtr);
  } else if (!strcmp(cmd, "filters")) {
    if (objc != 2) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						 "info filters");
    FilterList(in, cl->filters);
  } else if (!strcmp(cmd, "instinvar")) {
    if (objc != 2) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						 "info instinvar");
    AssertionList(in, cl->assertions.invariants);
  } else if (!strcmp(cmd, "instpre")) {
    ProcAssertion* procs;
    if (objc != 3) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						 "info instpre <proc>");
    procs = AssertionFindProcs (&cl->assertions, ObjStr(objv[2]));
    if (procs) AssertionList(in,procs->pre);
  } else if (!strcmp(cmd, "instpost")) {
    ProcAssertion* procs;
    if (objc != 3) return XOTclObjErrArgCnt(in,cl->object.cmdName,
						 "info instpost <proc>");
    procs = AssertionFindProcs (&cl->assertions, ObjStr(objv[2]));
    if (procs) AssertionList(in,procs->post);
  } else {
    int result = XOTclOInfoMethod(cd, in, objc, (Tcl_Obj **)objv);
    return result;
  }
  return TCL_OK;
}

static int
XOTclCParametersMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  XOTclObject* obj = XOTclAsObject(in, cd);
  Tcl_Obj     **pv=0;
  Tcl_Obj     **ov=0;
  int         i, elts, pc, oc, result = TCL_OK;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,"parameters ?params?");
  if ((result = Tcl_ListObjGetElements(in, objv[1], &pc, &pv)) == TCL_OK) {
    for (elts = 0; elts < pc; elts++) {
      if ((result = Tcl_ListObjGetElements(in, pv[elts], &oc, &ov)) == TCL_OK) {
	if (oc > 0) {
	  result = Tcl_VarEval (in, className(cl),
				" instproc ", ObjStr(ov[0])," args {\n", 
				"  if {[llength $args] == 0} {\n",
				"    return [::set [self]::[self proc]]\n", 
				"  } elseif {[llength $args] == 1} {\n",
				"    return [::set [self]::[self proc] [lindex $args 0]]\n",
				"  } else {\n",
				"    error \"wrong # args: [self] [self proc] ?value?\"\n",  
				"  }\n}",
				(char*) 0);
	  if (result != TCL_OK)
	    return result;
	}

	for (i=1; i < oc; i++) {
	  char *flag = ObjStr(ov[i]);
	  char* methodName;
	  int j, nrArgs;
  
	  if (*flag == '-') {
	    methodName = flag+1;
	  } else {
	    methodName = global_strings[DEFAULT];
	    i--;
	  }
	
	  for (j = i+1 , nrArgs=0; j < oc; j++, nrArgs++) {
	    char* arg = ObjStr(ov[j]);
	    if (*arg == '-') {
	      break;
	    }
	  }
	
	  {
	    Tcl_Obj *tmp;
	    int k;
	    int ac = 3+nrArgs;
	    XOTclClass* savedCl = cl->object.cl, *paramCl;

	    XOTclClass     *callCl = 0;
	    ClientData     cp = 0;
	    Tcl_ObjCmdProc *proc = 0;
	    Command        *procInfo;

	    DEFINE_NEW_TCL_OBJECTS_ON_STACK(ac,args);

	    
	    /* if the parameter class is not set by the user, use the
	       general parameter class */
	    if (!cl->parameterClass) {
	      paramCl = XOTclGetClass(in, global_strings[PARAM_CL]);
	    } else {
	      paramCl = XOTclGetClass(in, ObjStr(cl->parameterClass));
	    }
	 
	    if(!paramCl)
	      return XOTclVarErrMsg(in, "Parameters: Can't find ",
					 global_strings[PARAM_CL], (char*) 0);

	  
	    cl->object.cl = paramCl;

	    args[0] = cl->object.cmdName;
	    XOTclNewObj(tmp);
	    INCR_REF_COUNT(tmp);
	    args[1] = tmp;
	    Tcl_SetStringObj(args[1], methodName, -1);
	    args[2] = ov[0];
	    
	    for (k = 0; k < nrArgs; k++) {
	      args[3+k] = ov[1+k+i];
	    }

	    procInfo = FindMethod(methodName, obj->nsPtr);
	    if (procInfo) {
	      callCl = 0;
	    } else {
	      callCl = SearchCMethod(obj, obj->cl, NULL, methodName, &procInfo);
	    }
	    if (procInfo) {
	      proc = procInfo->objProc;
	      cp   = procInfo->objClientData;
	    } else {
	      proc = 0;
	      cp   = 0;
	    }
	    cl->object.cl = savedCl;

	    if (proc) {
	      if (cp == 0)
		cp = cd;
	      result = callProcCheck(cp, in, ac-1, args+1, procInfo,
				     obj, callCl, methodName, 0, 0, 0);

	    }
	    
	    FREE_TCL_OBJECTS_ON_STACK(args);

	    if (result != TCL_OK) {
	      Tcl_AppendResult(in, " during {", ObjStr(args[0]), "} {",
			       flag, "}", 0);
	      DECR_REF_COUNT(tmp);
	      return result;
	    }
	    DECR_REF_COUNT(tmp);
	  }
	  i += nrArgs;
	}
      } else
	break;
    }
  }
  return result;
}

static int
XOTclCParameterClassMethod(ClientData cd, Tcl_Interp* in, 
			   int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  char* paramClStr;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,"parameterclass cl");
  paramClStr = ObjStr(objv[1]);

  if ((paramClStr == NULL) || (*paramClStr == '\0')) {
    cl->parameterClass = 0;
  } else {
    cl->parameterClass = objv[1];
  }
  return TCL_OK;
}

static int
XOTclCInstProcMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  char *argStr, *bdyStr, *name;
  char* clName;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 4 && objc != 6)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,"instproc name args body ?preAssertion postAssertion?");

  /*
   * if the args list is "auto", the body is a script to load the proc
   */
  argStr = ObjStr(objv[2]);
  bdyStr = ObjStr(objv[3]);
  name = ObjStr(objv[1]);
  clName =  className(cl);
  if ((isDestroyString(name) && strcmp(clName,"::Object") == 0) ||
      (isCreateString(name) && strcmp(clName,"::Class") == 0)
      )
    return XOTclVarErrMsg(in, clName, " instproc: '", name, "' of ",
			       clName, " can not be overwritten. Derive a ",
			       "sub-class", (char*) 0);
#ifdef PROCCACHE
  ProcCacheCRemoveMethod (cl,name);
#endif

  if (isAutoString(argStr)) {
    /*
      CallStackContent *csc = CallStackGetFrame(in);
      fprintf(stderr,"Instpr -- self %s\n",xobjName(csc->self));
      fprintf(stderr,"Instpr -- c %s\n",ClassName(csc->class));
      fprintf(stderr,"Instpr -- p %s\n",csc->procName);
    */
    if (!MakeAuto(in,  bdyStr, name, cl->nsPtr))
      return XOTclMakeFailure(in, objv, 
				   "MakeAuto failed in XOTclCProcMethod\n");
  }
  else if (*argStr == 0 && *bdyStr == 0) {
    AssertionRemoveProc (&cl->assertions, name);
    NSDeleteCmd(in,cl->nsPtr,name);
  } else {
    MakeProc(cl->nsPtr, &cl->assertions, in, objc, (Tcl_Obj **) objv);
  }
    return TCL_OK;
}

static int
XOTclCMoveMethod(ClientData cd, Tcl_Interp* in,int objc, Tcl_Obj *objv[]){
  XOTclClass* cl = XOTclAsClass(in, cd);
  char *self, *newName, *newObjName, *newClName;
  Tcl_DString newClNameDS;
  int result;
  Namespace *oldClNS = cl->nsPtr, *oldObjNS = cl->object.nsPtr;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,"move newName");

  newName =  ObjStr(objv[1]);
  self = className(cl);

  if ((newName == NULL) || (*newName == '\0')) {
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(2,ov);
    ov[0] = cl->object.cmdName;
    ov[1] = global_objects[DESTROY];
    cl->object.destroyCalled = 1;

    result = ObjDispatch((ClientData) cl, in, 2, ov, 1);
    FREE_TCL_OBJECTS_ON_STACK(ov);

    return result;
  }

  NSAddXOTclClasses(newName, &newClNameDS);
  newClName = Tcl_DStringValue(&newClNameDS);
  newObjName = NSCutXOTclClasses(newClName);

  if ((result=NSRenameCommand(in, newObjName, self)) != TCL_OK) 
    return TCL_ERROR;

  cl->nsPtr = NSCreateAndCopy (in, cl->nsPtr, newClName, (ClientData) cl);
  cl->object.nsPtr = NSCreateAndCopy (in, cl->object.nsPtr, newObjName, (ClientData) cl);

  /* correct Info in cmdName */
  NSCorrectCmdName(in, &cl->object);
  Tcl_DStringFree (&newClNameDS);

  /* remove old Namespaces */
  Tcl_DeleteNamespace((Tcl_Namespace *) oldClNS);
  Tcl_DeleteNamespace((Tcl_Namespace *) oldObjNS);
  return result;
}

static int
XOTclCFilterMethod(ClientData cd, Tcl_Interp* in,int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  int i, result = TCL_OK;
  Tcl_Obj **ov; int oc;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) return XOTclObjErrArgCnt(in,cl->object.cmdName,
    "filter filterNameList");

  if (Tcl_ListObjGetElements(in, objv[1], &oc,&ov) != TCL_OK) 
    return TCL_ERROR;
  FilterRemoveList(cl);
  for (i = 0; i < oc; i ++) {
    result = FilterAdd (in, cl, ObjStr (ov[i]));
    if (result == TCL_ERROR) {
      return XOTclVarErrMsg(in,
       "filter: can't find filterproc or filterproc is not an instproc - Class: ",
       className(cl), " - Proc: ",
       ObjStr (ov[i]), (char*) NULL);
    }
  }
  return result;
}

/*
 *  Searches for filter on [self] and returns fully qualified name
 *  if it is not found it returns an empty string 
 */
static int
XOTclCFilterSearchMethod(ClientData cd, Tcl_Interp* in, int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);
  Command *cmdPtr;
  char *simpleName, *fullName;
  Tcl_DString fullNameDS;
  Tcl_HashEntry* hPtr;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc < 2) return XOTclObjErrArgCnt(in,cl->object.cmdName,
    "filtersearch filterName");

  Tcl_ResetResult(in);

  cmdPtr = FilterSearch(in, ObjStr(objv[1]), cl);
  if (!cmdPtr) {
    return TCL_OK;
  }

  simpleName = Tcl_GetCommandName (in, (Tcl_Command)cmdPtr);
  ALLOC_NAME_NS (&fullNameDS, cmdPtr->nsPtr->fullName, simpleName);
  fullName = Tcl_DStringValue(&fullNameDS);
  hPtr = Tcl_FindHashEntry(&XOTclFilterList, (char*)cmdPtr);

  if (hPtr) {
    Tcl_AppendElement(in, fullName);
  }

  Tcl_DStringFree(&fullNameDS);
  return TCL_OK;
}

static int
XOTclCInvariantsMethod(ClientData cd, Tcl_Interp* in,int objc, Tcl_Obj *objv[]) {
  XOTclClass* cl = XOTclAsClass(in, cd);

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,
				    "instinvar <invariantList>");

  AssertionDeleteList (cl->assertions.invariants);
  cl->assertions.invariants = AssertionNewList (in, objv[1]);

  return TCL_OK;
}

static int
XOTclCCopyMethod(ClientData cd, Tcl_Interp *in, 
		 int objc, Tcl_Obj *CONST objv[]) {
  XOTclClass *newcl, *cl = XOTclAsClass(in, cd);
  char *newClName, *newObjName;
  Tcl_DString newClNameDS;

  if (!cl) return XOTclObjErrType(in, objv[0], "Class");
  if (objc != 2)
      return XOTclObjErrArgCnt(in,cl->object.cmdName,
				    "copy <destination>");

  newcl = PrimitiveCCreate(in, ObjStr(objv[1]), cl->object.cl);
  if (newcl == 0) 
    return XOTclErrMsg(in,"Class copy failed", TCL_STATIC);

  NSAddXOTclClasses(ObjStr(objv[1]), &newClNameDS);
  newClName = Tcl_DStringValue(&newClNameDS);
  newObjName = NSCutXOTclClasses(newClName);

  NSCopy (in, cl->nsPtr, newClName, (ClientData) cl);
  NSCopy (in, cl->object.nsPtr, newObjName, (ClientData) cl);

  Tcl_DStringFree (&newClNameDS);
  return TCL_OK;
}

static int
XOTclCUnknownMethod(ClientData cd, Tcl_Interp *in, 
		    int objc, Tcl_Obj *CONST objv[]) {
  XOTclObject* obj = (XOTclObject*) cd;
  int result;
  char* procName, *self = ObjStr(obj->cmdName);

  if (objc < 2) return XOTclObjErrArgCnt(in,objv[0],"unknown args");

  procName =  ObjStr(objv[1]);

  if (isCreateString(self))
    return XOTclVarErrMsg(in, "error ", self, ": unable to dispatch '",
			  procName, "'", (char*)NULL);

  {
    int i, oc = objc+1;
    DEFINE_NEW_TCL_OBJECTS_ON_STACK(oc,ov);

    ov[0] = obj->cmdName;
    INCR_REF_COUNT(ov[0]);

    ov[1] = global_objects[CREATE];
    for (i = 2; i < oc; i++) {
      ov[i] = objv[i-1];
    }

    result =  XOTclObjDispatch(cd, in, oc, ov);

    DECR_REF_COUNT(ov[0]);
    FREE_TCL_OBJECTS_ON_STACK(ov);

    return result;
  }
}

/*
 * New Tcl Commands
 */
static int
XOTcl_InfoObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result, filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  if (objc > 1 && isLevelString(ObjStr(objv[1])) &&
      FilterEntryOnCallStack()) {
    filterFrameCutOff = 1;
    FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
  }
  result = Tcl_InfoObjCmd(cd,in,objc,objv);
  if (filterFrameCutOff)
    FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);
  return result;
}

static int
XOTcl_UplevelObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result;
  int filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  if (FilterEntryOnCallStack()) {
    filterFrameCutOff = 1;
    FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
  }
  result = Tcl_UplevelObjCmd(cd,in,objc,objv);
  if (filterFrameCutOff)
    FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);
  return result;
}

static int
XOTcl_UpvarObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  int result;
  int filterFrameCutOff = 0;
  CallFrame *savedFramePtr = 0, *savedVarFramePtr = 0;
  if (FilterEntryOnCallStack()) {
    filterFrameCutOff = 1;
    FilterCutInactiveFrames(in, &savedFramePtr, &savedVarFramePtr);
  }
  result = Tcl_UpvarObjCmd(cd,in,objc,objv);
  if (filterFrameCutOff)
    FilterRestoreFrames (in, savedFramePtr, savedVarFramePtr);
  return result;
}

static int
XOTcl_RenameObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  /* this call the Tcl_RenameObjCmd, but it ensures before that
     the renamed obj, functions, etc. are not part of XOTcl */
  char *oldName, *newName, *on;

  XOTclObject* obj; XOTclClass* cl;

  Tcl_Command cmd;
  Command *cmdPtr;
  int result;
  char* oldSimpleCmdName=0, *newSimpleCmdName;
  Tcl_DString oldDS;
  Tcl_Obj* objResult;

  /* wrong # args => normal Tcl ErrMsg*/
  if (objc != 3) return Tcl_RenameObjCmd(cd, in, objc, objv);

  oldName = ObjStr(objv[1]);
  on = NSCutXOTclClasses(oldName);
  newName = ObjStr(objv[2]);

  /* if an obj/cl should be renamed => call the XOTcl-rename command */
  cmd = Tcl_FindCommand(in, oldName, (Tcl_Namespace *) NULL,0);
  cmdPtr = (Command *) cmd;

  obj = XOTclGetObject(in,oldName);
  cl = XOTclGetClass (in,on);

  Tcl_DStringInit(&oldDS);

  if (cmdPtr != NULL) {
    if (obj || cl) {
      Tcl_Obj* args[3];
      ClientData clientData = (obj != NULL) ? (ClientData) obj : (ClientData) cl;
      args[0] = (obj ? obj->cmdName : cl->object.cmdName);
      args[1] = global_objects[MOVE];
      args[2] = objv[2];
      return XOTclObjDispatch(clientData,in,3,args);
    }

    /* Save the SimpleName */
    Tcl_DStringAppend(&oldDS, Tcl_GetCommandName(in,cmd), -1);
    oldSimpleCmdName = Tcl_DStringValue(&oldDS);

    /* saving obj/cl for proc's; they could be destroyed */

    if (cmdPtr->proc == XOTclDispatch) {
      obj = XOTclAsObject(in, cmdPtr->clientData);
      cl = XOTclIsClass(in, cmdPtr->clientData);
    } else {
      obj = 0;
      cl = 0;
    }
  }
  /* Do the rename */

  result = Tcl_RenameObjCmd(cd, in, objc, objv);

  /* has a XOTcl Proc/Instproc been renamed */
  if (result == TCL_OK && cmdPtr != NULL && (obj || cl)) {
    if ((newName == NULL) || (*newName == '\0')) {
      /* it was deleted -> delete from ProcCaches
	 if (cl) ProcCacheCRemoveMethod (cl, oldSimpleCmdName);*/
    } else {
      /* it was renamed -> correct the assertion lists*/
      Tcl_Obj* pre; Tcl_Obj* post;
      AssertionStore *as = (obj ? &obj->assertions : &cl->assertions);

      ProcAssertion *pa = AssertionFindProcs (as,oldSimpleCmdName);	
      if (pa) {
	newSimpleCmdName = Tcl_GetCommandName(in,cmd);
	
	objResult = Tcl_GetObjResult(in);
	AssertionList (in, pa->pre);
	pre = Tcl_GetObjResult(in);
	Tcl_ResetResult (in);
	AssertionList (in, pa->post);
	post = Tcl_GetObjResult(in);
	Tcl_SetObjResult(in, objResult);
	
	AssertionRemoveProc (as, oldSimpleCmdName);
	AssertionAddProc(in,newSimpleCmdName, as, pre, post);
	
      }
    }
  }
  Tcl_DStringFree (&oldDS);
  return result;
}

static int
XOTcl_TraceObjCmd(ClientData cd, Tcl_Interp *in, int objc, Tcl_Obj *CONST objv[]) {
  char* option;
  if (objc != 2)
    return XOTclObjErrArgCnt(in, NULL, "xotcltrace");

  option = ObjStr(objv[1]);
  if (strcmp(option,"stack") == 0) {
    XOTclStackTrace(in);
    return TCL_OK;
  }
  if (strcmp(option,"callstack") == 0) {
    XOTclCallStackTrace(in);
    return TCL_OK;
  }
  if (strcmp(option,"filters") == 0) {
    XOTclFilterTrace(in);
    return TCL_OK;
  }
  return XOTclVarErrMsg(in, "xotcltrace: unknown option", (char*) NULL);
}

/*
 * C interface routines for manipulating objects and classes
 */

static XOTCLINLINE XOTclClass*
XOTclIsClass(Tcl_Interp* in, ClientData cd) {
  return IsType((XOTclObject*)cd, InClass(in)) ? (XOTclClass*)cd : 0;
}

extern XOTclObject*
XOTclGetObject(Tcl_Interp* in, char* name) {
  Command *cmdPtr;
  XOTclObject* obj=0;
  assert(name);
  cmdPtr = (Command*)Tcl_FindCommand(in, name, NULL,0);
  if (cmdPtr && cmdPtr->proc == XOTclDispatch)
    obj = XOTclAsObject(in, cmdPtr->clientData);
  else
    obj = 0;
  return obj;
}

extern XOTclClass*
XOTclGetClass(Tcl_Interp* in, char* name) {
  Command *cmdPtr;
  XOTclClass* cl=0;
  assert(name);
  cmdPtr = (Command*) Tcl_FindCommand(in, name, NULL,0);

  if (cmdPtr && cmdPtr->proc == XOTclDispatch)
      cl = XOTclAsClass(in, cmdPtr->clientData);
  else
      cl = 0;
  return cl;
}

extern XOTclObject*
XOTclCreateObject(Tcl_Interp* in, Tcl_Obj* name, XOTclClass* cl) {
  Tcl_Obj* args[3];
  args[0] = cl->object.cmdName;
  args[1] = global_objects[CREATE];
  args[2] = name;
  INCR_REF_COUNT(args[2]);
  if (XOTclObjDispatch((ClientData)cl,in,3,args) != TCL_OK) return 0;
  DECR_REF_COUNT(args[2]);
  return XOTclGetObject(in, ObjStr(name));
}

extern XOTclClass*
XOTclCreateClass(Tcl_Interp* in, Tcl_Obj* name, XOTclClass* cl){
  Tcl_Obj* args[3];
  args[0] = cl->object.cmdName;
  args[1] = global_objects[CREATE];
  args[2] = name;
  INCR_REF_COUNT(args[2]);
  if (XOTclObjDispatch((ClientData)cl,in,3,args) != TCL_OK) return 0;
  DECR_REF_COUNT(args[2]);
  return XOTclGetClass(in, ObjStr(name));
}

int
XOTclDeleteObject(Tcl_Interp* in, XOTclObject* obj) {
  Tcl_Obj* args[2]; int result;
  args[0] = obj->cmdName;
  args[1] = global_objects[DESTROY];
  result =  XOTclObjDispatch((ClientData)obj, in, 2, args);
  return result;
}

int
XOTclDeleteClass(Tcl_Interp* in, XOTclClass* cl) {
  Tcl_Obj* args[2]; int result;
  args[0] =  cl->object.cmdName;
  args[1] = global_objects[DESTROY];
  result =  XOTclObjDispatch((ClientData)cl, in, 2, args);
  return result;
}

void
XOTclAddPMethod(Tcl_Interp* in, XOTclObject* obj, char* nm, Tcl_ObjCmdProc* proc,
               ClientData cd, Tcl_CmdDeleteProc* dp)
{
  Tcl_DString newCmd;
  ALLOC_NAME_NS (&newCmd, obj->nsPtr->fullName, nm);
  Tcl_CreateObjCommand(in, Tcl_DStringValue(&newCmd), proc, cd, dp);
  Tcl_DStringFree(&newCmd);
}

void
XOTclAddIMethod(Tcl_Interp* in, XOTclClass *cl, char* nm,
                  Tcl_ObjCmdProc* proc, ClientData cd, Tcl_CmdDeleteProc* dp)
{
  Tcl_DString newCmd;
  ALLOC_NAME_NS (&newCmd, cl->nsPtr->fullName, nm);
  Tcl_CreateObjCommand(in, Tcl_DStringValue(&newCmd), proc, cd, dp);
  Tcl_DStringFree(&newCmd);
}

int
XOTclRemovePMethod(Tcl_Interp* in, XOTclObject* obj, char* nm) {
  return NSDeleteCmd (in,obj->nsPtr, nm);
}

extern int
XOTclRemoveIMethod(Tcl_Interp* in, XOTclClass* cl, char* nm) {
  return NSDeleteCmd(in, cl->nsPtr, nm);
}

/*
 * Begin additions by Martin Andrews.
 */

extern Tcl_Obj*
XOTclOSetInstVar2(XOTclObject* obj,Tcl_Interp* in,Tcl_Obj* name1,Tcl_Obj* name2,
                 Tcl_Obj* value,int flgs) {
  Tcl_Obj* result;
  CallFrame frame, *oldFptr;

  oldFptr = ((Interp *)in)->varFramePtr;
  frame.nsPtr = obj->nsPtr;
  ((Interp *)in)->varFramePtr = &frame;

#if defined(V81)
  result = 
    Tcl_SetObjVar2(in, ObjStr(name1), 
		   ((name2==NULL) ? (char*)NULL: ObjStr(name2)), 
		   value, flgs|TCL_PARSE_PART1|TCL_NAMESPACE_ONLY);
#else
  result = 
    Tcl_ObjSetVar2(in, name1, name2, value, 
		   flgs|TCL_PARSE_PART1|TCL_NAMESPACE_ONLY);
#endif
  ((Interp *)in)->varFramePtr = oldFptr;

  return result;
}

extern Tcl_Obj*
XOTclOGetInstVar2(XOTclObject* obj, Tcl_Interp* in, Tcl_Obj* name1, Tcl_Obj *name2,
        int flgs){

  Tcl_Obj* result;
  CallFrame frame, *oldFptr;

  oldFptr = ((Interp *)in)->varFramePtr;
  frame.nsPtr = obj->nsPtr;
  ((Interp *)in)->varFramePtr = &frame;

#if defined(V81)
  result = 
    Tcl_GetObjVar2(in, ObjStr(name1), 
		   ((name2==NULL) ? (char*)NULL: ObjStr(name2)), 
		   flgs|TCL_PARSE_PART1|TCL_NAMESPACE_ONLY);
#else
  result = 
    Tcl_ObjGetVar2(in, name1, name2, 
		   flgs|TCL_PARSE_PART1|TCL_NAMESPACE_ONLY);
#endif
  ((Interp *)in)->varFramePtr = oldFptr;

  return result;
}

extern int
XOTclUnsetInstVar2(XOTclObject* obj, Tcl_Interp* in, char* name1, char *name2,
        int flgs) {
  int result;
  Tcl_DString fullName;
  ALLOC_NAME_NS (&fullName,obj->nsPtr->fullName,name1);
  result = Tcl_UnsetVar2(in, Tcl_DStringValue(&fullName),name2, flgs);
  Tcl_DStringFree (&fullName);
  return result;
}

extern int
XOTclInstVar(XOTclObject* obj, Tcl_Interp* in, char* name,char *destName) {
  int result;
  Tcl_DString ds, *dsp=&ds;

  ALLOC_NAME_NS (dsp, obj->nsPtr->fullName, name);

  result = XOTclInstVarDS(obj, in, dsp, destName);

  Tcl_DStringFree(dsp);
  return result;
}

extern int
XOTclInstVar2(XOTclObject* obj, Tcl_Interp* in, char* name1, 
	      char *name2, char* destName) {
  int result;
  Tcl_DString oneName;
  ALLOC_NAME_NS(&oneName, obj->nsPtr->fullName, name1);
  if (name2) {
    Tcl_DStringAppend(&oneName, "(", 1);
    Tcl_DStringAppend(&oneName, name2, -1);
    Tcl_DStringAppend(&oneName, ")", 1);
  }
  result = XOTclInstVarDS(obj,in,&oneName,destName);
  Tcl_DStringFree(&oneName);
  return result;
}

/*
 * End additions by Martin Andrews.
 */

extern void
XOTclSetObjectData(XOTclObject* obj, XOTclClass* cl, ClientData data) {
  Tcl_HashEntry *hPtr;
  int nw;

  if (!cl->objectdata) {
    cl->objectdata = (Tcl_HashTable*)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(cl->objectdata, TCL_ONE_WORD_KEYS);
  }
  hPtr = Tcl_CreateHashEntry(cl->objectdata, (char*)obj, &nw);
  Tcl_SetHashValue(hPtr, data);
}

extern int
XOTclGetObjectData(XOTclObject* obj, XOTclClass* cl, ClientData* data) {
  Tcl_HashEntry *hPtr;

  if (!cl->objectdata) return 0;
  hPtr = Tcl_FindHashEntry(cl->objectdata, (char*)obj);
  if (data) *data = hPtr ? Tcl_GetHashValue(hPtr) : 0;
  return (hPtr != 0);
}

extern int
XOTclUnsetObjectData(XOTclObject* obj, XOTclClass* cl) {
  Tcl_HashEntry *hPtr;

  if (!cl->objectdata) return 0;
  hPtr = Tcl_FindHashEntry(cl->objectdata, (char*)obj);
  if (hPtr) Tcl_DeleteHashEntry(hPtr);
  return (hPtr != 0);
}

/*
 *  Exit Handler
 */

static void
Exit_DestroyCall(Tcl_Interp* in, XOTclObject* obj) {
  Tcl_Obj *args[2];
  int result;

  /*
    fprintf(stderr,"XOTcl_ExitProc Objects <%s> destroyCalled=%d\n", 
    ObjStr(obj->cmdName), obj->destroyCalled);
  */ 
  args[0] = obj->cmdName;
  args[1] = global_objects[DESTROY];

  INCR_REF_COUNT(args[0]);
  result =  XOTclObjDispatch((ClientData)obj, in, 2, args);
  if (result != TCL_OK) {
    Tcl_VarEval(in, 
		"puts stderr \"ErrorCode: $::errorCode, ErrorInfo: $::errorInfo\"", 
		(char*) 0);
    panic("Exit Handler: Error in destroy call of object '%s' (line-info: %d)\n", 
	  ObjStr(args[0]), in->errorLine);
  }
  DECR_REF_COUNT(args[0]);
}

static void
XOTcl_ExitProc(ClientData cd) {
  Tcl_Interp* in = (Tcl_Interp*)  cd;
  Interp* iPtr = (Interp*) in;
  XOTclObject* obj;
 
  Tcl_Obj *exitHandler;
  Tcl_HashSearch hSrch;

#if defined(V81)
  exitHandler = Tcl_GetObjVar2(in, global_strings[EXIT_HANDLER], NULL, 0);
#else
  exitHandler = Tcl_ObjGetVar2(in, global_objects[EXIT_HANDLER], NULL, 0);
#endif /*  V81 */

  if (exitHandler) {
    int result = EvalObj(in, exitHandler);
    if (result != TCL_OK) {
      panic("User defined exit handler contains errors! (line-info: %d)\n Execution interruted.\n",
	    in->errorLine);
    }
  }

  if (iPtr && iPtr->globalNsPtr) {
    Tcl_HashTable* table = &iPtr->globalNsPtr->cmdTable;
    Tcl_HashEntry* hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
    
    /* delete objects */
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      char* key = Tcl_GetHashKey(table, hPtr);
      XOTclClass* cl = XOTclGetClass(in,key);
      
      obj = XOTclGetObject(in, key);
      if (!XOTclIsClass(in, (ClientData) cl))
	cl =0;
      if (obj && !cl && !(strcmp (key,"Object") == 0) && 
	  !(strcmp (key,"Class") == 0)) {
	Exit_DestroyCall(in, obj);
      }
    }

    /* delete classes */
    hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      char* key = Tcl_GetHashKey(table, hPtr);
      XOTclClass* cl = XOTclGetClass(in,key);

      obj = XOTclGetObject(in, key);
      if (obj && cl && !(strcmp (key,"Object") == 0) && 
	  !(strcmp (key,"Class") == 0) && 
	  !IsMetaClass(in, cl)) {
	Exit_DestroyCall(in, obj);
      }
    }

    /* delete meta-classes */
    hPtr = table ? Tcl_FirstHashEntry(table, &hSrch) : 0;
    for (; hPtr != 0; hPtr = Tcl_NextHashEntry(&hSrch)) {
      char* key = Tcl_GetHashKey(table, hPtr);
      obj = XOTclGetObject(in, key);
      if (obj && !(strcmp (key,"Object") == 0) && 
	  !(strcmp (key,"Class") == 0)) {
	Exit_DestroyCall(in, obj);
      }
    }
  }
}

/*
 * Tcl extension initialization routine
 */

#define MAXTCLPROC 4096

/* das hilft auch nicht fuer das package provide problem ...
extern int
Xotcl_SafeInit(Tcl_Interp* in) {
  Xotcl_Init(in);
}
*/

extern int
Xotcl_Init(Tcl_Interp* in) {
  XOTclClass* theobj = 0;
  XOTclClass* thecls = 0;
  XOTclClass* paramCl = 0;
  Tcl_HashEntry* hp1;
  Tcl_HashEntry* hp2;

  int i, nw1, nw2;
  char tm[MAXTCLPROC];
  
  /*
   * discover Tcl's hidden proc interpreter
   */

  if (ProcInterpId == 0) {
    char *name = "_fake_";
    Tcl_Obj* args[4];

    args[0] = Tcl_NewStringObj("proc",4); INCR_REF_COUNT(args[0]);
    args[1] = Tcl_NewStringObj(name,-1); INCR_REF_COUNT(args[1]);
    args[2] = Tcl_NewStringObj("",0); INCR_REF_COUNT(args[2]);
    args[3] = Tcl_NewStringObj("return",6); INCR_REF_COUNT(args[3]);
    if (Tcl_ProcObjCmd(0, in, 4, args) == TCL_OK) {
      Tcl_CmdInfo info;
      if (Tcl_GetCommandInfo(in, name, &info)) {
        ProcInterpId = info.proc;
        ProcObjInterpId = info.objProc;
        (void)Tcl_DeleteCommand(in, name);
      } else return XOTclErrMsg(in, "proc failed", TCL_STATIC);
    } else return TCL_ERROR;
  }

  RegisterObjTypes();

  /*
   * initialize global Tcl_Obj*
   */

  for (i = 0; i < nr_elements(global_strings); i++) {
    global_objects[i] = Tcl_NewStringObj(global_strings[i],-1);
    INCR_REF_COUNT(global_objects[i]);
  }

  /* XOTclClasses in separate Namespace / Objects in globalNS */

  XOTclClassesNS =  (Namespace*) Tcl_CreateNamespace (in,"XOTclClasses",
               (ClientData)NULL,(Tcl_NamespaceDeleteProc*)NULL);

  XOTclClassesNS->clientData = (ClientData) ckalloc (sizeof(CallStack));
  memset(((CallStack *)XOTclClassesNS->clientData)->content,
	 0,sizeof(CallStackContent));
  ((CallStack *)XOTclClassesNS->clientData)->top =
    ((CallStack *)XOTclClassesNS->clientData)->content;
  ((CallStack *)XOTclClassesNS->clientData)->frameCutRound = 0;

  /* register Exit Handler */
  Tcl_CreateExitHandler(XOTcl_ExitProc, (ClientData) in);

  /*
   * bootstrap the tables of base objects and classes
   */

  if (theObjects == 0) {
    theObjects = (Tcl_HashTable*)ckalloc(sizeof(Tcl_HashTable));
    if (!theObjects) return XOTclErrMsg(in, "Object table failed", TCL_STATIC);
    Tcl_InitHashTable(theObjects, TCL_ONE_WORD_KEYS);
  }

  if (theClasses == 0) {
    theClasses = (Tcl_HashTable*)ckalloc(sizeof(Tcl_HashTable));
    if (!theClasses) return XOTclErrMsg(in, "Class table failed", TCL_STATIC);
    Tcl_InitHashTable(theClasses, TCL_ONE_WORD_KEYS);
  }

  hp1 = Tcl_CreateHashEntry(theObjects, (char*)in, &nw1);
  if (nw1) theobj = PrimitiveCCreate(in, "Object", 0);
  hp2 = Tcl_CreateHashEntry(theClasses, (char*)in, &nw2);
  if (nw2) thecls = PrimitiveCCreate(in, "Class", 0);

  if (!nw1 && !nw2) {
    Tcl_SetObjResult(in,global_objects[ZERO]);
    return TCL_OK;
  } else if (!theobj || !thecls) {
    if (theobj) PrimitiveCDestroy((ClientData)theobj);
    if (thecls) PrimitiveCDestroy((ClientData)thecls);
    return XOTclErrMsg(in, "Object/Class failed", TCL_STATIC);
  }

  Tcl_SetHashValue(hp1, (char*)theobj);
  Tcl_SetHashValue(hp2, (char*)thecls);

  theobj->object.type = thecls;
  theobj->parent = 0;
  thecls->object.type = thecls;
  thecls->parent = theobj;

  AddInstance((XOTclObject*)theobj, thecls);
  AddInstance((XOTclObject*)thecls, thecls);
  AddSuper(thecls, theobj);

  /* global data -> to be moved into the interpreter structure ... */

  Tcl_InitHashTable(&XOTclFilterList, TCL_ONE_WORD_KEYS);

  /*
   * and fill them with functionality
   */

  XOTclAddPMethod(in, (XOTclObject*)theobj, "alloc", XOTclOAllocMethod,0 , 0);
  XOTclAddIMethod(in, theobj, "info", XOTclOInfoMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "destroy", XOTclODestroyMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "class", XOTclOClassMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "unset", XOTclOUnsetMethod, 0, 0);
  /*XOTclAddIMethod(in, theobj, "init", XOTclOInitMethod, 0, 0);*/
  XOTclAddIMethod(in, theobj, "proc", XOTclOProcMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "isobject", XOTclOIsObjectMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "isclass", XOTclOIsClassMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "ismetaclass", XOTclOIsMetaClassMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "set",  (Tcl_ObjCmdProc*)XOTclOSetMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "instvar", (Tcl_ObjCmdProc*)XOTclOInstVarMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "move", (Tcl_ObjCmdProc*)XOTclOMoveMethod,0,0);
  XOTclAddIMethod(in, theobj, "invar", (Tcl_ObjCmdProc*)XOTclOInvariantsMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "check", (Tcl_ObjCmdProc*)XOTclOCheckMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "autoname", XOTclOAutonameMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "metadata", (Tcl_ObjCmdProc*)XOTclOMetaDataMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "copy", (Tcl_ObjCmdProc*) XOTclOCopyMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "mixin", (Tcl_ObjCmdProc*)XOTclOMixinMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "procsearch", (Tcl_ObjCmdProc*)XOTclOProcSearchMethod, 0, 0);
  XOTclAddIMethod(in, theobj, "incr", (Tcl_ObjCmdProc*)XOTclOIncrMethod, 0, 0);

  XOTclAddPMethod(in, (XOTclObject*)thecls, "alloc", (Tcl_ObjCmdProc*)XOTclCAllocMethod,0, 0);
  XOTclAddIMethod(in, thecls, "superclass", (Tcl_ObjCmdProc*)XOTclCSuperClassMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "info", XOTclCInfoMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "copy", XOTclCCopyMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "parameters", (Tcl_ObjCmdProc*)XOTclCParametersMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "parameterclass", (Tcl_ObjCmdProc*)XOTclCParameterClassMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "move", (Tcl_ObjCmdProc*)XOTclCMoveMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "create", (Tcl_ObjCmdProc*)XOTclCCreateMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instproc", (Tcl_ObjCmdProc*)XOTclCInstProcMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "instinvar", (Tcl_ObjCmdProc*)XOTclCInvariantsMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "filter", (Tcl_ObjCmdProc*)XOTclCFilterMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "filtersearch", (Tcl_ObjCmdProc*)XOTclCFilterSearchMethod, 0, 0);
  XOTclAddIMethod(in, thecls, "unknown", (Tcl_ObjCmdProc*) XOTclCUnknownMethod, 0, 0);

  /* 
   * overwritten tcl obj cmds 
   */
  Tcl_CreateObjCommand(in, "rename", XOTcl_RenameObjCmd, 0, 0);
  Tcl_CreateObjCommand(in, "info", XOTcl_InfoObjCmd, 0, 0);
  Tcl_CreateObjCommand(in, "upvar", XOTcl_UpvarObjCmd, 0, 0);
  Tcl_CreateObjCommand(in, "uplevel", XOTcl_UplevelObjCmd, 0, 0);
  
  /* 
   * new tcl cmds
   */
  Tcl_CreateObjCommand(in, "xotcltrace", XOTcl_TraceObjCmd, 0, 0);

  /*
   *  Parameter Class
   */
  paramCl = PrimitiveCCreate(in, global_strings[PARAM_CL], thecls);
  XOTclAddIMethod(in, paramCl, "default", (Tcl_ObjCmdProc*) ParameterDefaultMethod, 0, 0);

  /*
   * with some methods and library procs in tcl - they could go in a
   * otcl.tcl file, but they're embedded here with Tcl_GlobalEval to avoid
   * the need to carry around a separate library.
   */
  (void)strcpy(tm,
"Object instproc array {opt ary args} {\n\
   [self] instvar $ary\n\
   eval ::array [list $opt] [list $ary] $args\n\
}\n\
Object instproc cset {vn arg} {\n\
  if {![::info exists [self]::$vn]} {\n\
    ::set [self]::$vn $arg\n\
  }\n\
}\n\
Object instproc filterappend f {\n\
  [self] filter [concat [[self] info filters] $f]\n\
}\n\
Object instproc mixinappend m {\n\
  [self] mixin [concat [[self] info mixins] $m]\n\
}\n\
Object proc getExitHandler {} {\n\
  if {[info exists [self]::__exitHandler]} {\n\
    return [[self] set __exitHandler]\n\
  } else {\n\
    return ""\n\
  }\n\
}\n\
Object proc setExitHandler body {\n\
  return [[self] set __exitHandler $body]\n\
}\n\
Object proc unsetExitHandler {} {\n\
  [self] unset __exitHandler\n\
}\n\
Class::Parameter instproc values {param args} {\n\
  set ci [[self] info instinvar]\n\
  set valueTest {}\n\
  foreach a $args {\n\
    lappend valueTest \"\\[\\[self\\] set $param\\] == [list $a]\"\n\
  }\n\
  lappend ci [join $valueTest \" || \"]\n\
  [self] instinvar $ci\n\
}\n\
Object instproc abstract {methtype methname arglist} {\n\
  if {$methtype  != \"proc\" && $methtype != \"instproc\"} {\n\
    error \"invalid method type '$methtype',\n\
	must be either 'proc' or 'instproc'.\"\n\
  }\n\
  [self] $methtype $methname $arglist \\\n\
      [list error \"Abstract method $methname $arglist called\"]\n\
}\n\
");
  if (Tcl_GlobalEval(in, tm) != TCL_OK) return TCL_ERROR;

  (void)strcpy(tm, "proc xotcl_load {obj file} {\n");
  (void)strcat(tm, "   source $file\n");
  (void)strcat(tm, "   foreach i [array names ::auto_index\\\n");
  (void)strcat(tm, "       [list $obj *proc *]] {\n");
  (void)strcat(tm, "     set type [lindex $i 1]\n");
  (void)strcat(tm, "     set meth [lindex $i 2]\n");
  (void)strcat(tm, "     if {[$obj info ${type}s $meth] == {}} {\n");
  (void)strcat(tm, "       $obj $type $meth {auto} $::auto_index($i)\n");
  (void)strcat(tm, "     }\n");
  (void)strcat(tm, "   }\n");
  (void)strcat(tm, " }\n");
  if (Tcl_GlobalEval(in, tm) != TCL_OK) return TCL_ERROR;

  (void)strcpy(tm, "proc xotcl_mkindex {meta dir args} {\n");
  (void)strcat(tm, "  set sp {[         ]+}\n");
  (void)strcat(tm, "  set st {^[        ]*}\n");
  (void)strcat(tm, "  set wd {([^       ]+)}\n");
  (void)strcat(tm, "  foreach creator $meta {\n");
  (void)strcat(tm, "    lappend cp \"$st$creator${sp}create$sp$wd\"\n");
  (void)strcat(tm, "    lappend ap \"$st$creator$sp$wd\"\n");
  (void)strcat(tm, "  }\n");
  (void)strcat(tm, "  foreach method {proc instproc} {\n");
  (void)strcat(tm, "    lappend mp \"$st$wd${sp}($method)$sp$wd\"\n");
  (void)strcat(tm, "  }\n");
  (void)strcat(tm, "  foreach cl [concat Class [Class info heritage]] {\n");
  (void)strcat(tm, "    eval lappend meths [$cl info instcommands]\n");
  (void)strcat(tm, "  }\n");
  (void)strcat(tm, "  set old [pwd]\n");
  (void)strcat(tm, "  cd $dir\n");
  (void)strcat(tm, "  append idx \"# Tcl autoload index file, \"\n");
  (void)strcat(tm, "  append idx \"version 2.0\\n\"\n");
  (void)strcat(tm, "  append idx \"# xotcl additions generated with \"\n");
  (void)strcat(tm, "  append idx \"\\\"xotcl_mkindex [list $meta] \"\n");
  (void)strcat(tm, "  append idx \"[list $dir] $args\\\"\\n\"\n");
  (void)strcat(tm, "  set oc 0\n");
  (void)strcat(tm, "  set mc 0\n");
  (void)strcat(tm, "  foreach file [eval glob -nocomplain -- $args] {\n");
  (void)strcat(tm, "    if {[catch {set f [open $file]} msg]} then {\n");
  (void)strcat(tm, "      catch {close $f}\n");
  (void)strcat(tm, "      cd $old\n");
  (void)strcat(tm, "      error $msg\n");
  (void)strcat(tm, "    }\n");
  (void)strcat(tm, "    while {[gets $f line] >= 0} {\n");
  (void)strcat(tm, "      foreach c $cp {\n");
  (void)strcat(tm, "        if {[regexp $c $line x obj]==1 &&\n");
  (void)strcat(tm, "            [string index $obj 0]!={$}} then {\n");
  (void)strcat(tm, "          incr oc\n");
  (void)strcat(tm, "          append idx \"set auto_index($obj) \"\n");
  (void)strcat(tm, "          append idx \"\\\"xotcl_load $obj \"\n");
  (void)strcat(tm, "          append idx \"\\$dir/$file\\\"\\n\"\n");
  (void)strcat(tm, "        }\n");
  (void)strcat(tm, "      }\n");
  (void)strcat(tm, "      foreach a $ap {\n");
  (void)strcat(tm, "        if {[regexp $a $line x obj]==1 &&\n");
  (void)strcat(tm, "            [string index $obj 0]!={$} &&\n");
  (void)strcat(tm, "            [lsearch -exact $meths $obj]==-1} {\n");
  (void)strcat(tm, "          incr oc\n");
  (void)strcat(tm, "          append idx \"set auto_index($obj) \"\n");
  (void)strcat(tm, "          append idx \"\\\"xotcl_load $obj \"\n");
  (void)strcat(tm, "          append idx \"\\$dir/$file\\\"\\n\"\n");
  (void)strcat(tm, "        }\n");
  (void)strcat(tm, "      }\n");
  (void)strcat(tm, "      foreach m $mp {\n");
  (void)strcat(tm, "        if {[regexp $m $line x obj ty pr]==1 &&\n");
  (void)strcat(tm, "            [string index $obj 0]!={$} &&\n");
  (void)strcat(tm, "            [string index $pr 0]!={$}} then {\n");
  (void)strcat(tm, "            incr mc\n");
  (void)strcat(tm, "            append idx \"set \\{auto_index($obj \"\n");
  (void)strcat(tm, "            append idx \"$ty $pr)\\} \\\"source \"\n");
  (void)strcat(tm, "            append idx \"\\$dir/$file\\\"\\n\"\n");
  (void)strcat(tm, "        }\n");
  (void)strcat(tm, "      }\n");
  (void)strcat(tm, "    }\n");
  (void)strcat(tm, "    close $f\n");
  (void)strcat(tm, "  }\n");
  (void)strcat(tm, "  set t [open tclIndex a+]\n");
  (void)strcat(tm, "  puts $t $idx nonewline\n");
  (void)strcat(tm, "  close $t\n");
  (void)strcat(tm, "  cd $old\n");
  (void)strcat(tm, "  return \"$oc objects, $mc methods\"\n");
  (void)strcat(tm, "}\n");
  if (Tcl_GlobalEval(in, tm) != TCL_OK) return TCL_ERROR;

  /* insert xotcl lib path after TCLLIBPATH if set */
  (void)sprintf(tm, "set auto_path [linsert $auto_path [expr {[llength $auto_path]-3}] %s]", XOLIBPKG);
  if (Tcl_GlobalEval(in, tm) != TCL_OK) return TCL_ERROR;

  Tcl_PkgProvide(in, "XOTcl", XOTCLVERSION);
  Tcl_SetObjResult(in,global_objects[ONE]);

  return TCL_OK;
}
